/*
  Copyright (c) 2007 Konosuke Watanabe <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <spu_mfcio.h>
#include <spu_intrinsics.h>
#include <stdio.h>
#include <string.h>

#include "spe_crypt.h"

static unsigned char sbox_0_1[4096] = SBOX_TBL_0_1;
static unsigned char sbox_2_3[4096] = SBOX_TBL_2_3;
static unsigned char sbox_4_5[4096] = SBOX_TBL_4_5;
static unsigned char sbox_6_7[4096] = SBOX_TBL_6_7;


static void
dump_vector(unsigned char *head, vec_uchar16 vec, int num)
{
    int i;
    unsigned char c;

    for (i = 0; i < num && i < 128; i ++) {
        c = (spu_extract(vec, i / 8) >> (i % 8)) & 1;

        if (i % 64 == 0)
            printf("%s: ", head);

        printf("%d", c);
        if (i == 63 && num > 64)
            printf("\n");
        else if (i % 8 == 7)
            printf(" ");
    }
    printf("\n");
}


void
dump_vector_lsb(unsigned char *head, vec_uchar16 vec, int num)
{
    int i;
    unsigned char c;
    
    printf("%s: ", head);

    for (i = 0; i < num && i < 16; i ++) {
        c = spu_extract(vec, i) & 1;
        printf("%d", c);
        if (i % 8 == 7)
            printf(" ");
    }
    printf("\n");
}



#define VZERO (vec_uchar16){0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}


/*
  uchar٥ȥ뤫顤16bitФơ
  uchar٥ȥγǤLSBݻ
*/
#define SCATTER_BIT_TO_LSB(_v, _shuffle, _mask)                     \
    spu_cntb(spu_and(spu_shuffle((_v), (_v), (_shuffle)), (_mask)))

/*
 spu_gatherη̤ (vector unsigned int) 
 Υ٥ȥΡ0β(LSB¦)16bit˥ե饰Ĥޤ
 äơϰʲΤ褦֤

 -------0 -------1 -------2 -------3 -------4 -------5 ...
                            

 ɥ쥹Ƭ                                      
 -------- -------- 01234567 89ABCDEF -------- -------- ...
*/

#define VGATHER (vec_uchar16){7, 6, 5, 4, 3, 2, 1, 0, 15, 14, 13, 12, 11, 10, 9, 8}

#define GATHER_BIT_FROM_LSB(_v)                     \
    spu_gather(spu_shuffle((_v), (_v), VGATHER))

#define VREV (vec_uchar16){15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0}


#define VPACK01 (vec_uchar16){2, 3, 18, 19, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
#define VPACK23 (vec_uchar16){0, 0, 0, 0, 2, 3, 18, 19, 0, 0, 0, 0, 0, 0, 0, 0}
#define VPACK45 (vec_uchar16){0, 0, 0, 0, 0, 0, 0, 0, 2, 3, 18, 19, 0, 0, 0, 0}

/*
  saltʸ
*/
static inline unsigned int
check_salt(unsigned char c)
{
    int out;

    out = (c < '.') 
        | ((c > '9') & (c < 'A'))
        | ((c > 'Z') & (c < 'a'))
        | (c > 'z');
   
    c = out ? '.' : c;

    c = c > 'Z' ? (c - 6) : c;
    c = c > '9' ? (c - 7) : c;

    c -= '.';

    return c;
}


/*
  
  ŬʤȤ 
  spu_write_decrementer(0);
  ȤäƻϤơŬʤȤ
  t0 += -spu_read_decrementer();
  Ȥäƻߤ봶
 */
#ifdef TIMER1
static int t0 = 0, t1 = 0, t2 = 0, t3 = 0, t4 = 0;
#endif




static vec_uchar16 vcrypt;


static inline char *
simd_crypt(const unsigned char *salt, const unsigned char *password)
{
    int i, j;

    unsigned int isalt;

    vec_uchar16 vpass;

    vec_uchar16 k0[16], k1[16], k2[16];

    /* ֥åϺǽ0 */
    vec_uchar16 vblock = VZERO;

    vec_uchar16 vl0, vl1, vr0, vr1;

    static unsigned char salt_cache[2] = {0};
    static vec_uchar16 veshfl0, veshfl1, veshfl2, vemask0, vemask1, vemask2;
    vec_uchar16 veswap;

    vec_uchar16 vdot = spu_splats((unsigned char)'.');
    vec_uchar16 vcnine = spu_splats((unsigned char)'9');
    vec_uchar16 vseven = spu_splats((unsigned char)7);
    vec_uchar16 vsix = spu_splats((unsigned char)6);
    vec_uchar16 vlz = spu_splats((unsigned char)'Z');
    vec_uchar16 vfinmask = spu_splats((unsigned char)0x3f);
    
    vec_uchar16 vb0, vb1, vb2, vb3;    

    vec_uchar16 vfin0, vfin1, vfin2, vfin3, vfin4, vfin5;
    vec_uchar16 vflag;


    /* ޤǥѥɤ٥ȥ */
    vpass = *(vec_uchar16 *)password;
    /*
     ξ֤Ǥ
      - 0bit7(LSB) ... 1ʸܤκǲbit
      - 0bit6      ... 1ʸܤβ2ܤbit
      - 0bit0      ... 1ʸܤκǾbit
      - 1bit7      ... 2ʸܤκǲbit
    */


    /* LSBΤߤѤ */
    k0[0] = SCATTER_BIT_TO_LSB(vpass, KSHFL0_0, KMASK0_0);
    k1[0] = SCATTER_BIT_TO_LSB(vpass, KSHFL0_1, KMASK0_1);
    k2[0] = SCATTER_BIT_TO_LSB(vpass, KSHFL0_2, KMASK0_2);

    k0[1] = SCATTER_BIT_TO_LSB(vpass, KSHFL1_0, KMASK1_0);
    k1[1] = SCATTER_BIT_TO_LSB(vpass, KSHFL1_1, KMASK1_1);
    k2[1] = SCATTER_BIT_TO_LSB(vpass, KSHFL1_2, KMASK1_2);

    k0[2] = SCATTER_BIT_TO_LSB(vpass, KSHFL2_0, KMASK2_0);
    k1[2] = SCATTER_BIT_TO_LSB(vpass, KSHFL2_1, KMASK2_1);
    k2[2] = SCATTER_BIT_TO_LSB(vpass, KSHFL2_2, KMASK2_2);

    k0[3] = SCATTER_BIT_TO_LSB(vpass, KSHFL3_0, KMASK3_0);
    k1[3] = SCATTER_BIT_TO_LSB(vpass, KSHFL3_1, KMASK3_1);
    k2[3] = SCATTER_BIT_TO_LSB(vpass, KSHFL3_2, KMASK3_2);

    k0[4] = SCATTER_BIT_TO_LSB(vpass, KSHFL4_0, KMASK4_0);
    k1[4] = SCATTER_BIT_TO_LSB(vpass, KSHFL4_1, KMASK4_1);
    k2[4] = SCATTER_BIT_TO_LSB(vpass, KSHFL4_2, KMASK4_2);

    k0[5] = SCATTER_BIT_TO_LSB(vpass, KSHFL5_0, KMASK5_0);
    k1[5] = SCATTER_BIT_TO_LSB(vpass, KSHFL5_1, KMASK5_1);
    k2[5] = SCATTER_BIT_TO_LSB(vpass, KSHFL5_2, KMASK5_2);

    k0[6] = SCATTER_BIT_TO_LSB(vpass, KSHFL6_0, KMASK6_0);
    k1[6] = SCATTER_BIT_TO_LSB(vpass, KSHFL6_1, KMASK6_1);
    k2[6] = SCATTER_BIT_TO_LSB(vpass, KSHFL6_2, KMASK6_2);

    k0[7] = SCATTER_BIT_TO_LSB(vpass, KSHFL7_0, KMASK7_0);
    k1[7] = SCATTER_BIT_TO_LSB(vpass, KSHFL7_1, KMASK7_1);
    k2[7] = SCATTER_BIT_TO_LSB(vpass, KSHFL7_2, KMASK7_2);

    k0[8] = SCATTER_BIT_TO_LSB(vpass, KSHFL8_0, KMASK8_0);
    k1[8] = SCATTER_BIT_TO_LSB(vpass, KSHFL8_1, KMASK8_1);
    k2[8] = SCATTER_BIT_TO_LSB(vpass, KSHFL8_2, KMASK8_2);

    k0[9] = SCATTER_BIT_TO_LSB(vpass, KSHFL9_0, KMASK9_0);
    k1[9] = SCATTER_BIT_TO_LSB(vpass, KSHFL9_1, KMASK9_1);
    k2[9] = SCATTER_BIT_TO_LSB(vpass, KSHFL9_2, KMASK9_2);

    k0[10] = SCATTER_BIT_TO_LSB(vpass, KSHFL10_0, KMASK10_0);
    k1[10] = SCATTER_BIT_TO_LSB(vpass, KSHFL10_1, KMASK10_1);
    k2[10] = SCATTER_BIT_TO_LSB(vpass, KSHFL10_2, KMASK10_2);

    k0[11] = SCATTER_BIT_TO_LSB(vpass, KSHFL11_0, KMASK11_0);
    k1[11] = SCATTER_BIT_TO_LSB(vpass, KSHFL11_1, KMASK11_1);
    k2[11] = SCATTER_BIT_TO_LSB(vpass, KSHFL11_2, KMASK11_2);

    k0[12] = SCATTER_BIT_TO_LSB(vpass, KSHFL12_0, KMASK12_0);
    k1[12] = SCATTER_BIT_TO_LSB(vpass, KSHFL12_1, KMASK12_1);
    k2[12] = SCATTER_BIT_TO_LSB(vpass, KSHFL12_2, KMASK12_2);

    k0[13] = SCATTER_BIT_TO_LSB(vpass, KSHFL13_0, KMASK13_0);
    k1[13] = SCATTER_BIT_TO_LSB(vpass, KSHFL13_1, KMASK13_1);
    k2[13] = SCATTER_BIT_TO_LSB(vpass, KSHFL13_2, KMASK13_2);

    k0[14] = SCATTER_BIT_TO_LSB(vpass, KSHFL14_0, KMASK14_0);
    k1[14] = SCATTER_BIT_TO_LSB(vpass, KSHFL14_1, KMASK14_1);
    k2[14] = SCATTER_BIT_TO_LSB(vpass, KSHFL14_2, KMASK14_2);

    k0[15] = SCATTER_BIT_TO_LSB(vpass, KSHFL15_0, KMASK15_0);
    k1[15] = SCATTER_BIT_TO_LSB(vpass, KSHFL15_1, KMASK15_1);
    k2[15] = SCATTER_BIT_TO_LSB(vpass, KSHFL15_2, KMASK15_2);

    
    /*
      žEsaltؤ

       Eϳž֤ԤΥӥåȰ֤Υꥹ

       Eϸ48ʬǡ011bit  2435bit ιԤ褬
       saltˤäؤ

       4ĤΥ٥ȥǡԤɽ

       E0: ******** ****++++  E1: ++++++++ --------
       E2: ******** ****++++  E3: ++++++++ --------

       *: ؤǽΤ
       +: ؤʤ
       -: ߡ

       ؤϡsaltγbitbyte˳ĥ 0x00 or 0xff Υޥˤ塤
       spu_selƤ֤ȤǹԤ
       - saltĥȡؤ٤Ȥ 0xff ˡ
         ǤʤȤ0x00ˤʤ
       - spu_selǤϡ0Υ٥ȥ뤫顤1ǸΥ٥ȥ뤫ФƤ
    */


    /* ޤӤȤsaltѤʤȤ */
    if (__builtin_expect(((salt[0] != salt_cache[0]) 
                          | (salt[1] != salt_cache[1])), 0)) {
        isalt = (check_salt(salt[1]) << 6) | check_salt(salt[0]);
        veswap = spu_shuffle(spu_maskb(isalt), spu_maskb(isalt), VREV);

        veshfl0 = spu_sel(ESHFL0_BASE, ESHFL2_BASE, veswap);
        veshfl2 = spu_sel(ESHFL2_BASE, ESHFL0_BASE, veswap);

        vemask0 = spu_sel(EMASK0_BASE, EMASK2_BASE, veswap);
        vemask2 = spu_sel(EMASK2_BASE, EMASK0_BASE, veswap);
        
        salt_cache[0] = salt[0];
        salt_cache[1] = salt[1];

        /* ǹԤؤλ */

#define SHFL_CONCAT_0_8_0_8                                             \
        (vec_uchar16){0, 1, 2, 3, 4, 5, 6, 7, 16, 17, 18, 19, 20, 21, 22, 23}

#define SHFL_CONCAT_8_8_0_8                                             \
        (vec_uchar16){8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23}

        /*
          ˡؤǼºݤ˻Ȥ褦ˡ3٥ȥ˵ͤ
          ƬΥ٥ȥϴˤĤޤäƤΤǡ
          E18byteE28byteͤᡤ
          E2θ8byteE38byte򤯤äĤE2ˤ
        */

        veshfl1 = spu_shuffle(ESHFL1_BASE, veshfl2, SHFL_CONCAT_0_8_0_8);
        vemask1 = spu_shuffle(EMASK1_BASE, vemask2, SHFL_CONCAT_0_8_0_8);
    
        veshfl2 = spu_shuffle(veshfl2, ESHFL3_BASE, SHFL_CONCAT_8_8_0_8);
        vemask2 = spu_shuffle(vemask2, EMASK3_BASE, SHFL_CONCAT_8_8_0_8);
    }


    for (j = 0; __builtin_expect((j < 25), 1); j ++) {
        vec_uchar16 vl_next0, vl_next1;
        vec_uchar16 vf;
        vec_uchar16 vrtmp;


#define SBOX_KEY_01 (vec_uchar16){31, 31, 31, 31, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0}

#define SBOX_KEY_23 (vec_uchar16){31, 31, 31, 31, 23, 22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 12}

#define SBOX_KEY_45 (vec_uchar16){31, 31, 31, 31, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8}

#define SBOX_KEY_67 (vec_uchar16){31, 31, 31, 31, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4}


        const vec_uchar16 sbox_key_01 = SBOX_KEY_01, 
            sbox_key_23 = SBOX_KEY_23, 
            sbox_key_45 = SBOX_KEY_45,
            sbox_key_67 = SBOX_KEY_67;

        /* ֥åž (ӥåȤ¤ʾ) */
        vl0 = SCATTER_BIT_TO_LSB(vblock, IPSHFL, IPLMASK0);
        vl1 = SCATTER_BIT_TO_LSB(vblock, IPSHFL, IPLMASK1);

        vr0 = SCATTER_BIT_TO_LSB(vblock, IPSHFL, IPRMASK0);
        vr1 = SCATTER_BIT_TO_LSB(vblock, IPSHFL, IPRMASK1);

        for (i = 0; __builtin_expect((i < 16), 1); i ++) {

            vec_uchar16 vpres0, vpres1, vpres2;
            vec_uchar16 vf1, vf2, vf3, vf4;
            unsigned int s01, s23, s45, s67;

            vl_next0 = vr0;
            vl_next1 = vr1;

            /*  vr0̡vr1̤ˤ褦ˤLSB鸵ΰ֤إѥå */
            vrtmp = (vec_uchar16)spu_shuffle(GATHER_BIT_FROM_LSB(vr0), 
                                             GATHER_BIT_FROM_LSB(vr1),
                                             VPACK01);

            /* RnEǳž֤ơθKnxor */
            vpres0 = spu_xor(SCATTER_BIT_TO_LSB(vrtmp, veshfl0, vemask0),
                             k0[i]);

            vpres1 = spu_xor(SCATTER_BIT_TO_LSB(vrtmp, veshfl1, vemask1),
                             k1[i]);

            vpres2 = spu_xor(SCATTER_BIT_TO_LSB(vrtmp, veshfl2, vemask2),
                             k2[i]);


            /* 12bitļФơS8bitѴ */
            s01 = spu_extract(spu_gather(spu_shuffle(vpres0, VZERO,
                                                     sbox_key_01)), 0);
            s23 = spu_extract(spu_gather(spu_shuffle(vpres0, vpres1, 
                                                     sbox_key_23)), 0);
            s45 = spu_extract(spu_gather(spu_shuffle(vpres1, vpres2, 
                                                     sbox_key_45)), 0);
            s67 = spu_extract(spu_gather(spu_shuffle(vpres2, VZERO,
                                                     sbox_key_67)), 0);

            vf1 = spu_insert(sbox_0_1[s01 & 0xfff], VZERO, 0);
            vf2 = spu_insert(sbox_2_3[s23 & 0xfff], VZERO, 1);
            vf3 = spu_insert(sbox_4_5[s45 & 0xfff], VZERO, 2);
            vf4 = spu_insert(sbox_6_7[s67 & 0xfff], VZERO, 3);

            vf = spu_or(spu_or(vf1, vf2), spu_or(vf3, vf4));

                                            
            /* Pž */
            vr0 = spu_xor(vl0, SCATTER_BIT_TO_LSB(vf, PSHFL0, PMASK0));
            vr1 = spu_xor(vl1, SCATTER_BIT_TO_LSB(vf, PSHFL1, PMASK1));

            vl0 = vl_next0;
            vl1 = vl_next1;



        }

        /* 
           spu_rlqwbytespu_rlmaskqwbytebyteñ̤ǥեȤor
           åե뤷®롥
           ȸ䤹
         */

        vblock = (vec_uchar16)spu_or(spu_shuffle(GATHER_BIT_FROM_LSB(vr0), 
                                                 GATHER_BIT_FROM_LSB(vr1),
                                                 VPACK01),
                                     spu_shuffle(GATHER_BIT_FROM_LSB(vl0),
                                                 GATHER_BIT_FROM_LSB(vl1),
                                                 VPACK23));
        
        /* ǽž */
        vb0 = SCATTER_BIT_TO_LSB(vblock, FPSHFL, FPMASK0);
        vb1 = SCATTER_BIT_TO_LSB(vblock, FPSHFL, FPMASK1);
        vb2 = SCATTER_BIT_TO_LSB(vblock, FPSHFL, FPMASK2);
        vb3 = SCATTER_BIT_TO_LSB(vblock, FPSHFL, FPMASK3);

        vblock = (vec_uchar16)spu_or(spu_shuffle(GATHER_BIT_FROM_LSB(vb0), 
                                                 GATHER_BIT_FROM_LSB(vb1),
                                                 VPACK01),
                                     spu_shuffle(GATHER_BIT_FROM_LSB(vb2),
                                                 GATHER_BIT_FROM_LSB(vb3),
                                                 VPACK23));

    }





#define SHFL_FINAL0 (vec_uchar16){31, 31, 0, 1, 2, 3, 4, 5, 31, 31, 6, 7, 8, 9, 10, 11}
#define SHFL_FINAL1 (vec_uchar16){31, 31, 12, 13, 14, 15, 16, 17, 31, 31, 18, 19, 20, 21, 22, 23}
#define SHFL_FINAL2 (vec_uchar16){31, 31, 8, 9, 10, 11, 12, 13, 31, 31, 14, 15, 16, 17, 18, 19}

#define SHFL_FINAL3 (vec_uchar16){31, 31, 4, 5, 6, 7, 8, 9, 31, 31, 10, 11, 12, 13, 14, 15}


    vfin0 = (vec_uchar16)spu_or(spu_shuffle(spu_gather(spu_shuffle(vb0, vb0, 
                                                                   SHFL_FINAL0)),
                                            spu_gather(spu_shuffle(vb0, vb1, 
                                                                   SHFL_FINAL1)),
                                            VPACK01),
                                spu_shuffle(spu_gather(spu_shuffle(vb1, vb2, 
                                                                   SHFL_FINAL2)), 
                                            spu_gather(spu_shuffle(vb2, vb2,
                                                                   SHFL_FINAL3)),
                                            VPACK23));

    vfin1 = (vec_uchar16)spu_shuffle(spu_gather(spu_shuffle(vb3, vb3, SHFL_FINAL0)),
                                     spu_gather(spu_shuffle(vb3, VZERO, SHFL_FINAL1)),
                                     VPACK45);

    vfin2 = spu_and(vfinmask, spu_or(vfin0, vfin1));



    vfin3 = (vec_uchar16)spu_add((vec_ushort8)vfin2, (vec_ushort8)vdot);

    /* > 9 */
    vflag = spu_cmpgt(vfin3, vcnine);

    vfin4 = (vec_uchar16)spu_add((vec_ushort8)vfin3,
                                 (vec_ushort8)spu_and(vflag, vseven));

    /* > Z */
    vflag = spu_cmpgt(vfin4, vlz);

    vfin5 = (vec_uchar16)spu_add((vec_ushort8)vfin4,
                                 (vec_ushort8)spu_and(vflag, vsix));


    vcrypt = vfin5;

   

    return (char *)&vcrypt;
}


int
main(unsigned long long argc, unsigned long long argv)
{
    unsigned char password[16] __attribute__((aligned(16)));
    unsigned char salt[2];
    int n, o, p;
    char *c;

    /*
    password[0] = 0xc9;
    password[1] = 0xcf;
    password[2] = 0xc8;
    password[3] = 0xba;
    */
    password[0] = 'I';
    password[1] = 'O';
    password[2] = 'H';
    password[3] = ':';
    password[4] = 'd';
    password[5] = 'a';
    password[6] = 'D';
    password[7] = '!';


    c = simd_crypt((unsigned char *)"..", password); 
    c[11] = '\0';
    printf("%s\n", (unsigned char *)c);


    salt[0] = password[2];
    salt[1] = password[3];

#if 1
    for (n = 0x20; __builtin_expect((n < 0x7f), 1); n ++) {
        password[5] = n;

        for (o = 0x20; __builtin_expect((o < 0x7f), 1); o ++) {
            password[6] = o;
        
            for (p = 0x20; __builtin_expect((p < 0x7f), 1); p ++) {
                password[7] = p;
                c = simd_crypt(&password[1], password); 
            }
        }
    }
#endif

#if 0
    password[0] = 'I';
    password[1] = 'O';
    password[2] = 'H';
    password[3] = ':';
    password[4] = 'd';
    password[5] = 'a';
    password[6] = 'D';
    password[7] = '!';

    c = simd_crypt((unsigned char *)"..", password); 
    c[11] = '\0';
    printf("%s\n", (unsigned char *)c);
#endif

#ifdef TIMER1
    printf("t0 = %d\n", t0);
    printf("t1 = %d\n", t1);
    printf("t2 = %d\n", t2);
    printf("t3 = %d\n", t3);
#endif

    return 0;
}
