/*
  Copyright (c) 2007 Konosuke Watanabe <sasugaanija@gmail.com>

  All rights reserved.

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef _SPE_CRYPT_H_
#define _SPE_CRYPT_H_

#define SBOX_TBL_0_1 {0xf7, 0xf2, 0xfc, 0xff, 0xf4, 0xfb, 0xfa, 0xfc, \
            0xfb, 0xf7, 0xf6, 0xf9, 0xfd, 0xf4, 0xf0, 0xfa,           \
            0xf2, 0xf8, 0xf5, 0xf3, 0xff, 0xf6, 0xf9, 0xf5,           \
            0xf8, 0xf1, 0xf3, 0xfe, 0xf1, 0xfd, 0xfe, 0xf0,           \
            0xf0, 0xff, 0xf5, 0xfa, 0xf7, 0xf2, 0xf9, 0xf5,           \
            0xfe, 0xf1, 0xf3, 0xfc, 0xfb, 0xf8, 0xfc, 0xf6,           \
            0xff, 0xf3, 0xf6, 0xfd, 0xf4, 0xf9, 0xfa, 0xf0,           \
            0xf2, 0xf4, 0xfd, 0xf7, 0xf8, 0xfe, 0xf1, 0xfb,           \
            0x07, 0x02, 0x0c, 0x0f, 0x04, 0x0b, 0x0a, 0x0c,           \
            0x0b, 0x07, 0x06, 0x09, 0x0d, 0x04, 0x00, 0x0a,           \
            0x02, 0x08, 0x05, 0x03, 0x0f, 0x06, 0x09, 0x05,           \
            0x08, 0x01, 0x03, 0x0e, 0x01, 0x0d, 0x0e, 0x00,           \
            0x00, 0x0f, 0x05, 0x0a, 0x07, 0x02, 0x09, 0x05,           \
            0x0e, 0x01, 0x03, 0x0c, 0x0b, 0x08, 0x0c, 0x06,           \
            0x0f, 0x03, 0x06, 0x0d, 0x04, 0x09, 0x0a, 0x00,           \
            0x02, 0x04, 0x0d, 0x07, 0x08, 0x0e, 0x01, 0x0b,           \
            0x97, 0x92, 0x9c, 0x9f, 0x94, 0x9b, 0x9a, 0x9c,           \
            0x9b, 0x97, 0x96, 0x99, 0x9d, 0x94, 0x90, 0x9a,           \
            0x92, 0x98, 0x95, 0x93, 0x9f, 0x96, 0x99, 0x95,           \
            0x98, 0x91, 0x93, 0x9e, 0x91, 0x9d, 0x9e, 0x90,           \
            0x90, 0x9f, 0x95, 0x9a, 0x97, 0x92, 0x99, 0x95,           \
            0x9e, 0x91, 0x93, 0x9c, 0x9b, 0x98, 0x9c, 0x96,           \
            0x9f, 0x93, 0x96, 0x9d, 0x94, 0x99, 0x9a, 0x90,           \
            0x92, 0x94, 0x9d, 0x97, 0x98, 0x9e, 0x91, 0x9b,           \
            0xa7, 0xa2, 0xac, 0xaf, 0xa4, 0xab, 0xaa, 0xac,           \
            0xab, 0xa7, 0xa6, 0xa9, 0xad, 0xa4, 0xa0, 0xaa,           \
            0xa2, 0xa8, 0xa5, 0xa3, 0xaf, 0xa6, 0xa9, 0xa5,           \
            0xa8, 0xa1, 0xa3, 0xae, 0xa1, 0xad, 0xae, 0xa0,           \
            0xa0, 0xaf, 0xa5, 0xaa, 0xa7, 0xa2, 0xa9, 0xa5,           \
            0xae, 0xa1, 0xa3, 0xac, 0xab, 0xa8, 0xac, 0xa6,           \
            0xaf, 0xa3, 0xa6, 0xad, 0xa4, 0xa9, 0xaa, 0xa0,           \
            0xa2, 0xa4, 0xad, 0xa7, 0xa8, 0xae, 0xa1, 0xab,           \
            0x67, 0x62, 0x6c, 0x6f, 0x64, 0x6b, 0x6a, 0x6c,           \
            0x6b, 0x67, 0x66, 0x69, 0x6d, 0x64, 0x60, 0x6a,           \
            0x62, 0x68, 0x65, 0x63, 0x6f, 0x66, 0x69, 0x65,           \
            0x68, 0x61, 0x63, 0x6e, 0x61, 0x6d, 0x6e, 0x60,           \
            0x60, 0x6f, 0x65, 0x6a, 0x67, 0x62, 0x69, 0x65,           \
            0x6e, 0x61, 0x63, 0x6c, 0x6b, 0x68, 0x6c, 0x66,           \
            0x6f, 0x63, 0x66, 0x6d, 0x64, 0x69, 0x6a, 0x60,           \
            0x62, 0x64, 0x6d, 0x67, 0x68, 0x6e, 0x61, 0x6b,           \
            0x57, 0x52, 0x5c, 0x5f, 0x54, 0x5b, 0x5a, 0x5c,           \
            0x5b, 0x57, 0x56, 0x59, 0x5d, 0x54, 0x50, 0x5a,           \
            0x52, 0x58, 0x55, 0x53, 0x5f, 0x56, 0x59, 0x55,           \
            0x58, 0x51, 0x53, 0x5e, 0x51, 0x5d, 0x5e, 0x50,           \
            0x50, 0x5f, 0x55, 0x5a, 0x57, 0x52, 0x59, 0x55,           \
            0x5e, 0x51, 0x53, 0x5c, 0x5b, 0x58, 0x5c, 0x56,           \
            0x5f, 0x53, 0x56, 0x5d, 0x54, 0x59, 0x5a, 0x50,           \
            0x52, 0x54, 0x5d, 0x57, 0x58, 0x5e, 0x51, 0x5b,           \
            0x37, 0x32, 0x3c, 0x3f, 0x34, 0x3b, 0x3a, 0x3c,           \
            0x3b, 0x37, 0x36, 0x39, 0x3d, 0x34, 0x30, 0x3a,           \
            0x32, 0x38, 0x35, 0x33, 0x3f, 0x36, 0x39, 0x35,           \
            0x38, 0x31, 0x33, 0x3e, 0x31, 0x3d, 0x3e, 0x30,           \
            0x30, 0x3f, 0x35, 0x3a, 0x37, 0x32, 0x39, 0x35,           \
            0x3e, 0x31, 0x33, 0x3c, 0x3b, 0x38, 0x3c, 0x36,           \
            0x3f, 0x33, 0x36, 0x3d, 0x34, 0x39, 0x3a, 0x30,           \
            0x32, 0x34, 0x3d, 0x37, 0x38, 0x3e, 0x31, 0x3b,           \
            0x97, 0x92, 0x9c, 0x9f, 0x94, 0x9b, 0x9a, 0x9c,           \
            0x9b, 0x97, 0x96, 0x99, 0x9d, 0x94, 0x90, 0x9a,           \
            0x92, 0x98, 0x95, 0x93, 0x9f, 0x96, 0x99, 0x95,           \
            0x98, 0x91, 0x93, 0x9e, 0x91, 0x9d, 0x9e, 0x90,           \
            0x90, 0x9f, 0x95, 0x9a, 0x97, 0x92, 0x99, 0x95,           \
            0x9e, 0x91, 0x93, 0x9c, 0x9b, 0x98, 0x9c, 0x96,           \
            0x9f, 0x93, 0x96, 0x9d, 0x94, 0x99, 0x9a, 0x90,           \
            0x92, 0x94, 0x9d, 0x97, 0x98, 0x9e, 0x91, 0x9b,           \
            0x17, 0x12, 0x1c, 0x1f, 0x14, 0x1b, 0x1a, 0x1c,           \
            0x1b, 0x17, 0x16, 0x19, 0x1d, 0x14, 0x10, 0x1a,           \
            0x12, 0x18, 0x15, 0x13, 0x1f, 0x16, 0x19, 0x15,           \
            0x18, 0x11, 0x13, 0x1e, 0x11, 0x1d, 0x1e, 0x10,           \
            0x10, 0x1f, 0x15, 0x1a, 0x17, 0x12, 0x19, 0x15,           \
            0x1e, 0x11, 0x13, 0x1c, 0x1b, 0x18, 0x1c, 0x16,           \
            0x1f, 0x13, 0x16, 0x1d, 0x14, 0x19, 0x1a, 0x10,           \
            0x12, 0x14, 0x1d, 0x17, 0x18, 0x1e, 0x11, 0x1b,           \
            0xe7, 0xe2, 0xec, 0xef, 0xe4, 0xeb, 0xea, 0xec,           \
            0xeb, 0xe7, 0xe6, 0xe9, 0xed, 0xe4, 0xe0, 0xea,           \
            0xe2, 0xe8, 0xe5, 0xe3, 0xef, 0xe6, 0xe9, 0xe5,           \
            0xe8, 0xe1, 0xe3, 0xee, 0xe1, 0xed, 0xee, 0xe0,           \
            0xe0, 0xef, 0xe5, 0xea, 0xe7, 0xe2, 0xe9, 0xe5,           \
            0xee, 0xe1, 0xe3, 0xec, 0xeb, 0xe8, 0xec, 0xe6,           \
            0xef, 0xe3, 0xe6, 0xed, 0xe4, 0xe9, 0xea, 0xe0,           \
            0xe2, 0xe4, 0xed, 0xe7, 0xe8, 0xee, 0xe1, 0xeb,           \
            0x47, 0x42, 0x4c, 0x4f, 0x44, 0x4b, 0x4a, 0x4c,           \
            0x4b, 0x47, 0x46, 0x49, 0x4d, 0x44, 0x40, 0x4a,           \
            0x42, 0x48, 0x45, 0x43, 0x4f, 0x46, 0x49, 0x45,           \
            0x48, 0x41, 0x43, 0x4e, 0x41, 0x4d, 0x4e, 0x40,           \
            0x40, 0x4f, 0x45, 0x4a, 0x47, 0x42, 0x49, 0x45,           \
            0x4e, 0x41, 0x43, 0x4c, 0x4b, 0x48, 0x4c, 0x46,           \
            0x4f, 0x43, 0x46, 0x4d, 0x44, 0x49, 0x4a, 0x40,           \
            0x42, 0x44, 0x4d, 0x47, 0x48, 0x4e, 0x41, 0x4b,           \
            0x37, 0x32, 0x3c, 0x3f, 0x34, 0x3b, 0x3a, 0x3c,           \
            0x3b, 0x37, 0x36, 0x39, 0x3d, 0x34, 0x30, 0x3a,           \
            0x32, 0x38, 0x35, 0x33, 0x3f, 0x36, 0x39, 0x35,           \
            0x38, 0x31, 0x33, 0x3e, 0x31, 0x3d, 0x3e, 0x30,           \
            0x30, 0x3f, 0x35, 0x3a, 0x37, 0x32, 0x39, 0x35,           \
            0x3e, 0x31, 0x33, 0x3c, 0x3b, 0x38, 0x3c, 0x36,           \
            0x3f, 0x33, 0x36, 0x3d, 0x34, 0x39, 0x3a, 0x30,           \
            0x32, 0x34, 0x3d, 0x37, 0x38, 0x3e, 0x31, 0x3b,           \
            0xc7, 0xc2, 0xcc, 0xcf, 0xc4, 0xcb, 0xca, 0xcc,           \
            0xcb, 0xc7, 0xc6, 0xc9, 0xcd, 0xc4, 0xc0, 0xca,           \
            0xc2, 0xc8, 0xc5, 0xc3, 0xcf, 0xc6, 0xc9, 0xc5,           \
            0xc8, 0xc1, 0xc3, 0xce, 0xc1, 0xcd, 0xce, 0xc0,           \
            0xc0, 0xcf, 0xc5, 0xca, 0xc7, 0xc2, 0xc9, 0xc5,           \
            0xce, 0xc1, 0xc3, 0xcc, 0xcb, 0xc8, 0xcc, 0xc6,           \
            0xcf, 0xc3, 0xc6, 0xcd, 0xc4, 0xc9, 0xca, 0xc0,           \
            0xc2, 0xc4, 0xcd, 0xc7, 0xc8, 0xce, 0xc1, 0xcb,           \
            0xb7, 0xb2, 0xbc, 0xbf, 0xb4, 0xbb, 0xba, 0xbc,           \
            0xbb, 0xb7, 0xb6, 0xb9, 0xbd, 0xb4, 0xb0, 0xba,           \
            0xb2, 0xb8, 0xb5, 0xb3, 0xbf, 0xb6, 0xb9, 0xb5,           \
            0xb8, 0xb1, 0xb3, 0xbe, 0xb1, 0xbd, 0xbe, 0xb0,           \
            0xb0, 0xbf, 0xb5, 0xba, 0xb7, 0xb2, 0xb9, 0xb5,           \
            0xbe, 0xb1, 0xb3, 0xbc, 0xbb, 0xb8, 0xbc, 0xb6,           \
            0xbf, 0xb3, 0xb6, 0xbd, 0xb4, 0xb9, 0xba, 0xb0,           \
            0xb2, 0xb4, 0xbd, 0xb7, 0xb8, 0xbe, 0xb1, 0xbb,           \
            0xa7, 0xa2, 0xac, 0xaf, 0xa4, 0xab, 0xaa, 0xac,           \
            0xab, 0xa7, 0xa6, 0xa9, 0xad, 0xa4, 0xa0, 0xaa,           \
            0xa2, 0xa8, 0xa5, 0xa3, 0xaf, 0xa6, 0xa9, 0xa5,           \
            0xa8, 0xa1, 0xa3, 0xae, 0xa1, 0xad, 0xae, 0xa0,           \
            0xa0, 0xaf, 0xa5, 0xaa, 0xa7, 0xa2, 0xa9, 0xa5,           \
            0xae, 0xa1, 0xa3, 0xac, 0xab, 0xa8, 0xac, 0xa6,           \
            0xaf, 0xa3, 0xa6, 0xad, 0xa4, 0xa9, 0xaa, 0xa0,           \
            0xa2, 0xa4, 0xad, 0xa7, 0xa8, 0xae, 0xa1, 0xab,           \
            0x47, 0x42, 0x4c, 0x4f, 0x44, 0x4b, 0x4a, 0x4c,           \
            0x4b, 0x47, 0x46, 0x49, 0x4d, 0x44, 0x40, 0x4a,           \
            0x42, 0x48, 0x45, 0x43, 0x4f, 0x46, 0x49, 0x45,           \
            0x48, 0x41, 0x43, 0x4e, 0x41, 0x4d, 0x4e, 0x40,           \
            0x40, 0x4f, 0x45, 0x4a, 0x47, 0x42, 0x49, 0x45,           \
            0x4e, 0x41, 0x43, 0x4c, 0x4b, 0x48, 0x4c, 0x46,           \
            0x4f, 0x43, 0x46, 0x4d, 0x44, 0x49, 0x4a, 0x40,           \
            0x42, 0x44, 0x4d, 0x47, 0x48, 0x4e, 0x41, 0x4b,           \
            0x87, 0x82, 0x8c, 0x8f, 0x84, 0x8b, 0x8a, 0x8c,           \
            0x8b, 0x87, 0x86, 0x89, 0x8d, 0x84, 0x80, 0x8a,           \
            0x82, 0x88, 0x85, 0x83, 0x8f, 0x86, 0x89, 0x85,           \
            0x88, 0x81, 0x83, 0x8e, 0x81, 0x8d, 0x8e, 0x80,           \
            0x80, 0x8f, 0x85, 0x8a, 0x87, 0x82, 0x89, 0x85,           \
            0x8e, 0x81, 0x83, 0x8c, 0x8b, 0x88, 0x8c, 0x86,           \
            0x8f, 0x83, 0x86, 0x8d, 0x84, 0x89, 0x8a, 0x80,           \
            0x82, 0x84, 0x8d, 0x87, 0x88, 0x8e, 0x81, 0x8b,           \
            0x77, 0x72, 0x7c, 0x7f, 0x74, 0x7b, 0x7a, 0x7c,           \
            0x7b, 0x77, 0x76, 0x79, 0x7d, 0x74, 0x70, 0x7a,           \
            0x72, 0x78, 0x75, 0x73, 0x7f, 0x76, 0x79, 0x75,           \
            0x78, 0x71, 0x73, 0x7e, 0x71, 0x7d, 0x7e, 0x70,           \
            0x70, 0x7f, 0x75, 0x7a, 0x77, 0x72, 0x79, 0x75,           \
            0x7e, 0x71, 0x73, 0x7c, 0x7b, 0x78, 0x7c, 0x76,           \
            0x7f, 0x73, 0x76, 0x7d, 0x74, 0x79, 0x7a, 0x70,           \
            0x72, 0x74, 0x7d, 0x77, 0x78, 0x7e, 0x71, 0x7b,           \
            0xe7, 0xe2, 0xec, 0xef, 0xe4, 0xeb, 0xea, 0xec,           \
            0xeb, 0xe7, 0xe6, 0xe9, 0xed, 0xe4, 0xe0, 0xea,           \
            0xe2, 0xe8, 0xe5, 0xe3, 0xef, 0xe6, 0xe9, 0xe5,           \
            0xe8, 0xe1, 0xe3, 0xee, 0xe1, 0xed, 0xee, 0xe0,           \
            0xe0, 0xef, 0xe5, 0xea, 0xe7, 0xe2, 0xe9, 0xe5,           \
            0xee, 0xe1, 0xe3, 0xec, 0xeb, 0xe8, 0xec, 0xe6,           \
            0xef, 0xe3, 0xe6, 0xed, 0xe4, 0xe9, 0xea, 0xe0,           \
            0xe2, 0xe4, 0xed, 0xe7, 0xe8, 0xee, 0xe1, 0xeb,           \
            0x17, 0x12, 0x1c, 0x1f, 0x14, 0x1b, 0x1a, 0x1c,           \
            0x1b, 0x17, 0x16, 0x19, 0x1d, 0x14, 0x10, 0x1a,           \
            0x12, 0x18, 0x15, 0x13, 0x1f, 0x16, 0x19, 0x15,           \
            0x18, 0x11, 0x13, 0x1e, 0x11, 0x1d, 0x1e, 0x10,           \
            0x10, 0x1f, 0x15, 0x1a, 0x17, 0x12, 0x19, 0x15,           \
            0x1e, 0x11, 0x13, 0x1c, 0x1b, 0x18, 0x1c, 0x16,           \
            0x1f, 0x13, 0x16, 0x1d, 0x14, 0x19, 0x1a, 0x10,           \
            0x12, 0x14, 0x1d, 0x17, 0x18, 0x1e, 0x11, 0x1b,           \
            0xd7, 0xd2, 0xdc, 0xdf, 0xd4, 0xdb, 0xda, 0xdc,           \
            0xdb, 0xd7, 0xd6, 0xd9, 0xdd, 0xd4, 0xd0, 0xda,           \
            0xd2, 0xd8, 0xd5, 0xd3, 0xdf, 0xd6, 0xd9, 0xd5,           \
            0xd8, 0xd1, 0xd3, 0xde, 0xd1, 0xdd, 0xde, 0xd0,           \
            0xd0, 0xdf, 0xd5, 0xda, 0xd7, 0xd2, 0xd9, 0xd5,           \
            0xde, 0xd1, 0xd3, 0xdc, 0xdb, 0xd8, 0xdc, 0xd6,           \
            0xdf, 0xd3, 0xd6, 0xdd, 0xd4, 0xd9, 0xda, 0xd0,           \
            0xd2, 0xd4, 0xdd, 0xd7, 0xd8, 0xde, 0xd1, 0xdb,           \
            0x27, 0x22, 0x2c, 0x2f, 0x24, 0x2b, 0x2a, 0x2c,           \
            0x2b, 0x27, 0x26, 0x29, 0x2d, 0x24, 0x20, 0x2a,           \
            0x22, 0x28, 0x25, 0x23, 0x2f, 0x26, 0x29, 0x25,           \
            0x28, 0x21, 0x23, 0x2e, 0x21, 0x2d, 0x2e, 0x20,           \
            0x20, 0x2f, 0x25, 0x2a, 0x27, 0x22, 0x29, 0x25,           \
            0x2e, 0x21, 0x23, 0x2c, 0x2b, 0x28, 0x2c, 0x26,           \
            0x2f, 0x23, 0x26, 0x2d, 0x24, 0x29, 0x2a, 0x20,           \
            0x22, 0x24, 0x2d, 0x27, 0x28, 0x2e, 0x21, 0x2b,           \
            0x07, 0x02, 0x0c, 0x0f, 0x04, 0x0b, 0x0a, 0x0c,           \
            0x0b, 0x07, 0x06, 0x09, 0x0d, 0x04, 0x00, 0x0a,           \
            0x02, 0x08, 0x05, 0x03, 0x0f, 0x06, 0x09, 0x05,           \
            0x08, 0x01, 0x03, 0x0e, 0x01, 0x0d, 0x0e, 0x00,           \
            0x00, 0x0f, 0x05, 0x0a, 0x07, 0x02, 0x09, 0x05,           \
            0x0e, 0x01, 0x03, 0x0c, 0x0b, 0x08, 0x0c, 0x06,           \
            0x0f, 0x03, 0x06, 0x0d, 0x04, 0x09, 0x0a, 0x00,           \
            0x02, 0x04, 0x0d, 0x07, 0x08, 0x0e, 0x01, 0x0b,           \
            0xc7, 0xc2, 0xcc, 0xcf, 0xc4, 0xcb, 0xca, 0xcc,           \
            0xcb, 0xc7, 0xc6, 0xc9, 0xcd, 0xc4, 0xc0, 0xca,           \
            0xc2, 0xc8, 0xc5, 0xc3, 0xcf, 0xc6, 0xc9, 0xc5,           \
            0xc8, 0xc1, 0xc3, 0xce, 0xc1, 0xcd, 0xce, 0xc0,           \
            0xc0, 0xcf, 0xc5, 0xca, 0xc7, 0xc2, 0xc9, 0xc5,           \
            0xce, 0xc1, 0xc3, 0xcc, 0xcb, 0xc8, 0xcc, 0xc6,           \
            0xcf, 0xc3, 0xc6, 0xcd, 0xc4, 0xc9, 0xca, 0xc0,           \
            0xc2, 0xc4, 0xcd, 0xc7, 0xc8, 0xce, 0xc1, 0xcb,           \
            0x77, 0x72, 0x7c, 0x7f, 0x74, 0x7b, 0x7a, 0x7c,           \
            0x7b, 0x77, 0x76, 0x79, 0x7d, 0x74, 0x70, 0x7a,           \
            0x72, 0x78, 0x75, 0x73, 0x7f, 0x76, 0x79, 0x75,           \
            0x78, 0x71, 0x73, 0x7e, 0x71, 0x7d, 0x7e, 0x70,           \
            0x70, 0x7f, 0x75, 0x7a, 0x77, 0x72, 0x79, 0x75,           \
            0x7e, 0x71, 0x73, 0x7c, 0x7b, 0x78, 0x7c, 0x76,           \
            0x7f, 0x73, 0x76, 0x7d, 0x74, 0x79, 0x7a, 0x70,           \
            0x72, 0x74, 0x7d, 0x77, 0x78, 0x7e, 0x71, 0x7b,           \
            0xd7, 0xd2, 0xdc, 0xdf, 0xd4, 0xdb, 0xda, 0xdc,           \
            0xdb, 0xd7, 0xd6, 0xd9, 0xdd, 0xd4, 0xd0, 0xda,           \
            0xd2, 0xd8, 0xd5, 0xd3, 0xdf, 0xd6, 0xd9, 0xd5,           \
            0xd8, 0xd1, 0xd3, 0xde, 0xd1, 0xdd, 0xde, 0xd0,           \
            0xd0, 0xdf, 0xd5, 0xda, 0xd7, 0xd2, 0xd9, 0xd5,           \
            0xde, 0xd1, 0xd3, 0xdc, 0xdb, 0xd8, 0xdc, 0xd6,           \
            0xdf, 0xd3, 0xd6, 0xdd, 0xd4, 0xd9, 0xda, 0xd0,           \
            0xd2, 0xd4, 0xdd, 0xd7, 0xd8, 0xde, 0xd1, 0xdb,           \
            0xb7, 0xb2, 0xbc, 0xbf, 0xb4, 0xbb, 0xba, 0xbc,           \
            0xbb, 0xb7, 0xb6, 0xb9, 0xbd, 0xb4, 0xb0, 0xba,           \
            0xb2, 0xb8, 0xb5, 0xb3, 0xbf, 0xb6, 0xb9, 0xb5,           \
            0xb8, 0xb1, 0xb3, 0xbe, 0xb1, 0xbd, 0xbe, 0xb0,           \
            0xb0, 0xbf, 0xb5, 0xba, 0xb7, 0xb2, 0xb9, 0xb5,           \
            0xbe, 0xb1, 0xb3, 0xbc, 0xbb, 0xb8, 0xbc, 0xb6,           \
            0xbf, 0xb3, 0xb6, 0xbd, 0xb4, 0xb9, 0xba, 0xb0,           \
            0xb2, 0xb4, 0xbd, 0xb7, 0xb8, 0xbe, 0xb1, 0xbb,           \
            0x67, 0x62, 0x6c, 0x6f, 0x64, 0x6b, 0x6a, 0x6c,           \
            0x6b, 0x67, 0x66, 0x69, 0x6d, 0x64, 0x60, 0x6a,           \
            0x62, 0x68, 0x65, 0x63, 0x6f, 0x66, 0x69, 0x65,           \
            0x68, 0x61, 0x63, 0x6e, 0x61, 0x6d, 0x6e, 0x60,           \
            0x60, 0x6f, 0x65, 0x6a, 0x67, 0x62, 0x69, 0x65,           \
            0x6e, 0x61, 0x63, 0x6c, 0x6b, 0x68, 0x6c, 0x66,           \
            0x6f, 0x63, 0x66, 0x6d, 0x64, 0x69, 0x6a, 0x60,           \
            0x62, 0x64, 0x6d, 0x67, 0x68, 0x6e, 0x61, 0x6b,           \
            0x27, 0x22, 0x2c, 0x2f, 0x24, 0x2b, 0x2a, 0x2c,           \
            0x2b, 0x27, 0x26, 0x29, 0x2d, 0x24, 0x20, 0x2a,           \
            0x22, 0x28, 0x25, 0x23, 0x2f, 0x26, 0x29, 0x25,           \
            0x28, 0x21, 0x23, 0x2e, 0x21, 0x2d, 0x2e, 0x20,           \
            0x20, 0x2f, 0x25, 0x2a, 0x27, 0x22, 0x29, 0x25,           \
            0x2e, 0x21, 0x23, 0x2c, 0x2b, 0x28, 0x2c, 0x26,           \
            0x2f, 0x23, 0x26, 0x2d, 0x24, 0x29, 0x2a, 0x20,           \
            0x22, 0x24, 0x2d, 0x27, 0x28, 0x2e, 0x21, 0x2b,           \
            0x87, 0x82, 0x8c, 0x8f, 0x84, 0x8b, 0x8a, 0x8c,           \
            0x8b, 0x87, 0x86, 0x89, 0x8d, 0x84, 0x80, 0x8a,           \
            0x82, 0x88, 0x85, 0x83, 0x8f, 0x86, 0x89, 0x85,           \
            0x88, 0x81, 0x83, 0x8e, 0x81, 0x8d, 0x8e, 0x80,           \
            0x80, 0x8f, 0x85, 0x8a, 0x87, 0x82, 0x89, 0x85,           \
            0x8e, 0x81, 0x83, 0x8c, 0x8b, 0x88, 0x8c, 0x86,           \
            0x8f, 0x83, 0x86, 0x8d, 0x84, 0x89, 0x8a, 0x80,           \
            0x82, 0x84, 0x8d, 0x87, 0x88, 0x8e, 0x81, 0x8b,           \
            0x57, 0x52, 0x5c, 0x5f, 0x54, 0x5b, 0x5a, 0x5c,           \
            0x5b, 0x57, 0x56, 0x59, 0x5d, 0x54, 0x50, 0x5a,           \
            0x52, 0x58, 0x55, 0x53, 0x5f, 0x56, 0x59, 0x55,           \
            0x58, 0x51, 0x53, 0x5e, 0x51, 0x5d, 0x5e, 0x50,           \
            0x50, 0x5f, 0x55, 0x5a, 0x57, 0x52, 0x59, 0x55,           \
            0x5e, 0x51, 0x53, 0x5c, 0x5b, 0x58, 0x5c, 0x56,           \
            0x5f, 0x53, 0x56, 0x5d, 0x54, 0x59, 0x5a, 0x50,           \
            0x52, 0x54, 0x5d, 0x57, 0x58, 0x5e, 0x51, 0x5b,           \
            0xf7, 0xf2, 0xfc, 0xff, 0xf4, 0xfb, 0xfa, 0xfc,           \
            0xfb, 0xf7, 0xf6, 0xf9, 0xfd, 0xf4, 0xf0, 0xfa,           \
            0xf2, 0xf8, 0xf5, 0xf3, 0xff, 0xf6, 0xf9, 0xf5,           \
            0xf8, 0xf1, 0xf3, 0xfe, 0xf1, 0xfd, 0xfe, 0xf0,           \
            0xf0, 0xff, 0xf5, 0xfa, 0xf7, 0xf2, 0xf9, 0xf5,           \
            0xfe, 0xf1, 0xf3, 0xfc, 0xfb, 0xf8, 0xfc, 0xf6,           \
            0xff, 0xf3, 0xf6, 0xfd, 0xf4, 0xf9, 0xfa, 0xf0,           \
            0xf2, 0xf4, 0xfd, 0xf7, 0xf8, 0xfe, 0xf1, 0xfb,           \
            0xc7, 0xc2, 0xcc, 0xcf, 0xc4, 0xcb, 0xca, 0xcc,           \
            0xcb, 0xc7, 0xc6, 0xc9, 0xcd, 0xc4, 0xc0, 0xca,           \
            0xc2, 0xc8, 0xc5, 0xc3, 0xcf, 0xc6, 0xc9, 0xc5,           \
            0xc8, 0xc1, 0xc3, 0xce, 0xc1, 0xcd, 0xce, 0xc0,           \
            0xc0, 0xcf, 0xc5, 0xca, 0xc7, 0xc2, 0xc9, 0xc5,           \
            0xce, 0xc1, 0xc3, 0xcc, 0xcb, 0xc8, 0xcc, 0xc6,           \
            0xcf, 0xc3, 0xc6, 0xcd, 0xc4, 0xc9, 0xca, 0xc0,           \
            0xc2, 0xc4, 0xcd, 0xc7, 0xc8, 0xce, 0xc1, 0xcb,           \
            0xb7, 0xb2, 0xbc, 0xbf, 0xb4, 0xbb, 0xba, 0xbc,           \
            0xbb, 0xb7, 0xb6, 0xb9, 0xbd, 0xb4, 0xb0, 0xba,           \
            0xb2, 0xb8, 0xb5, 0xb3, 0xbf, 0xb6, 0xb9, 0xb5,           \
            0xb8, 0xb1, 0xb3, 0xbe, 0xb1, 0xbd, 0xbe, 0xb0,           \
            0xb0, 0xbf, 0xb5, 0xba, 0xb7, 0xb2, 0xb9, 0xb5,           \
            0xbe, 0xb1, 0xb3, 0xbc, 0xbb, 0xb8, 0xbc, 0xb6,           \
            0xbf, 0xb3, 0xb6, 0xbd, 0xb4, 0xb9, 0xba, 0xb0,           \
            0xb2, 0xb4, 0xbd, 0xb7, 0xb8, 0xbe, 0xb1, 0xbb,           \
            0x37, 0x32, 0x3c, 0x3f, 0x34, 0x3b, 0x3a, 0x3c,           \
            0x3b, 0x37, 0x36, 0x39, 0x3d, 0x34, 0x30, 0x3a,           \
            0x32, 0x38, 0x35, 0x33, 0x3f, 0x36, 0x39, 0x35,           \
            0x38, 0x31, 0x33, 0x3e, 0x31, 0x3d, 0x3e, 0x30,           \
            0x30, 0x3f, 0x35, 0x3a, 0x37, 0x32, 0x39, 0x35,           \
            0x3e, 0x31, 0x33, 0x3c, 0x3b, 0x38, 0x3c, 0x36,           \
            0x3f, 0x33, 0x36, 0x3d, 0x34, 0x39, 0x3a, 0x30,           \
            0x32, 0x34, 0x3d, 0x37, 0x38, 0x3e, 0x31, 0x3b,           \
            0xd7, 0xd2, 0xdc, 0xdf, 0xd4, 0xdb, 0xda, 0xdc,           \
            0xdb, 0xd7, 0xd6, 0xd9, 0xdd, 0xd4, 0xd0, 0xda,           \
            0xd2, 0xd8, 0xd5, 0xd3, 0xdf, 0xd6, 0xd9, 0xd5,           \
            0xd8, 0xd1, 0xd3, 0xde, 0xd1, 0xdd, 0xde, 0xd0,           \
            0xd0, 0xdf, 0xd5, 0xda, 0xd7, 0xd2, 0xd9, 0xd5,           \
            0xde, 0xd1, 0xd3, 0xdc, 0xdb, 0xd8, 0xdc, 0xd6,           \
            0xdf, 0xd3, 0xd6, 0xdd, 0xd4, 0xd9, 0xda, 0xd0,           \
            0xd2, 0xd4, 0xdd, 0xd7, 0xd8, 0xde, 0xd1, 0xdb,           \
            0xf7, 0xf2, 0xfc, 0xff, 0xf4, 0xfb, 0xfa, 0xfc,           \
            0xfb, 0xf7, 0xf6, 0xf9, 0xfd, 0xf4, 0xf0, 0xfa,           \
            0xf2, 0xf8, 0xf5, 0xf3, 0xff, 0xf6, 0xf9, 0xf5,           \
            0xf8, 0xf1, 0xf3, 0xfe, 0xf1, 0xfd, 0xfe, 0xf0,           \
            0xf0, 0xff, 0xf5, 0xfa, 0xf7, 0xf2, 0xf9, 0xf5,           \
            0xfe, 0xf1, 0xf3, 0xfc, 0xfb, 0xf8, 0xfc, 0xf6,           \
            0xff, 0xf3, 0xf6, 0xfd, 0xf4, 0xf9, 0xfa, 0xf0,           \
            0xf2, 0xf4, 0xfd, 0xf7, 0xf8, 0xfe, 0xf1, 0xfb,           \
            0xc7, 0xc2, 0xcc, 0xcf, 0xc4, 0xcb, 0xca, 0xcc,           \
            0xcb, 0xc7, 0xc6, 0xc9, 0xcd, 0xc4, 0xc0, 0xca,           \
            0xc2, 0xc8, 0xc5, 0xc3, 0xcf, 0xc6, 0xc9, 0xc5,           \
            0xc8, 0xc1, 0xc3, 0xce, 0xc1, 0xcd, 0xce, 0xc0,           \
            0xc0, 0xcf, 0xc5, 0xca, 0xc7, 0xc2, 0xc9, 0xc5,           \
            0xce, 0xc1, 0xc3, 0xcc, 0xcb, 0xc8, 0xcc, 0xc6,           \
            0xcf, 0xc3, 0xc6, 0xcd, 0xc4, 0xc9, 0xca, 0xc0,           \
            0xc2, 0xc4, 0xcd, 0xc7, 0xc8, 0xce, 0xc1, 0xcb,           \
            0x67, 0x62, 0x6c, 0x6f, 0x64, 0x6b, 0x6a, 0x6c,           \
            0x6b, 0x67, 0x66, 0x69, 0x6d, 0x64, 0x60, 0x6a,           \
            0x62, 0x68, 0x65, 0x63, 0x6f, 0x66, 0x69, 0x65,           \
            0x68, 0x61, 0x63, 0x6e, 0x61, 0x6d, 0x6e, 0x60,           \
            0x60, 0x6f, 0x65, 0x6a, 0x67, 0x62, 0x69, 0x65,           \
            0x6e, 0x61, 0x63, 0x6c, 0x6b, 0x68, 0x6c, 0x66,           \
            0x6f, 0x63, 0x66, 0x6d, 0x64, 0x69, 0x6a, 0x60,           \
            0x62, 0x64, 0x6d, 0x67, 0x68, 0x6e, 0x61, 0x6b,           \
            0x07, 0x02, 0x0c, 0x0f, 0x04, 0x0b, 0x0a, 0x0c,           \
            0x0b, 0x07, 0x06, 0x09, 0x0d, 0x04, 0x00, 0x0a,           \
            0x02, 0x08, 0x05, 0x03, 0x0f, 0x06, 0x09, 0x05,           \
            0x08, 0x01, 0x03, 0x0e, 0x01, 0x0d, 0x0e, 0x00,           \
            0x00, 0x0f, 0x05, 0x0a, 0x07, 0x02, 0x09, 0x05,           \
            0x0e, 0x01, 0x03, 0x0c, 0x0b, 0x08, 0x0c, 0x06,           \
            0x0f, 0x03, 0x06, 0x0d, 0x04, 0x09, 0x0a, 0x00,           \
            0x02, 0x04, 0x0d, 0x07, 0x08, 0x0e, 0x01, 0x0b,           \
            0x27, 0x22, 0x2c, 0x2f, 0x24, 0x2b, 0x2a, 0x2c,           \
            0x2b, 0x27, 0x26, 0x29, 0x2d, 0x24, 0x20, 0x2a,           \
            0x22, 0x28, 0x25, 0x23, 0x2f, 0x26, 0x29, 0x25,           \
            0x28, 0x21, 0x23, 0x2e, 0x21, 0x2d, 0x2e, 0x20,           \
            0x20, 0x2f, 0x25, 0x2a, 0x27, 0x22, 0x29, 0x25,           \
            0x2e, 0x21, 0x23, 0x2c, 0x2b, 0x28, 0x2c, 0x26,           \
            0x2f, 0x23, 0x26, 0x2d, 0x24, 0x29, 0x2a, 0x20,           \
            0x22, 0x24, 0x2d, 0x27, 0x28, 0x2e, 0x21, 0x2b,           \
            0x57, 0x52, 0x5c, 0x5f, 0x54, 0x5b, 0x5a, 0x5c,           \
            0x5b, 0x57, 0x56, 0x59, 0x5d, 0x54, 0x50, 0x5a,           \
            0x52, 0x58, 0x55, 0x53, 0x5f, 0x56, 0x59, 0x55,           \
            0x58, 0x51, 0x53, 0x5e, 0x51, 0x5d, 0x5e, 0x50,           \
            0x50, 0x5f, 0x55, 0x5a, 0x57, 0x52, 0x59, 0x55,           \
            0x5e, 0x51, 0x53, 0x5c, 0x5b, 0x58, 0x5c, 0x56,           \
            0x5f, 0x53, 0x56, 0x5d, 0x54, 0x59, 0x5a, 0x50,           \
            0x52, 0x54, 0x5d, 0x57, 0x58, 0x5e, 0x51, 0x5b,           \
            0x87, 0x82, 0x8c, 0x8f, 0x84, 0x8b, 0x8a, 0x8c,           \
            0x8b, 0x87, 0x86, 0x89, 0x8d, 0x84, 0x80, 0x8a,           \
            0x82, 0x88, 0x85, 0x83, 0x8f, 0x86, 0x89, 0x85,           \
            0x88, 0x81, 0x83, 0x8e, 0x81, 0x8d, 0x8e, 0x80,           \
            0x80, 0x8f, 0x85, 0x8a, 0x87, 0x82, 0x89, 0x85,           \
            0x8e, 0x81, 0x83, 0x8c, 0x8b, 0x88, 0x8c, 0x86,           \
            0x8f, 0x83, 0x86, 0x8d, 0x84, 0x89, 0x8a, 0x80,           \
            0x82, 0x84, 0x8d, 0x87, 0x88, 0x8e, 0x81, 0x8b,           \
            0xe7, 0xe2, 0xec, 0xef, 0xe4, 0xeb, 0xea, 0xec,           \
            0xeb, 0xe7, 0xe6, 0xe9, 0xed, 0xe4, 0xe0, 0xea,           \
            0xe2, 0xe8, 0xe5, 0xe3, 0xef, 0xe6, 0xe9, 0xe5,           \
            0xe8, 0xe1, 0xe3, 0xee, 0xe1, 0xed, 0xee, 0xe0,           \
            0xe0, 0xef, 0xe5, 0xea, 0xe7, 0xe2, 0xe9, 0xe5,           \
            0xee, 0xe1, 0xe3, 0xec, 0xeb, 0xe8, 0xec, 0xe6,           \
            0xef, 0xe3, 0xe6, 0xed, 0xe4, 0xe9, 0xea, 0xe0,           \
            0xe2, 0xe4, 0xed, 0xe7, 0xe8, 0xee, 0xe1, 0xeb,           \
            0x17, 0x12, 0x1c, 0x1f, 0x14, 0x1b, 0x1a, 0x1c,           \
            0x1b, 0x17, 0x16, 0x19, 0x1d, 0x14, 0x10, 0x1a,           \
            0x12, 0x18, 0x15, 0x13, 0x1f, 0x16, 0x19, 0x15,           \
            0x18, 0x11, 0x13, 0x1e, 0x11, 0x1d, 0x1e, 0x10,           \
            0x10, 0x1f, 0x15, 0x1a, 0x17, 0x12, 0x19, 0x15,           \
            0x1e, 0x11, 0x13, 0x1c, 0x1b, 0x18, 0x1c, 0x16,           \
            0x1f, 0x13, 0x16, 0x1d, 0x14, 0x19, 0x1a, 0x10,           \
            0x12, 0x14, 0x1d, 0x17, 0x18, 0x1e, 0x11, 0x1b,           \
            0x27, 0x22, 0x2c, 0x2f, 0x24, 0x2b, 0x2a, 0x2c,           \
            0x2b, 0x27, 0x26, 0x29, 0x2d, 0x24, 0x20, 0x2a,           \
            0x22, 0x28, 0x25, 0x23, 0x2f, 0x26, 0x29, 0x25,           \
            0x28, 0x21, 0x23, 0x2e, 0x21, 0x2d, 0x2e, 0x20,           \
            0x20, 0x2f, 0x25, 0x2a, 0x27, 0x22, 0x29, 0x25,           \
            0x2e, 0x21, 0x23, 0x2c, 0x2b, 0x28, 0x2c, 0x26,           \
            0x2f, 0x23, 0x26, 0x2d, 0x24, 0x29, 0x2a, 0x20,           \
            0x22, 0x24, 0x2d, 0x27, 0x28, 0x2e, 0x21, 0x2b,           \
            0xd7, 0xd2, 0xdc, 0xdf, 0xd4, 0xdb, 0xda, 0xdc,           \
            0xdb, 0xd7, 0xd6, 0xd9, 0xdd, 0xd4, 0xd0, 0xda,           \
            0xd2, 0xd8, 0xd5, 0xd3, 0xdf, 0xd6, 0xd9, 0xd5,           \
            0xd8, 0xd1, 0xd3, 0xde, 0xd1, 0xdd, 0xde, 0xd0,           \
            0xd0, 0xdf, 0xd5, 0xda, 0xd7, 0xd2, 0xd9, 0xd5,           \
            0xde, 0xd1, 0xd3, 0xdc, 0xdb, 0xd8, 0xdc, 0xd6,           \
            0xdf, 0xd3, 0xd6, 0xdd, 0xd4, 0xd9, 0xda, 0xd0,           \
            0xd2, 0xd4, 0xdd, 0xd7, 0xd8, 0xde, 0xd1, 0xdb,           \
            0x77, 0x72, 0x7c, 0x7f, 0x74, 0x7b, 0x7a, 0x7c,           \
            0x7b, 0x77, 0x76, 0x79, 0x7d, 0x74, 0x70, 0x7a,           \
            0x72, 0x78, 0x75, 0x73, 0x7f, 0x76, 0x79, 0x75,           \
            0x78, 0x71, 0x73, 0x7e, 0x71, 0x7d, 0x7e, 0x70,           \
            0x70, 0x7f, 0x75, 0x7a, 0x77, 0x72, 0x79, 0x75,           \
            0x7e, 0x71, 0x73, 0x7c, 0x7b, 0x78, 0x7c, 0x76,           \
            0x7f, 0x73, 0x76, 0x7d, 0x74, 0x79, 0x7a, 0x70,           \
            0x72, 0x74, 0x7d, 0x77, 0x78, 0x7e, 0x71, 0x7b,           \
            0xb7, 0xb2, 0xbc, 0xbf, 0xb4, 0xbb, 0xba, 0xbc,           \
            0xbb, 0xb7, 0xb6, 0xb9, 0xbd, 0xb4, 0xb0, 0xba,           \
            0xb2, 0xb8, 0xb5, 0xb3, 0xbf, 0xb6, 0xb9, 0xb5,           \
            0xb8, 0xb1, 0xb3, 0xbe, 0xb1, 0xbd, 0xbe, 0xb0,           \
            0xb0, 0xbf, 0xb5, 0xba, 0xb7, 0xb2, 0xb9, 0xb5,           \
            0xbe, 0xb1, 0xb3, 0xbc, 0xbb, 0xb8, 0xbc, 0xb6,           \
            0xbf, 0xb3, 0xb6, 0xbd, 0xb4, 0xb9, 0xba, 0xb0,           \
            0xb2, 0xb4, 0xbd, 0xb7, 0xb8, 0xbe, 0xb1, 0xbb,           \
            0x17, 0x12, 0x1c, 0x1f, 0x14, 0x1b, 0x1a, 0x1c,           \
            0x1b, 0x17, 0x16, 0x19, 0x1d, 0x14, 0x10, 0x1a,           \
            0x12, 0x18, 0x15, 0x13, 0x1f, 0x16, 0x19, 0x15,           \
            0x18, 0x11, 0x13, 0x1e, 0x11, 0x1d, 0x1e, 0x10,           \
            0x10, 0x1f, 0x15, 0x1a, 0x17, 0x12, 0x19, 0x15,           \
            0x1e, 0x11, 0x13, 0x1c, 0x1b, 0x18, 0x1c, 0x16,           \
            0x1f, 0x13, 0x16, 0x1d, 0x14, 0x19, 0x1a, 0x10,           \
            0x12, 0x14, 0x1d, 0x17, 0x18, 0x1e, 0x11, 0x1b,           \
            0x07, 0x02, 0x0c, 0x0f, 0x04, 0x0b, 0x0a, 0x0c,           \
            0x0b, 0x07, 0x06, 0x09, 0x0d, 0x04, 0x00, 0x0a,           \
            0x02, 0x08, 0x05, 0x03, 0x0f, 0x06, 0x09, 0x05,           \
            0x08, 0x01, 0x03, 0x0e, 0x01, 0x0d, 0x0e, 0x00,           \
            0x00, 0x0f, 0x05, 0x0a, 0x07, 0x02, 0x09, 0x05,           \
            0x0e, 0x01, 0x03, 0x0c, 0x0b, 0x08, 0x0c, 0x06,           \
            0x0f, 0x03, 0x06, 0x0d, 0x04, 0x09, 0x0a, 0x00,           \
            0x02, 0x04, 0x0d, 0x07, 0x08, 0x0e, 0x01, 0x0b,           \
            0x67, 0x62, 0x6c, 0x6f, 0x64, 0x6b, 0x6a, 0x6c,           \
            0x6b, 0x67, 0x66, 0x69, 0x6d, 0x64, 0x60, 0x6a,           \
            0x62, 0x68, 0x65, 0x63, 0x6f, 0x66, 0x69, 0x65,           \
            0x68, 0x61, 0x63, 0x6e, 0x61, 0x6d, 0x6e, 0x60,           \
            0x60, 0x6f, 0x65, 0x6a, 0x67, 0x62, 0x69, 0x65,           \
            0x6e, 0x61, 0x63, 0x6c, 0x6b, 0x68, 0x6c, 0x66,           \
            0x6f, 0x63, 0x66, 0x6d, 0x64, 0x69, 0x6a, 0x60,           \
            0x62, 0x64, 0x6d, 0x67, 0x68, 0x6e, 0x61, 0x6b,           \
            0x47, 0x42, 0x4c, 0x4f, 0x44, 0x4b, 0x4a, 0x4c,           \
            0x4b, 0x47, 0x46, 0x49, 0x4d, 0x44, 0x40, 0x4a,           \
            0x42, 0x48, 0x45, 0x43, 0x4f, 0x46, 0x49, 0x45,           \
            0x48, 0x41, 0x43, 0x4e, 0x41, 0x4d, 0x4e, 0x40,           \
            0x40, 0x4f, 0x45, 0x4a, 0x47, 0x42, 0x49, 0x45,           \
            0x4e, 0x41, 0x43, 0x4c, 0x4b, 0x48, 0x4c, 0x46,           \
            0x4f, 0x43, 0x46, 0x4d, 0x44, 0x49, 0x4a, 0x40,           \
            0x42, 0x44, 0x4d, 0x47, 0x48, 0x4e, 0x41, 0x4b,           \
            0xf7, 0xf2, 0xfc, 0xff, 0xf4, 0xfb, 0xfa, 0xfc,           \
            0xfb, 0xf7, 0xf6, 0xf9, 0xfd, 0xf4, 0xf0, 0xfa,           \
            0xf2, 0xf8, 0xf5, 0xf3, 0xff, 0xf6, 0xf9, 0xf5,           \
            0xf8, 0xf1, 0xf3, 0xfe, 0xf1, 0xfd, 0xfe, 0xf0,           \
            0xf0, 0xff, 0xf5, 0xfa, 0xf7, 0xf2, 0xf9, 0xf5,           \
            0xfe, 0xf1, 0xf3, 0xfc, 0xfb, 0xf8, 0xfc, 0xf6,           \
            0xff, 0xf3, 0xf6, 0xfd, 0xf4, 0xf9, 0xfa, 0xf0,           \
            0xf2, 0xf4, 0xfd, 0xf7, 0xf8, 0xfe, 0xf1, 0xfb,           \
            0x97, 0x92, 0x9c, 0x9f, 0x94, 0x9b, 0x9a, 0x9c,           \
            0x9b, 0x97, 0x96, 0x99, 0x9d, 0x94, 0x90, 0x9a,           \
            0x92, 0x98, 0x95, 0x93, 0x9f, 0x96, 0x99, 0x95,           \
            0x98, 0x91, 0x93, 0x9e, 0x91, 0x9d, 0x9e, 0x90,           \
            0x90, 0x9f, 0x95, 0x9a, 0x97, 0x92, 0x99, 0x95,           \
            0x9e, 0x91, 0x93, 0x9c, 0x9b, 0x98, 0x9c, 0x96,           \
            0x9f, 0x93, 0x96, 0x9d, 0x94, 0x99, 0x9a, 0x90,           \
            0x92, 0x94, 0x9d, 0x97, 0x98, 0x9e, 0x91, 0x9b,           \
            0xa7, 0xa2, 0xac, 0xaf, 0xa4, 0xab, 0xaa, 0xac,           \
            0xab, 0xa7, 0xa6, 0xa9, 0xad, 0xa4, 0xa0, 0xaa,           \
            0xa2, 0xa8, 0xa5, 0xa3, 0xaf, 0xa6, 0xa9, 0xa5,           \
            0xa8, 0xa1, 0xa3, 0xae, 0xa1, 0xad, 0xae, 0xa0,           \
            0xa0, 0xaf, 0xa5, 0xaa, 0xa7, 0xa2, 0xa9, 0xa5,           \
            0xae, 0xa1, 0xa3, 0xac, 0xab, 0xa8, 0xac, 0xa6,           \
            0xaf, 0xa3, 0xa6, 0xad, 0xa4, 0xa9, 0xaa, 0xa0,           \
            0xa2, 0xa4, 0xad, 0xa7, 0xa8, 0xae, 0xa1, 0xab,           \
            0xe7, 0xe2, 0xec, 0xef, 0xe4, 0xeb, 0xea, 0xec,           \
            0xeb, 0xe7, 0xe6, 0xe9, 0xed, 0xe4, 0xe0, 0xea,           \
            0xe2, 0xe8, 0xe5, 0xe3, 0xef, 0xe6, 0xe9, 0xe5,           \
            0xe8, 0xe1, 0xe3, 0xee, 0xe1, 0xed, 0xee, 0xe0,           \
            0xe0, 0xef, 0xe5, 0xea, 0xe7, 0xe2, 0xe9, 0xe5,           \
            0xee, 0xe1, 0xe3, 0xec, 0xeb, 0xe8, 0xec, 0xe6,           \
            0xef, 0xe3, 0xe6, 0xed, 0xe4, 0xe9, 0xea, 0xe0,           \
            0xe2, 0xe4, 0xed, 0xe7, 0xe8, 0xee, 0xe1, 0xeb,           \
            0x87, 0x82, 0x8c, 0x8f, 0x84, 0x8b, 0x8a, 0x8c,           \
            0x8b, 0x87, 0x86, 0x89, 0x8d, 0x84, 0x80, 0x8a,           \
            0x82, 0x88, 0x85, 0x83, 0x8f, 0x86, 0x89, 0x85,           \
            0x88, 0x81, 0x83, 0x8e, 0x81, 0x8d, 0x8e, 0x80,           \
            0x80, 0x8f, 0x85, 0x8a, 0x87, 0x82, 0x89, 0x85,           \
            0x8e, 0x81, 0x83, 0x8c, 0x8b, 0x88, 0x8c, 0x86,           \
            0x8f, 0x83, 0x86, 0x8d, 0x84, 0x89, 0x8a, 0x80,           \
            0x82, 0x84, 0x8d, 0x87, 0x88, 0x8e, 0x81, 0x8b,           \
            0x57, 0x52, 0x5c, 0x5f, 0x54, 0x5b, 0x5a, 0x5c,           \
            0x5b, 0x57, 0x56, 0x59, 0x5d, 0x54, 0x50, 0x5a,           \
            0x52, 0x58, 0x55, 0x53, 0x5f, 0x56, 0x59, 0x55,           \
            0x58, 0x51, 0x53, 0x5e, 0x51, 0x5d, 0x5e, 0x50,           \
            0x50, 0x5f, 0x55, 0x5a, 0x57, 0x52, 0x59, 0x55,           \
            0x5e, 0x51, 0x53, 0x5c, 0x5b, 0x58, 0x5c, 0x56,           \
            0x5f, 0x53, 0x56, 0x5d, 0x54, 0x59, 0x5a, 0x50,           \
            0x52, 0x54, 0x5d, 0x57, 0x58, 0x5e, 0x51, 0x5b,           \
            0x37, 0x32, 0x3c, 0x3f, 0x34, 0x3b, 0x3a, 0x3c,           \
            0x3b, 0x37, 0x36, 0x39, 0x3d, 0x34, 0x30, 0x3a,           \
            0x32, 0x38, 0x35, 0x33, 0x3f, 0x36, 0x39, 0x35,           \
            0x38, 0x31, 0x33, 0x3e, 0x31, 0x3d, 0x3e, 0x30,           \
            0x30, 0x3f, 0x35, 0x3a, 0x37, 0x32, 0x39, 0x35,           \
            0x3e, 0x31, 0x33, 0x3c, 0x3b, 0x38, 0x3c, 0x36,           \
            0x3f, 0x33, 0x36, 0x3d, 0x34, 0x39, 0x3a, 0x30,           \
            0x32, 0x34, 0x3d, 0x37, 0x38, 0x3e, 0x31, 0x3b,           \
            0x77, 0x72, 0x7c, 0x7f, 0x74, 0x7b, 0x7a, 0x7c,           \
            0x7b, 0x77, 0x76, 0x79, 0x7d, 0x74, 0x70, 0x7a,           \
            0x72, 0x78, 0x75, 0x73, 0x7f, 0x76, 0x79, 0x75,           \
            0x78, 0x71, 0x73, 0x7e, 0x71, 0x7d, 0x7e, 0x70,           \
            0x70, 0x7f, 0x75, 0x7a, 0x77, 0x72, 0x79, 0x75,           \
            0x7e, 0x71, 0x73, 0x7c, 0x7b, 0x78, 0x7c, 0x76,           \
            0x7f, 0x73, 0x76, 0x7d, 0x74, 0x79, 0x7a, 0x70,           \
            0x72, 0x74, 0x7d, 0x77, 0x78, 0x7e, 0x71, 0x7b,           \
            0x47, 0x42, 0x4c, 0x4f, 0x44, 0x4b, 0x4a, 0x4c,           \
            0x4b, 0x47, 0x46, 0x49, 0x4d, 0x44, 0x40, 0x4a,           \
            0x42, 0x48, 0x45, 0x43, 0x4f, 0x46, 0x49, 0x45,           \
            0x48, 0x41, 0x43, 0x4e, 0x41, 0x4d, 0x4e, 0x40,           \
            0x40, 0x4f, 0x45, 0x4a, 0x47, 0x42, 0x49, 0x45,           \
            0x4e, 0x41, 0x43, 0x4c, 0x4b, 0x48, 0x4c, 0x46,           \
            0x4f, 0x43, 0x46, 0x4d, 0x44, 0x49, 0x4a, 0x40,           \
            0x42, 0x44, 0x4d, 0x47, 0x48, 0x4e, 0x41, 0x4b,           \
            0xa7, 0xa2, 0xac, 0xaf, 0xa4, 0xab, 0xaa, 0xac,           \
            0xab, 0xa7, 0xa6, 0xa9, 0xad, 0xa4, 0xa0, 0xaa,           \
            0xa2, 0xa8, 0xa5, 0xa3, 0xaf, 0xa6, 0xa9, 0xa5,           \
            0xa8, 0xa1, 0xa3, 0xae, 0xa1, 0xad, 0xae, 0xa0,           \
            0xa0, 0xaf, 0xa5, 0xaa, 0xa7, 0xa2, 0xa9, 0xa5,           \
            0xae, 0xa1, 0xa3, 0xac, 0xab, 0xa8, 0xac, 0xa6,           \
            0xaf, 0xa3, 0xa6, 0xad, 0xa4, 0xa9, 0xaa, 0xa0,           \
            0xa2, 0xa4, 0xad, 0xa7, 0xa8, 0xae, 0xa1, 0xab,           \
            0x97, 0x92, 0x9c, 0x9f, 0x94, 0x9b, 0x9a, 0x9c,           \
            0x9b, 0x97, 0x96, 0x99, 0x9d, 0x94, 0x90, 0x9a,           \
            0x92, 0x98, 0x95, 0x93, 0x9f, 0x96, 0x99, 0x95,           \
            0x98, 0x91, 0x93, 0x9e, 0x91, 0x9d, 0x9e, 0x90,           \
            0x90, 0x9f, 0x95, 0x9a, 0x97, 0x92, 0x99, 0x95,           \
            0x9e, 0x91, 0x93, 0x9c, 0x9b, 0x98, 0x9c, 0x96,           \
            0x9f, 0x93, 0x96, 0x9d, 0x94, 0x99, 0x9a, 0x90,           \
            0x92, 0x94, 0x9d, 0x97, 0x98, 0x9e, 0x91, 0x9b}

#define SBOX_TBL_2_3 {0xe5, 0xeb, 0xe8, 0xed, 0xe6, 0xe1, 0xed, 0xea, \
            0xe9, 0xe2, 0xe3, 0xe4, 0xef, 0xec, 0xe4, 0xe7,           \
            0xe0, 0xe6, 0xeb, 0xe8, 0xec, 0xef, 0xe2, 0xe5,           \
            0xe7, 0xe9, 0xee, 0xe3, 0xea, 0xe0, 0xe1, 0xee,           \
            0xeb, 0xe8, 0xe4, 0xe2, 0xec, 0xe6, 0xe3, 0xed,           \
            0xe0, 0xeb, 0xea, 0xe7, 0xe6, 0xe1, 0xef, 0xe4,           \
            0xee, 0xe5, 0xe1, 0xef, 0xe2, 0xe9, 0xed, 0xea,           \
            0xe9, 0xe0, 0xe7, 0xec, 0xe5, 0xee, 0xe8, 0xe3,           \
            0x55, 0x5b, 0x58, 0x5d, 0x56, 0x51, 0x5d, 0x5a,           \
            0x59, 0x52, 0x53, 0x54, 0x5f, 0x5c, 0x54, 0x57,           \
            0x50, 0x56, 0x5b, 0x58, 0x5c, 0x5f, 0x52, 0x55,           \
            0x57, 0x59, 0x5e, 0x53, 0x5a, 0x50, 0x51, 0x5e,           \
            0x5b, 0x58, 0x54, 0x52, 0x5c, 0x56, 0x53, 0x5d,           \
            0x50, 0x5b, 0x5a, 0x57, 0x56, 0x51, 0x5f, 0x54,           \
            0x5e, 0x55, 0x51, 0x5f, 0x52, 0x59, 0x5d, 0x5a,           \
            0x59, 0x50, 0x57, 0x5c, 0x55, 0x5e, 0x58, 0x53,           \
            0x85, 0x8b, 0x88, 0x8d, 0x86, 0x81, 0x8d, 0x8a,           \
            0x89, 0x82, 0x83, 0x84, 0x8f, 0x8c, 0x84, 0x87,           \
            0x80, 0x86, 0x8b, 0x88, 0x8c, 0x8f, 0x82, 0x85,           \
            0x87, 0x89, 0x8e, 0x83, 0x8a, 0x80, 0x81, 0x8e,           \
            0x8b, 0x88, 0x84, 0x82, 0x8c, 0x86, 0x83, 0x8d,           \
            0x80, 0x8b, 0x8a, 0x87, 0x86, 0x81, 0x8f, 0x84,           \
            0x8e, 0x85, 0x81, 0x8f, 0x82, 0x89, 0x8d, 0x8a,           \
            0x89, 0x80, 0x87, 0x8c, 0x85, 0x8e, 0x88, 0x83,           \
            0xf5, 0xfb, 0xf8, 0xfd, 0xf6, 0xf1, 0xfd, 0xfa,           \
            0xf9, 0xf2, 0xf3, 0xf4, 0xff, 0xfc, 0xf4, 0xf7,           \
            0xf0, 0xf6, 0xfb, 0xf8, 0xfc, 0xff, 0xf2, 0xf5,           \
            0xf7, 0xf9, 0xfe, 0xf3, 0xfa, 0xf0, 0xf1, 0xfe,           \
            0xfb, 0xf8, 0xf4, 0xf2, 0xfc, 0xf6, 0xf3, 0xfd,           \
            0xf0, 0xfb, 0xfa, 0xf7, 0xf6, 0xf1, 0xff, 0xf4,           \
            0xfe, 0xf5, 0xf1, 0xff, 0xf2, 0xf9, 0xfd, 0xfa,           \
            0xf9, 0xf0, 0xf7, 0xfc, 0xf5, 0xfe, 0xf8, 0xf3,           \
            0x05, 0x0b, 0x08, 0x0d, 0x06, 0x01, 0x0d, 0x0a,           \
            0x09, 0x02, 0x03, 0x04, 0x0f, 0x0c, 0x04, 0x07,           \
            0x00, 0x06, 0x0b, 0x08, 0x0c, 0x0f, 0x02, 0x05,           \
            0x07, 0x09, 0x0e, 0x03, 0x0a, 0x00, 0x01, 0x0e,           \
            0x0b, 0x08, 0x04, 0x02, 0x0c, 0x06, 0x03, 0x0d,           \
            0x00, 0x0b, 0x0a, 0x07, 0x06, 0x01, 0x0f, 0x04,           \
            0x0e, 0x05, 0x01, 0x0f, 0x02, 0x09, 0x0d, 0x0a,           \
            0x09, 0x00, 0x07, 0x0c, 0x05, 0x0e, 0x08, 0x03,           \
            0x35, 0x3b, 0x38, 0x3d, 0x36, 0x31, 0x3d, 0x3a,           \
            0x39, 0x32, 0x33, 0x34, 0x3f, 0x3c, 0x34, 0x37,           \
            0x30, 0x36, 0x3b, 0x38, 0x3c, 0x3f, 0x32, 0x35,           \
            0x37, 0x39, 0x3e, 0x33, 0x3a, 0x30, 0x31, 0x3e,           \
            0x3b, 0x38, 0x34, 0x32, 0x3c, 0x36, 0x33, 0x3d,           \
            0x30, 0x3b, 0x3a, 0x37, 0x36, 0x31, 0x3f, 0x34,           \
            0x3e, 0x35, 0x31, 0x3f, 0x32, 0x39, 0x3d, 0x3a,           \
            0x39, 0x30, 0x37, 0x3c, 0x35, 0x3e, 0x38, 0x33,           \
            0xd5, 0xdb, 0xd8, 0xdd, 0xd6, 0xd1, 0xdd, 0xda,           \
            0xd9, 0xd2, 0xd3, 0xd4, 0xdf, 0xdc, 0xd4, 0xd7,           \
            0xd0, 0xd6, 0xdb, 0xd8, 0xdc, 0xdf, 0xd2, 0xd5,           \
            0xd7, 0xd9, 0xde, 0xd3, 0xda, 0xd0, 0xd1, 0xde,           \
            0xdb, 0xd8, 0xd4, 0xd2, 0xdc, 0xd6, 0xd3, 0xdd,           \
            0xd0, 0xdb, 0xda, 0xd7, 0xd6, 0xd1, 0xdf, 0xd4,           \
            0xde, 0xd5, 0xd1, 0xdf, 0xd2, 0xd9, 0xdd, 0xda,           \
            0xd9, 0xd0, 0xd7, 0xdc, 0xd5, 0xde, 0xd8, 0xd3,           \
            0xa5, 0xab, 0xa8, 0xad, 0xa6, 0xa1, 0xad, 0xaa,           \
            0xa9, 0xa2, 0xa3, 0xa4, 0xaf, 0xac, 0xa4, 0xa7,           \
            0xa0, 0xa6, 0xab, 0xa8, 0xac, 0xaf, 0xa2, 0xa5,           \
            0xa7, 0xa9, 0xae, 0xa3, 0xaa, 0xa0, 0xa1, 0xae,           \
            0xab, 0xa8, 0xa4, 0xa2, 0xac, 0xa6, 0xa3, 0xad,           \
            0xa0, 0xab, 0xaa, 0xa7, 0xa6, 0xa1, 0xaf, 0xa4,           \
            0xae, 0xa5, 0xa1, 0xaf, 0xa2, 0xa9, 0xad, 0xaa,           \
            0xa9, 0xa0, 0xa7, 0xac, 0xa5, 0xae, 0xa8, 0xa3,           \
            0x75, 0x7b, 0x78, 0x7d, 0x76, 0x71, 0x7d, 0x7a,           \
            0x79, 0x72, 0x73, 0x74, 0x7f, 0x7c, 0x74, 0x77,           \
            0x70, 0x76, 0x7b, 0x78, 0x7c, 0x7f, 0x72, 0x75,           \
            0x77, 0x79, 0x7e, 0x73, 0x7a, 0x70, 0x71, 0x7e,           \
            0x7b, 0x78, 0x74, 0x72, 0x7c, 0x76, 0x73, 0x7d,           \
            0x70, 0x7b, 0x7a, 0x77, 0x76, 0x71, 0x7f, 0x74,           \
            0x7e, 0x75, 0x71, 0x7f, 0x72, 0x79, 0x7d, 0x7a,           \
            0x79, 0x70, 0x77, 0x7c, 0x75, 0x7e, 0x78, 0x73,           \
            0x95, 0x9b, 0x98, 0x9d, 0x96, 0x91, 0x9d, 0x9a,           \
            0x99, 0x92, 0x93, 0x94, 0x9f, 0x9c, 0x94, 0x97,           \
            0x90, 0x96, 0x9b, 0x98, 0x9c, 0x9f, 0x92, 0x95,           \
            0x97, 0x99, 0x9e, 0x93, 0x9a, 0x90, 0x91, 0x9e,           \
            0x9b, 0x98, 0x94, 0x92, 0x9c, 0x96, 0x93, 0x9d,           \
            0x90, 0x9b, 0x9a, 0x97, 0x96, 0x91, 0x9f, 0x94,           \
            0x9e, 0x95, 0x91, 0x9f, 0x92, 0x99, 0x9d, 0x9a,           \
            0x99, 0x90, 0x97, 0x9c, 0x95, 0x9e, 0x98, 0x93,           \
            0x15, 0x1b, 0x18, 0x1d, 0x16, 0x11, 0x1d, 0x1a,           \
            0x19, 0x12, 0x13, 0x14, 0x1f, 0x1c, 0x14, 0x17,           \
            0x10, 0x16, 0x1b, 0x18, 0x1c, 0x1f, 0x12, 0x15,           \
            0x17, 0x19, 0x1e, 0x13, 0x1a, 0x10, 0x11, 0x1e,           \
            0x1b, 0x18, 0x14, 0x12, 0x1c, 0x16, 0x13, 0x1d,           \
            0x10, 0x1b, 0x1a, 0x17, 0x16, 0x11, 0x1f, 0x14,           \
            0x1e, 0x15, 0x11, 0x1f, 0x12, 0x19, 0x1d, 0x1a,           \
            0x19, 0x10, 0x17, 0x1c, 0x15, 0x1e, 0x18, 0x13,           \
            0xc5, 0xcb, 0xc8, 0xcd, 0xc6, 0xc1, 0xcd, 0xca,           \
            0xc9, 0xc2, 0xc3, 0xc4, 0xcf, 0xcc, 0xc4, 0xc7,           \
            0xc0, 0xc6, 0xcb, 0xc8, 0xcc, 0xcf, 0xc2, 0xc5,           \
            0xc7, 0xc9, 0xce, 0xc3, 0xca, 0xc0, 0xc1, 0xce,           \
            0xcb, 0xc8, 0xc4, 0xc2, 0xcc, 0xc6, 0xc3, 0xcd,           \
            0xc0, 0xcb, 0xca, 0xc7, 0xc6, 0xc1, 0xcf, 0xc4,           \
            0xce, 0xc5, 0xc1, 0xcf, 0xc2, 0xc9, 0xcd, 0xca,           \
            0xc9, 0xc0, 0xc7, 0xcc, 0xc5, 0xce, 0xc8, 0xc3,           \
            0x95, 0x9b, 0x98, 0x9d, 0x96, 0x91, 0x9d, 0x9a,           \
            0x99, 0x92, 0x93, 0x94, 0x9f, 0x9c, 0x94, 0x97,           \
            0x90, 0x96, 0x9b, 0x98, 0x9c, 0x9f, 0x92, 0x95,           \
            0x97, 0x99, 0x9e, 0x93, 0x9a, 0x90, 0x91, 0x9e,           \
            0x9b, 0x98, 0x94, 0x92, 0x9c, 0x96, 0x93, 0x9d,           \
            0x90, 0x9b, 0x9a, 0x97, 0x96, 0x91, 0x9f, 0x94,           \
            0x9e, 0x95, 0x91, 0x9f, 0x92, 0x99, 0x9d, 0x9a,           \
            0x99, 0x90, 0x97, 0x9c, 0x95, 0x9e, 0x98, 0x93,           \
            0xe5, 0xeb, 0xe8, 0xed, 0xe6, 0xe1, 0xed, 0xea,           \
            0xe9, 0xe2, 0xe3, 0xe4, 0xef, 0xec, 0xe4, 0xe7,           \
            0xe0, 0xe6, 0xeb, 0xe8, 0xec, 0xef, 0xe2, 0xe5,           \
            0xe7, 0xe9, 0xee, 0xe3, 0xea, 0xe0, 0xe1, 0xee,           \
            0xeb, 0xe8, 0xe4, 0xe2, 0xec, 0xe6, 0xe3, 0xed,           \
            0xe0, 0xeb, 0xea, 0xe7, 0xe6, 0xe1, 0xef, 0xe4,           \
            0xee, 0xe5, 0xe1, 0xef, 0xe2, 0xe9, 0xed, 0xea,           \
            0xe9, 0xe0, 0xe7, 0xec, 0xe5, 0xee, 0xe8, 0xe3,           \
            0x25, 0x2b, 0x28, 0x2d, 0x26, 0x21, 0x2d, 0x2a,           \
            0x29, 0x22, 0x23, 0x24, 0x2f, 0x2c, 0x24, 0x27,           \
            0x20, 0x26, 0x2b, 0x28, 0x2c, 0x2f, 0x22, 0x25,           \
            0x27, 0x29, 0x2e, 0x23, 0x2a, 0x20, 0x21, 0x2e,           \
            0x2b, 0x28, 0x24, 0x22, 0x2c, 0x26, 0x23, 0x2d,           \
            0x20, 0x2b, 0x2a, 0x27, 0x26, 0x21, 0x2f, 0x24,           \
            0x2e, 0x25, 0x21, 0x2f, 0x22, 0x29, 0x2d, 0x2a,           \
            0x29, 0x20, 0x27, 0x2c, 0x25, 0x2e, 0x28, 0x23,           \
            0x15, 0x1b, 0x18, 0x1d, 0x16, 0x11, 0x1d, 0x1a,           \
            0x19, 0x12, 0x13, 0x14, 0x1f, 0x1c, 0x14, 0x17,           \
            0x10, 0x16, 0x1b, 0x18, 0x1c, 0x1f, 0x12, 0x15,           \
            0x17, 0x19, 0x1e, 0x13, 0x1a, 0x10, 0x11, 0x1e,           \
            0x1b, 0x18, 0x14, 0x12, 0x1c, 0x16, 0x13, 0x1d,           \
            0x10, 0x1b, 0x1a, 0x17, 0x16, 0x11, 0x1f, 0x14,           \
            0x1e, 0x15, 0x11, 0x1f, 0x12, 0x19, 0x1d, 0x1a,           \
            0x19, 0x10, 0x17, 0x1c, 0x15, 0x1e, 0x18, 0x13,           \
            0xb5, 0xbb, 0xb8, 0xbd, 0xb6, 0xb1, 0xbd, 0xba,           \
            0xb9, 0xb2, 0xb3, 0xb4, 0xbf, 0xbc, 0xb4, 0xb7,           \
            0xb0, 0xb6, 0xbb, 0xb8, 0xbc, 0xbf, 0xb2, 0xb5,           \
            0xb7, 0xb9, 0xbe, 0xb3, 0xba, 0xb0, 0xb1, 0xbe,           \
            0xbb, 0xb8, 0xb4, 0xb2, 0xbc, 0xb6, 0xb3, 0xbd,           \
            0xb0, 0xbb, 0xba, 0xb7, 0xb6, 0xb1, 0xbf, 0xb4,           \
            0xbe, 0xb5, 0xb1, 0xbf, 0xb2, 0xb9, 0xbd, 0xba,           \
            0xb9, 0xb0, 0xb7, 0xbc, 0xb5, 0xbe, 0xb8, 0xb3,           \
            0x65, 0x6b, 0x68, 0x6d, 0x66, 0x61, 0x6d, 0x6a,           \
            0x69, 0x62, 0x63, 0x64, 0x6f, 0x6c, 0x64, 0x67,           \
            0x60, 0x66, 0x6b, 0x68, 0x6c, 0x6f, 0x62, 0x65,           \
            0x67, 0x69, 0x6e, 0x63, 0x6a, 0x60, 0x61, 0x6e,           \
            0x6b, 0x68, 0x64, 0x62, 0x6c, 0x66, 0x63, 0x6d,           \
            0x60, 0x6b, 0x6a, 0x67, 0x66, 0x61, 0x6f, 0x64,           \
            0x6e, 0x65, 0x61, 0x6f, 0x62, 0x69, 0x6d, 0x6a,           \
            0x69, 0x60, 0x67, 0x6c, 0x65, 0x6e, 0x68, 0x63,           \
            0x45, 0x4b, 0x48, 0x4d, 0x46, 0x41, 0x4d, 0x4a,           \
            0x49, 0x42, 0x43, 0x44, 0x4f, 0x4c, 0x44, 0x47,           \
            0x40, 0x46, 0x4b, 0x48, 0x4c, 0x4f, 0x42, 0x45,           \
            0x47, 0x49, 0x4e, 0x43, 0x4a, 0x40, 0x41, 0x4e,           \
            0x4b, 0x48, 0x44, 0x42, 0x4c, 0x46, 0x43, 0x4d,           \
            0x40, 0x4b, 0x4a, 0x47, 0x46, 0x41, 0x4f, 0x44,           \
            0x4e, 0x45, 0x41, 0x4f, 0x42, 0x49, 0x4d, 0x4a,           \
            0x49, 0x40, 0x47, 0x4c, 0x45, 0x4e, 0x48, 0x43,           \
            0x85, 0x8b, 0x88, 0x8d, 0x86, 0x81, 0x8d, 0x8a,           \
            0x89, 0x82, 0x83, 0x84, 0x8f, 0x8c, 0x84, 0x87,           \
            0x80, 0x86, 0x8b, 0x88, 0x8c, 0x8f, 0x82, 0x85,           \
            0x87, 0x89, 0x8e, 0x83, 0x8a, 0x80, 0x81, 0x8e,           \
            0x8b, 0x88, 0x84, 0x82, 0x8c, 0x86, 0x83, 0x8d,           \
            0x80, 0x8b, 0x8a, 0x87, 0x86, 0x81, 0x8f, 0x84,           \
            0x8e, 0x85, 0x81, 0x8f, 0x82, 0x89, 0x8d, 0x8a,           \
            0x89, 0x80, 0x87, 0x8c, 0x85, 0x8e, 0x88, 0x83,           \
            0x65, 0x6b, 0x68, 0x6d, 0x66, 0x61, 0x6d, 0x6a,           \
            0x69, 0x62, 0x63, 0x64, 0x6f, 0x6c, 0x64, 0x67,           \
            0x60, 0x66, 0x6b, 0x68, 0x6c, 0x6f, 0x62, 0x65,           \
            0x67, 0x69, 0x6e, 0x63, 0x6a, 0x60, 0x61, 0x6e,           \
            0x6b, 0x68, 0x64, 0x62, 0x6c, 0x66, 0x63, 0x6d,           \
            0x60, 0x6b, 0x6a, 0x67, 0x66, 0x61, 0x6f, 0x64,           \
            0x6e, 0x65, 0x61, 0x6f, 0x62, 0x69, 0x6d, 0x6a,           \
            0x69, 0x60, 0x67, 0x6c, 0x65, 0x6e, 0x68, 0x63,           \
            0xd5, 0xdb, 0xd8, 0xdd, 0xd6, 0xd1, 0xdd, 0xda,           \
            0xd9, 0xd2, 0xd3, 0xd4, 0xdf, 0xdc, 0xd4, 0xd7,           \
            0xd0, 0xd6, 0xdb, 0xd8, 0xdc, 0xdf, 0xd2, 0xd5,           \
            0xd7, 0xd9, 0xde, 0xd3, 0xda, 0xd0, 0xd1, 0xde,           \
            0xdb, 0xd8, 0xd4, 0xd2, 0xdc, 0xd6, 0xd3, 0xdd,           \
            0xd0, 0xdb, 0xda, 0xd7, 0xd6, 0xd1, 0xdf, 0xd4,           \
            0xde, 0xd5, 0xd1, 0xdf, 0xd2, 0xd9, 0xdd, 0xda,           \
            0xd9, 0xd0, 0xd7, 0xdc, 0xd5, 0xde, 0xd8, 0xd3,           \
            0x35, 0x3b, 0x38, 0x3d, 0x36, 0x31, 0x3d, 0x3a,           \
            0x39, 0x32, 0x33, 0x34, 0x3f, 0x3c, 0x34, 0x37,           \
            0x30, 0x36, 0x3b, 0x38, 0x3c, 0x3f, 0x32, 0x35,           \
            0x37, 0x39, 0x3e, 0x33, 0x3a, 0x30, 0x31, 0x3e,           \
            0x3b, 0x38, 0x34, 0x32, 0x3c, 0x36, 0x33, 0x3d,           \
            0x30, 0x3b, 0x3a, 0x37, 0x36, 0x31, 0x3f, 0x34,           \
            0x3e, 0x35, 0x31, 0x3f, 0x32, 0x39, 0x3d, 0x3a,           \
            0x39, 0x30, 0x37, 0x3c, 0x35, 0x3e, 0x38, 0x33,           \
            0x45, 0x4b, 0x48, 0x4d, 0x46, 0x41, 0x4d, 0x4a,           \
            0x49, 0x42, 0x43, 0x44, 0x4f, 0x4c, 0x44, 0x47,           \
            0x40, 0x46, 0x4b, 0x48, 0x4c, 0x4f, 0x42, 0x45,           \
            0x47, 0x49, 0x4e, 0x43, 0x4a, 0x40, 0x41, 0x4e,           \
            0x4b, 0x48, 0x44, 0x42, 0x4c, 0x46, 0x43, 0x4d,           \
            0x40, 0x4b, 0x4a, 0x47, 0x46, 0x41, 0x4f, 0x44,           \
            0x4e, 0x45, 0x41, 0x4f, 0x42, 0x49, 0x4d, 0x4a,           \
            0x49, 0x40, 0x47, 0x4c, 0x45, 0x4e, 0x48, 0x43,           \
            0xc5, 0xcb, 0xc8, 0xcd, 0xc6, 0xc1, 0xcd, 0xca,           \
            0xc9, 0xc2, 0xc3, 0xc4, 0xcf, 0xcc, 0xc4, 0xc7,           \
            0xc0, 0xc6, 0xcb, 0xc8, 0xcc, 0xcf, 0xc2, 0xc5,           \
            0xc7, 0xc9, 0xce, 0xc3, 0xca, 0xc0, 0xc1, 0xce,           \
            0xcb, 0xc8, 0xc4, 0xc2, 0xcc, 0xc6, 0xc3, 0xcd,           \
            0xc0, 0xcb, 0xca, 0xc7, 0xc6, 0xc1, 0xcf, 0xc4,           \
            0xce, 0xc5, 0xc1, 0xcf, 0xc2, 0xc9, 0xcd, 0xca,           \
            0xc9, 0xc0, 0xc7, 0xcc, 0xc5, 0xce, 0xc8, 0xc3,           \
            0x05, 0x0b, 0x08, 0x0d, 0x06, 0x01, 0x0d, 0x0a,           \
            0x09, 0x02, 0x03, 0x04, 0x0f, 0x0c, 0x04, 0x07,           \
            0x00, 0x06, 0x0b, 0x08, 0x0c, 0x0f, 0x02, 0x05,           \
            0x07, 0x09, 0x0e, 0x03, 0x0a, 0x00, 0x01, 0x0e,           \
            0x0b, 0x08, 0x04, 0x02, 0x0c, 0x06, 0x03, 0x0d,           \
            0x00, 0x0b, 0x0a, 0x07, 0x06, 0x01, 0x0f, 0x04,           \
            0x0e, 0x05, 0x01, 0x0f, 0x02, 0x09, 0x0d, 0x0a,           \
            0x09, 0x00, 0x07, 0x0c, 0x05, 0x0e, 0x08, 0x03,           \
            0xa5, 0xab, 0xa8, 0xad, 0xa6, 0xa1, 0xad, 0xaa,           \
            0xa9, 0xa2, 0xa3, 0xa4, 0xaf, 0xac, 0xa4, 0xa7,           \
            0xa0, 0xa6, 0xab, 0xa8, 0xac, 0xaf, 0xa2, 0xa5,           \
            0xa7, 0xa9, 0xae, 0xa3, 0xaa, 0xa0, 0xa1, 0xae,           \
            0xab, 0xa8, 0xa4, 0xa2, 0xac, 0xa6, 0xa3, 0xad,           \
            0xa0, 0xab, 0xaa, 0xa7, 0xa6, 0xa1, 0xaf, 0xa4,           \
            0xae, 0xa5, 0xa1, 0xaf, 0xa2, 0xa9, 0xad, 0xaa,           \
            0xa9, 0xa0, 0xa7, 0xac, 0xa5, 0xae, 0xa8, 0xa3,           \
            0x75, 0x7b, 0x78, 0x7d, 0x76, 0x71, 0x7d, 0x7a,           \
            0x79, 0x72, 0x73, 0x74, 0x7f, 0x7c, 0x74, 0x77,           \
            0x70, 0x76, 0x7b, 0x78, 0x7c, 0x7f, 0x72, 0x75,           \
            0x77, 0x79, 0x7e, 0x73, 0x7a, 0x70, 0x71, 0x7e,           \
            0x7b, 0x78, 0x74, 0x72, 0x7c, 0x76, 0x73, 0x7d,           \
            0x70, 0x7b, 0x7a, 0x77, 0x76, 0x71, 0x7f, 0x74,           \
            0x7e, 0x75, 0x71, 0x7f, 0x72, 0x79, 0x7d, 0x7a,           \
            0x79, 0x70, 0x77, 0x7c, 0x75, 0x7e, 0x78, 0x73,           \
            0x55, 0x5b, 0x58, 0x5d, 0x56, 0x51, 0x5d, 0x5a,           \
            0x59, 0x52, 0x53, 0x54, 0x5f, 0x5c, 0x54, 0x57,           \
            0x50, 0x56, 0x5b, 0x58, 0x5c, 0x5f, 0x52, 0x55,           \
            0x57, 0x59, 0x5e, 0x53, 0x5a, 0x50, 0x51, 0x5e,           \
            0x5b, 0x58, 0x54, 0x52, 0x5c, 0x56, 0x53, 0x5d,           \
            0x50, 0x5b, 0x5a, 0x57, 0x56, 0x51, 0x5f, 0x54,           \
            0x5e, 0x55, 0x51, 0x5f, 0x52, 0x59, 0x5d, 0x5a,           \
            0x59, 0x50, 0x57, 0x5c, 0x55, 0x5e, 0x58, 0x53,           \
            0xb5, 0xbb, 0xb8, 0xbd, 0xb6, 0xb1, 0xbd, 0xba,           \
            0xb9, 0xb2, 0xb3, 0xb4, 0xbf, 0xbc, 0xb4, 0xb7,           \
            0xb0, 0xb6, 0xbb, 0xb8, 0xbc, 0xbf, 0xb2, 0xb5,           \
            0xb7, 0xb9, 0xbe, 0xb3, 0xba, 0xb0, 0xb1, 0xbe,           \
            0xbb, 0xb8, 0xb4, 0xb2, 0xbc, 0xb6, 0xb3, 0xbd,           \
            0xb0, 0xbb, 0xba, 0xb7, 0xb6, 0xb1, 0xbf, 0xb4,           \
            0xbe, 0xb5, 0xb1, 0xbf, 0xb2, 0xb9, 0xbd, 0xba,           \
            0xb9, 0xb0, 0xb7, 0xbc, 0xb5, 0xbe, 0xb8, 0xb3,           \
            0xf5, 0xfb, 0xf8, 0xfd, 0xf6, 0xf1, 0xfd, 0xfa,           \
            0xf9, 0xf2, 0xf3, 0xf4, 0xff, 0xfc, 0xf4, 0xf7,           \
            0xf0, 0xf6, 0xfb, 0xf8, 0xfc, 0xff, 0xf2, 0xf5,           \
            0xf7, 0xf9, 0xfe, 0xf3, 0xfa, 0xf0, 0xf1, 0xfe,           \
            0xfb, 0xf8, 0xf4, 0xf2, 0xfc, 0xf6, 0xf3, 0xfd,           \
            0xf0, 0xfb, 0xfa, 0xf7, 0xf6, 0xf1, 0xff, 0xf4,           \
            0xfe, 0xf5, 0xf1, 0xff, 0xf2, 0xf9, 0xfd, 0xfa,           \
            0xf9, 0xf0, 0xf7, 0xfc, 0xf5, 0xfe, 0xf8, 0xf3,           \
            0x25, 0x2b, 0x28, 0x2d, 0x26, 0x21, 0x2d, 0x2a,           \
            0x29, 0x22, 0x23, 0x24, 0x2f, 0x2c, 0x24, 0x27,           \
            0x20, 0x26, 0x2b, 0x28, 0x2c, 0x2f, 0x22, 0x25,           \
            0x27, 0x29, 0x2e, 0x23, 0x2a, 0x20, 0x21, 0x2e,           \
            0x2b, 0x28, 0x24, 0x22, 0x2c, 0x26, 0x23, 0x2d,           \
            0x20, 0x2b, 0x2a, 0x27, 0x26, 0x21, 0x2f, 0x24,           \
            0x2e, 0x25, 0x21, 0x2f, 0x22, 0x29, 0x2d, 0x2a,           \
            0x29, 0x20, 0x27, 0x2c, 0x25, 0x2e, 0x28, 0x23,           \
            0xb5, 0xbb, 0xb8, 0xbd, 0xb6, 0xb1, 0xbd, 0xba,           \
            0xb9, 0xb2, 0xb3, 0xb4, 0xbf, 0xbc, 0xb4, 0xb7,           \
            0xb0, 0xb6, 0xbb, 0xb8, 0xbc, 0xbf, 0xb2, 0xb5,           \
            0xb7, 0xb9, 0xbe, 0xb3, 0xba, 0xb0, 0xb1, 0xbe,           \
            0xbb, 0xb8, 0xb4, 0xb2, 0xbc, 0xb6, 0xb3, 0xbd,           \
            0xb0, 0xbb, 0xba, 0xb7, 0xb6, 0xb1, 0xbf, 0xb4,           \
            0xbe, 0xb5, 0xb1, 0xbf, 0xb2, 0xb9, 0xbd, 0xba,           \
            0xb9, 0xb0, 0xb7, 0xbc, 0xb5, 0xbe, 0xb8, 0xb3,           \
            0xc5, 0xcb, 0xc8, 0xcd, 0xc6, 0xc1, 0xcd, 0xca,           \
            0xc9, 0xc2, 0xc3, 0xc4, 0xcf, 0xcc, 0xc4, 0xc7,           \
            0xc0, 0xc6, 0xcb, 0xc8, 0xcc, 0xcf, 0xc2, 0xc5,           \
            0xc7, 0xc9, 0xce, 0xc3, 0xca, 0xc0, 0xc1, 0xce,           \
            0xcb, 0xc8, 0xc4, 0xc2, 0xcc, 0xc6, 0xc3, 0xcd,           \
            0xc0, 0xcb, 0xca, 0xc7, 0xc6, 0xc1, 0xcf, 0xc4,           \
            0xce, 0xc5, 0xc1, 0xcf, 0xc2, 0xc9, 0xcd, 0xca,           \
            0xc9, 0xc0, 0xc7, 0xcc, 0xc5, 0xce, 0xc8, 0xc3,           \
            0x25, 0x2b, 0x28, 0x2d, 0x26, 0x21, 0x2d, 0x2a,           \
            0x29, 0x22, 0x23, 0x24, 0x2f, 0x2c, 0x24, 0x27,           \
            0x20, 0x26, 0x2b, 0x28, 0x2c, 0x2f, 0x22, 0x25,           \
            0x27, 0x29, 0x2e, 0x23, 0x2a, 0x20, 0x21, 0x2e,           \
            0x2b, 0x28, 0x24, 0x22, 0x2c, 0x26, 0x23, 0x2d,           \
            0x20, 0x2b, 0x2a, 0x27, 0x26, 0x21, 0x2f, 0x24,           \
            0x2e, 0x25, 0x21, 0x2f, 0x22, 0x29, 0x2d, 0x2a,           \
            0x29, 0x20, 0x27, 0x2c, 0x25, 0x2e, 0x28, 0x23,           \
            0x95, 0x9b, 0x98, 0x9d, 0x96, 0x91, 0x9d, 0x9a,           \
            0x99, 0x92, 0x93, 0x94, 0x9f, 0x9c, 0x94, 0x97,           \
            0x90, 0x96, 0x9b, 0x98, 0x9c, 0x9f, 0x92, 0x95,           \
            0x97, 0x99, 0x9e, 0x93, 0x9a, 0x90, 0x91, 0x9e,           \
            0x9b, 0x98, 0x94, 0x92, 0x9c, 0x96, 0x93, 0x9d,           \
            0x90, 0x9b, 0x9a, 0x97, 0x96, 0x91, 0x9f, 0x94,           \
            0x9e, 0x95, 0x91, 0x9f, 0x92, 0x99, 0x9d, 0x9a,           \
            0x99, 0x90, 0x97, 0x9c, 0x95, 0x9e, 0x98, 0x93,           \
            0x65, 0x6b, 0x68, 0x6d, 0x66, 0x61, 0x6d, 0x6a,           \
            0x69, 0x62, 0x63, 0x64, 0x6f, 0x6c, 0x64, 0x67,           \
            0x60, 0x66, 0x6b, 0x68, 0x6c, 0x6f, 0x62, 0x65,           \
            0x67, 0x69, 0x6e, 0x63, 0x6a, 0x60, 0x61, 0x6e,           \
            0x6b, 0x68, 0x64, 0x62, 0x6c, 0x66, 0x63, 0x6d,           \
            0x60, 0x6b, 0x6a, 0x67, 0x66, 0x61, 0x6f, 0x64,           \
            0x6e, 0x65, 0x61, 0x6f, 0x62, 0x69, 0x6d, 0x6a,           \
            0x69, 0x60, 0x67, 0x6c, 0x65, 0x6e, 0x68, 0x63,           \
            0x55, 0x5b, 0x58, 0x5d, 0x56, 0x51, 0x5d, 0x5a,           \
            0x59, 0x52, 0x53, 0x54, 0x5f, 0x5c, 0x54, 0x57,           \
            0x50, 0x56, 0x5b, 0x58, 0x5c, 0x5f, 0x52, 0x55,           \
            0x57, 0x59, 0x5e, 0x53, 0x5a, 0x50, 0x51, 0x5e,           \
            0x5b, 0x58, 0x54, 0x52, 0x5c, 0x56, 0x53, 0x5d,           \
            0x50, 0x5b, 0x5a, 0x57, 0x56, 0x51, 0x5f, 0x54,           \
            0x5e, 0x55, 0x51, 0x5f, 0x52, 0x59, 0x5d, 0x5a,           \
            0x59, 0x50, 0x57, 0x5c, 0x55, 0x5e, 0x58, 0x53,           \
            0x85, 0x8b, 0x88, 0x8d, 0x86, 0x81, 0x8d, 0x8a,           \
            0x89, 0x82, 0x83, 0x84, 0x8f, 0x8c, 0x84, 0x87,           \
            0x80, 0x86, 0x8b, 0x88, 0x8c, 0x8f, 0x82, 0x85,           \
            0x87, 0x89, 0x8e, 0x83, 0x8a, 0x80, 0x81, 0x8e,           \
            0x8b, 0x88, 0x84, 0x82, 0x8c, 0x86, 0x83, 0x8d,           \
            0x80, 0x8b, 0x8a, 0x87, 0x86, 0x81, 0x8f, 0x84,           \
            0x8e, 0x85, 0x81, 0x8f, 0x82, 0x89, 0x8d, 0x8a,           \
            0x89, 0x80, 0x87, 0x8c, 0x85, 0x8e, 0x88, 0x83,           \
            0x35, 0x3b, 0x38, 0x3d, 0x36, 0x31, 0x3d, 0x3a,           \
            0x39, 0x32, 0x33, 0x34, 0x3f, 0x3c, 0x34, 0x37,           \
            0x30, 0x36, 0x3b, 0x38, 0x3c, 0x3f, 0x32, 0x35,           \
            0x37, 0x39, 0x3e, 0x33, 0x3a, 0x30, 0x31, 0x3e,           \
            0x3b, 0x38, 0x34, 0x32, 0x3c, 0x36, 0x33, 0x3d,           \
            0x30, 0x3b, 0x3a, 0x37, 0x36, 0x31, 0x3f, 0x34,           \
            0x3e, 0x35, 0x31, 0x3f, 0x32, 0x39, 0x3d, 0x3a,           \
            0x39, 0x30, 0x37, 0x3c, 0x35, 0x3e, 0x38, 0x33,           \
            0xd5, 0xdb, 0xd8, 0xdd, 0xd6, 0xd1, 0xdd, 0xda,           \
            0xd9, 0xd2, 0xd3, 0xd4, 0xdf, 0xdc, 0xd4, 0xd7,           \
            0xd0, 0xd6, 0xdb, 0xd8, 0xdc, 0xdf, 0xd2, 0xd5,           \
            0xd7, 0xd9, 0xde, 0xd3, 0xda, 0xd0, 0xd1, 0xde,           \
            0xdb, 0xd8, 0xd4, 0xd2, 0xdc, 0xd6, 0xd3, 0xdd,           \
            0xd0, 0xdb, 0xda, 0xd7, 0xd6, 0xd1, 0xdf, 0xd4,           \
            0xde, 0xd5, 0xd1, 0xdf, 0xd2, 0xd9, 0xdd, 0xda,           \
            0xd9, 0xd0, 0xd7, 0xdc, 0xd5, 0xde, 0xd8, 0xd3,           \
            0x05, 0x0b, 0x08, 0x0d, 0x06, 0x01, 0x0d, 0x0a,           \
            0x09, 0x02, 0x03, 0x04, 0x0f, 0x0c, 0x04, 0x07,           \
            0x00, 0x06, 0x0b, 0x08, 0x0c, 0x0f, 0x02, 0x05,           \
            0x07, 0x09, 0x0e, 0x03, 0x0a, 0x00, 0x01, 0x0e,           \
            0x0b, 0x08, 0x04, 0x02, 0x0c, 0x06, 0x03, 0x0d,           \
            0x00, 0x0b, 0x0a, 0x07, 0x06, 0x01, 0x0f, 0x04,           \
            0x0e, 0x05, 0x01, 0x0f, 0x02, 0x09, 0x0d, 0x0a,           \
            0x09, 0x00, 0x07, 0x0c, 0x05, 0x0e, 0x08, 0x03,           \
            0x45, 0x4b, 0x48, 0x4d, 0x46, 0x41, 0x4d, 0x4a,           \
            0x49, 0x42, 0x43, 0x44, 0x4f, 0x4c, 0x44, 0x47,           \
            0x40, 0x46, 0x4b, 0x48, 0x4c, 0x4f, 0x42, 0x45,           \
            0x47, 0x49, 0x4e, 0x43, 0x4a, 0x40, 0x41, 0x4e,           \
            0x4b, 0x48, 0x44, 0x42, 0x4c, 0x46, 0x43, 0x4d,           \
            0x40, 0x4b, 0x4a, 0x47, 0x46, 0x41, 0x4f, 0x44,           \
            0x4e, 0x45, 0x41, 0x4f, 0x42, 0x49, 0x4d, 0x4a,           \
            0x49, 0x40, 0x47, 0x4c, 0x45, 0x4e, 0x48, 0x43,           \
            0xa5, 0xab, 0xa8, 0xad, 0xa6, 0xa1, 0xad, 0xaa,           \
            0xa9, 0xa2, 0xa3, 0xa4, 0xaf, 0xac, 0xa4, 0xa7,           \
            0xa0, 0xa6, 0xab, 0xa8, 0xac, 0xaf, 0xa2, 0xa5,           \
            0xa7, 0xa9, 0xae, 0xa3, 0xaa, 0xa0, 0xa1, 0xae,           \
            0xab, 0xa8, 0xa4, 0xa2, 0xac, 0xa6, 0xa3, 0xad,           \
            0xa0, 0xab, 0xaa, 0xa7, 0xa6, 0xa1, 0xaf, 0xa4,           \
            0xae, 0xa5, 0xa1, 0xaf, 0xa2, 0xa9, 0xad, 0xaa,           \
            0xa9, 0xa0, 0xa7, 0xac, 0xa5, 0xae, 0xa8, 0xa3,           \
            0x05, 0x0b, 0x08, 0x0d, 0x06, 0x01, 0x0d, 0x0a,           \
            0x09, 0x02, 0x03, 0x04, 0x0f, 0x0c, 0x04, 0x07,           \
            0x00, 0x06, 0x0b, 0x08, 0x0c, 0x0f, 0x02, 0x05,           \
            0x07, 0x09, 0x0e, 0x03, 0x0a, 0x00, 0x01, 0x0e,           \
            0x0b, 0x08, 0x04, 0x02, 0x0c, 0x06, 0x03, 0x0d,           \
            0x00, 0x0b, 0x0a, 0x07, 0x06, 0x01, 0x0f, 0x04,           \
            0x0e, 0x05, 0x01, 0x0f, 0x02, 0x09, 0x0d, 0x0a,           \
            0x09, 0x00, 0x07, 0x0c, 0x05, 0x0e, 0x08, 0x03,           \
            0xb5, 0xbb, 0xb8, 0xbd, 0xb6, 0xb1, 0xbd, 0xba,           \
            0xb9, 0xb2, 0xb3, 0xb4, 0xbf, 0xbc, 0xb4, 0xb7,           \
            0xb0, 0xb6, 0xbb, 0xb8, 0xbc, 0xbf, 0xb2, 0xb5,           \
            0xb7, 0xb9, 0xbe, 0xb3, 0xba, 0xb0, 0xb1, 0xbe,           \
            0xbb, 0xb8, 0xb4, 0xb2, 0xbc, 0xb6, 0xb3, 0xbd,           \
            0xb0, 0xbb, 0xba, 0xb7, 0xb6, 0xb1, 0xbf, 0xb4,           \
            0xbe, 0xb5, 0xb1, 0xbf, 0xb2, 0xb9, 0xbd, 0xba,           \
            0xb9, 0xb0, 0xb7, 0xbc, 0xb5, 0xbe, 0xb8, 0xb3,           \
            0x75, 0x7b, 0x78, 0x7d, 0x76, 0x71, 0x7d, 0x7a,           \
            0x79, 0x72, 0x73, 0x74, 0x7f, 0x7c, 0x74, 0x77,           \
            0x70, 0x76, 0x7b, 0x78, 0x7c, 0x7f, 0x72, 0x75,           \
            0x77, 0x79, 0x7e, 0x73, 0x7a, 0x70, 0x71, 0x7e,           \
            0x7b, 0x78, 0x74, 0x72, 0x7c, 0x76, 0x73, 0x7d,           \
            0x70, 0x7b, 0x7a, 0x77, 0x76, 0x71, 0x7f, 0x74,           \
            0x7e, 0x75, 0x71, 0x7f, 0x72, 0x79, 0x7d, 0x7a,           \
            0x79, 0x70, 0x77, 0x7c, 0x75, 0x7e, 0x78, 0x73,           \
            0x45, 0x4b, 0x48, 0x4d, 0x46, 0x41, 0x4d, 0x4a,           \
            0x49, 0x42, 0x43, 0x44, 0x4f, 0x4c, 0x44, 0x47,           \
            0x40, 0x46, 0x4b, 0x48, 0x4c, 0x4f, 0x42, 0x45,           \
            0x47, 0x49, 0x4e, 0x43, 0x4a, 0x40, 0x41, 0x4e,           \
            0x4b, 0x48, 0x44, 0x42, 0x4c, 0x46, 0x43, 0x4d,           \
            0x40, 0x4b, 0x4a, 0x47, 0x46, 0x41, 0x4f, 0x44,           \
            0x4e, 0x45, 0x41, 0x4f, 0x42, 0x49, 0x4d, 0x4a,           \
            0x49, 0x40, 0x47, 0x4c, 0x45, 0x4e, 0x48, 0x43,           \
            0x15, 0x1b, 0x18, 0x1d, 0x16, 0x11, 0x1d, 0x1a,           \
            0x19, 0x12, 0x13, 0x14, 0x1f, 0x1c, 0x14, 0x17,           \
            0x10, 0x16, 0x1b, 0x18, 0x1c, 0x1f, 0x12, 0x15,           \
            0x17, 0x19, 0x1e, 0x13, 0x1a, 0x10, 0x11, 0x1e,           \
            0x1b, 0x18, 0x14, 0x12, 0x1c, 0x16, 0x13, 0x1d,           \
            0x10, 0x1b, 0x1a, 0x17, 0x16, 0x11, 0x1f, 0x14,           \
            0x1e, 0x15, 0x11, 0x1f, 0x12, 0x19, 0x1d, 0x1a,           \
            0x19, 0x10, 0x17, 0x1c, 0x15, 0x1e, 0x18, 0x13,           \
            0xf5, 0xfb, 0xf8, 0xfd, 0xf6, 0xf1, 0xfd, 0xfa,           \
            0xf9, 0xf2, 0xf3, 0xf4, 0xff, 0xfc, 0xf4, 0xf7,           \
            0xf0, 0xf6, 0xfb, 0xf8, 0xfc, 0xff, 0xf2, 0xf5,           \
            0xf7, 0xf9, 0xfe, 0xf3, 0xfa, 0xf0, 0xf1, 0xfe,           \
            0xfb, 0xf8, 0xf4, 0xf2, 0xfc, 0xf6, 0xf3, 0xfd,           \
            0xf0, 0xfb, 0xfa, 0xf7, 0xf6, 0xf1, 0xff, 0xf4,           \
            0xfe, 0xf5, 0xf1, 0xff, 0xf2, 0xf9, 0xfd, 0xfa,           \
            0xf9, 0xf0, 0xf7, 0xfc, 0xf5, 0xfe, 0xf8, 0xf3,           \
            0xe5, 0xeb, 0xe8, 0xed, 0xe6, 0xe1, 0xed, 0xea,           \
            0xe9, 0xe2, 0xe3, 0xe4, 0xef, 0xec, 0xe4, 0xe7,           \
            0xe0, 0xe6, 0xeb, 0xe8, 0xec, 0xef, 0xe2, 0xe5,           \
            0xe7, 0xe9, 0xee, 0xe3, 0xea, 0xe0, 0xe1, 0xee,           \
            0xeb, 0xe8, 0xe4, 0xe2, 0xec, 0xe6, 0xe3, 0xed,           \
            0xe0, 0xeb, 0xea, 0xe7, 0xe6, 0xe1, 0xef, 0xe4,           \
            0xee, 0xe5, 0xe1, 0xef, 0xe2, 0xe9, 0xed, 0xea,           \
            0xe9, 0xe0, 0xe7, 0xec, 0xe5, 0xee, 0xe8, 0xe3,           \
            0x25, 0x2b, 0x28, 0x2d, 0x26, 0x21, 0x2d, 0x2a,           \
            0x29, 0x22, 0x23, 0x24, 0x2f, 0x2c, 0x24, 0x27,           \
            0x20, 0x26, 0x2b, 0x28, 0x2c, 0x2f, 0x22, 0x25,           \
            0x27, 0x29, 0x2e, 0x23, 0x2a, 0x20, 0x21, 0x2e,           \
            0x2b, 0x28, 0x24, 0x22, 0x2c, 0x26, 0x23, 0x2d,           \
            0x20, 0x2b, 0x2a, 0x27, 0x26, 0x21, 0x2f, 0x24,           \
            0x2e, 0x25, 0x21, 0x2f, 0x22, 0x29, 0x2d, 0x2a,           \
            0x29, 0x20, 0x27, 0x2c, 0x25, 0x2e, 0x28, 0x23,           \
            0xf5, 0xfb, 0xf8, 0xfd, 0xf6, 0xf1, 0xfd, 0xfa,           \
            0xf9, 0xf2, 0xf3, 0xf4, 0xff, 0xfc, 0xf4, 0xf7,           \
            0xf0, 0xf6, 0xfb, 0xf8, 0xfc, 0xff, 0xf2, 0xf5,           \
            0xf7, 0xf9, 0xfe, 0xf3, 0xfa, 0xf0, 0xf1, 0xfe,           \
            0xfb, 0xf8, 0xf4, 0xf2, 0xfc, 0xf6, 0xf3, 0xfd,           \
            0xf0, 0xfb, 0xfa, 0xf7, 0xf6, 0xf1, 0xff, 0xf4,           \
            0xfe, 0xf5, 0xf1, 0xff, 0xf2, 0xf9, 0xfd, 0xfa,           \
            0xf9, 0xf0, 0xf7, 0xfc, 0xf5, 0xfe, 0xf8, 0xf3,           \
            0x85, 0x8b, 0x88, 0x8d, 0x86, 0x81, 0x8d, 0x8a,           \
            0x89, 0x82, 0x83, 0x84, 0x8f, 0x8c, 0x84, 0x87,           \
            0x80, 0x86, 0x8b, 0x88, 0x8c, 0x8f, 0x82, 0x85,           \
            0x87, 0x89, 0x8e, 0x83, 0x8a, 0x80, 0x81, 0x8e,           \
            0x8b, 0x88, 0x84, 0x82, 0x8c, 0x86, 0x83, 0x8d,           \
            0x80, 0x8b, 0x8a, 0x87, 0x86, 0x81, 0x8f, 0x84,           \
            0x8e, 0x85, 0x81, 0x8f, 0x82, 0x89, 0x8d, 0x8a,           \
            0x89, 0x80, 0x87, 0x8c, 0x85, 0x8e, 0x88, 0x83,           \
            0x55, 0x5b, 0x58, 0x5d, 0x56, 0x51, 0x5d, 0x5a,           \
            0x59, 0x52, 0x53, 0x54, 0x5f, 0x5c, 0x54, 0x57,           \
            0x50, 0x56, 0x5b, 0x58, 0x5c, 0x5f, 0x52, 0x55,           \
            0x57, 0x59, 0x5e, 0x53, 0x5a, 0x50, 0x51, 0x5e,           \
            0x5b, 0x58, 0x54, 0x52, 0x5c, 0x56, 0x53, 0x5d,           \
            0x50, 0x5b, 0x5a, 0x57, 0x56, 0x51, 0x5f, 0x54,           \
            0x5e, 0x55, 0x51, 0x5f, 0x52, 0x59, 0x5d, 0x5a,           \
            0x59, 0x50, 0x57, 0x5c, 0x55, 0x5e, 0x58, 0x53,           \
            0xe5, 0xeb, 0xe8, 0xed, 0xe6, 0xe1, 0xed, 0xea,           \
            0xe9, 0xe2, 0xe3, 0xe4, 0xef, 0xec, 0xe4, 0xe7,           \
            0xe0, 0xe6, 0xeb, 0xe8, 0xec, 0xef, 0xe2, 0xe5,           \
            0xe7, 0xe9, 0xee, 0xe3, 0xea, 0xe0, 0xe1, 0xee,           \
            0xeb, 0xe8, 0xe4, 0xe2, 0xec, 0xe6, 0xe3, 0xed,           \
            0xe0, 0xeb, 0xea, 0xe7, 0xe6, 0xe1, 0xef, 0xe4,           \
            0xee, 0xe5, 0xe1, 0xef, 0xe2, 0xe9, 0xed, 0xea,           \
            0xe9, 0xe0, 0xe7, 0xec, 0xe5, 0xee, 0xe8, 0xe3,           \
            0xa5, 0xab, 0xa8, 0xad, 0xa6, 0xa1, 0xad, 0xaa,           \
            0xa9, 0xa2, 0xa3, 0xa4, 0xaf, 0xac, 0xa4, 0xa7,           \
            0xa0, 0xa6, 0xab, 0xa8, 0xac, 0xaf, 0xa2, 0xa5,           \
            0xa7, 0xa9, 0xae, 0xa3, 0xaa, 0xa0, 0xa1, 0xae,           \
            0xab, 0xa8, 0xa4, 0xa2, 0xac, 0xa6, 0xa3, 0xad,           \
            0xa0, 0xab, 0xaa, 0xa7, 0xa6, 0xa1, 0xaf, 0xa4,           \
            0xae, 0xa5, 0xa1, 0xaf, 0xa2, 0xa9, 0xad, 0xaa,           \
            0xa9, 0xa0, 0xa7, 0xac, 0xa5, 0xae, 0xa8, 0xa3,           \
            0x65, 0x6b, 0x68, 0x6d, 0x66, 0x61, 0x6d, 0x6a,           \
            0x69, 0x62, 0x63, 0x64, 0x6f, 0x6c, 0x64, 0x67,           \
            0x60, 0x66, 0x6b, 0x68, 0x6c, 0x6f, 0x62, 0x65,           \
            0x67, 0x69, 0x6e, 0x63, 0x6a, 0x60, 0x61, 0x6e,           \
            0x6b, 0x68, 0x64, 0x62, 0x6c, 0x66, 0x63, 0x6d,           \
            0x60, 0x6b, 0x6a, 0x67, 0x66, 0x61, 0x6f, 0x64,           \
            0x6e, 0x65, 0x61, 0x6f, 0x62, 0x69, 0x6d, 0x6a,           \
            0x69, 0x60, 0x67, 0x6c, 0x65, 0x6e, 0x68, 0x63,           \
            0x35, 0x3b, 0x38, 0x3d, 0x36, 0x31, 0x3d, 0x3a,           \
            0x39, 0x32, 0x33, 0x34, 0x3f, 0x3c, 0x34, 0x37,           \
            0x30, 0x36, 0x3b, 0x38, 0x3c, 0x3f, 0x32, 0x35,           \
            0x37, 0x39, 0x3e, 0x33, 0x3a, 0x30, 0x31, 0x3e,           \
            0x3b, 0x38, 0x34, 0x32, 0x3c, 0x36, 0x33, 0x3d,           \
            0x30, 0x3b, 0x3a, 0x37, 0x36, 0x31, 0x3f, 0x34,           \
            0x3e, 0x35, 0x31, 0x3f, 0x32, 0x39, 0x3d, 0x3a,           \
            0x39, 0x30, 0x37, 0x3c, 0x35, 0x3e, 0x38, 0x33,           \
            0xd5, 0xdb, 0xd8, 0xdd, 0xd6, 0xd1, 0xdd, 0xda,           \
            0xd9, 0xd2, 0xd3, 0xd4, 0xdf, 0xdc, 0xd4, 0xd7,           \
            0xd0, 0xd6, 0xdb, 0xd8, 0xdc, 0xdf, 0xd2, 0xd5,           \
            0xd7, 0xd9, 0xde, 0xd3, 0xda, 0xd0, 0xd1, 0xde,           \
            0xdb, 0xd8, 0xd4, 0xd2, 0xdc, 0xd6, 0xd3, 0xdd,           \
            0xd0, 0xdb, 0xda, 0xd7, 0xd6, 0xd1, 0xdf, 0xd4,           \
            0xde, 0xd5, 0xd1, 0xdf, 0xd2, 0xd9, 0xdd, 0xda,           \
            0xd9, 0xd0, 0xd7, 0xdc, 0xd5, 0xde, 0xd8, 0xd3,           \
            0xc5, 0xcb, 0xc8, 0xcd, 0xc6, 0xc1, 0xcd, 0xca,           \
            0xc9, 0xc2, 0xc3, 0xc4, 0xcf, 0xcc, 0xc4, 0xc7,           \
            0xc0, 0xc6, 0xcb, 0xc8, 0xcc, 0xcf, 0xc2, 0xc5,           \
            0xc7, 0xc9, 0xce, 0xc3, 0xca, 0xc0, 0xc1, 0xce,           \
            0xcb, 0xc8, 0xc4, 0xc2, 0xcc, 0xc6, 0xc3, 0xcd,           \
            0xc0, 0xcb, 0xca, 0xc7, 0xc6, 0xc1, 0xcf, 0xc4,           \
            0xce, 0xc5, 0xc1, 0xcf, 0xc2, 0xc9, 0xcd, 0xca,           \
            0xc9, 0xc0, 0xc7, 0xcc, 0xc5, 0xce, 0xc8, 0xc3,           \
            0x15, 0x1b, 0x18, 0x1d, 0x16, 0x11, 0x1d, 0x1a,           \
            0x19, 0x12, 0x13, 0x14, 0x1f, 0x1c, 0x14, 0x17,           \
            0x10, 0x16, 0x1b, 0x18, 0x1c, 0x1f, 0x12, 0x15,           \
            0x17, 0x19, 0x1e, 0x13, 0x1a, 0x10, 0x11, 0x1e,           \
            0x1b, 0x18, 0x14, 0x12, 0x1c, 0x16, 0x13, 0x1d,           \
            0x10, 0x1b, 0x1a, 0x17, 0x16, 0x11, 0x1f, 0x14,           \
            0x1e, 0x15, 0x11, 0x1f, 0x12, 0x19, 0x1d, 0x1a,           \
            0x19, 0x10, 0x17, 0x1c, 0x15, 0x1e, 0x18, 0x13,           \
            0x95, 0x9b, 0x98, 0x9d, 0x96, 0x91, 0x9d, 0x9a,           \
            0x99, 0x92, 0x93, 0x94, 0x9f, 0x9c, 0x94, 0x97,           \
            0x90, 0x96, 0x9b, 0x98, 0x9c, 0x9f, 0x92, 0x95,           \
            0x97, 0x99, 0x9e, 0x93, 0x9a, 0x90, 0x91, 0x9e,           \
            0x9b, 0x98, 0x94, 0x92, 0x9c, 0x96, 0x93, 0x9d,           \
            0x90, 0x9b, 0x9a, 0x97, 0x96, 0x91, 0x9f, 0x94,           \
            0x9e, 0x95, 0x91, 0x9f, 0x92, 0x99, 0x9d, 0x9a,           \
            0x99, 0x90, 0x97, 0x9c, 0x95, 0x9e, 0x98, 0x93,           \
            0x75, 0x7b, 0x78, 0x7d, 0x76, 0x71, 0x7d, 0x7a,           \
            0x79, 0x72, 0x73, 0x74, 0x7f, 0x7c, 0x74, 0x77,           \
            0x70, 0x76, 0x7b, 0x78, 0x7c, 0x7f, 0x72, 0x75,           \
            0x77, 0x79, 0x7e, 0x73, 0x7a, 0x70, 0x71, 0x7e,           \
            0x7b, 0x78, 0x74, 0x72, 0x7c, 0x76, 0x73, 0x7d,           \
            0x70, 0x7b, 0x7a, 0x77, 0x76, 0x71, 0x7f, 0x74,           \
            0x7e, 0x75, 0x71, 0x7f, 0x72, 0x79, 0x7d, 0x7a,           \
            0x79, 0x70, 0x77, 0x7c, 0x75, 0x7e, 0x78, 0x73}

#define SBOX_TBL_4_5 {0x34, 0x32, 0x31, 0x3f, 0x3e, 0x35, 0x3b, 0x36, \
            0x32, 0x38, 0x3c, 0x33, 0x3d, 0x3e, 0x37, 0x30,           \
            0x33, 0x34, 0x3a, 0x39, 0x35, 0x3b, 0x30, 0x3c,           \
            0x38, 0x3d, 0x3f, 0x3a, 0x36, 0x31, 0x39, 0x37,           \
            0x37, 0x3d, 0x3a, 0x36, 0x32, 0x38, 0x3c, 0x35,           \
            0x34, 0x33, 0x3f, 0x30, 0x3b, 0x34, 0x31, 0x3a,           \
            0x3d, 0x31, 0x30, 0x3f, 0x3e, 0x37, 0x39, 0x32,           \
            0x33, 0x3e, 0x35, 0x39, 0x38, 0x3b, 0x36, 0x3c,           \
            0x94, 0x92, 0x91, 0x9f, 0x9e, 0x95, 0x9b, 0x96,           \
            0x92, 0x98, 0x9c, 0x93, 0x9d, 0x9e, 0x97, 0x90,           \
            0x93, 0x94, 0x9a, 0x99, 0x95, 0x9b, 0x90, 0x9c,           \
            0x98, 0x9d, 0x9f, 0x9a, 0x96, 0x91, 0x99, 0x97,           \
            0x97, 0x9d, 0x9a, 0x96, 0x92, 0x98, 0x9c, 0x95,           \
            0x94, 0x93, 0x9f, 0x90, 0x9b, 0x94, 0x91, 0x9a,           \
            0x9d, 0x91, 0x90, 0x9f, 0x9e, 0x97, 0x99, 0x92,           \
            0x93, 0x9e, 0x95, 0x99, 0x98, 0x9b, 0x96, 0x9c,           \
            0x04, 0x02, 0x01, 0x0f, 0x0e, 0x05, 0x0b, 0x06,           \
            0x02, 0x08, 0x0c, 0x03, 0x0d, 0x0e, 0x07, 0x00,           \
            0x03, 0x04, 0x0a, 0x09, 0x05, 0x0b, 0x00, 0x0c,           \
            0x08, 0x0d, 0x0f, 0x0a, 0x06, 0x01, 0x09, 0x07,           \
            0x07, 0x0d, 0x0a, 0x06, 0x02, 0x08, 0x0c, 0x05,           \
            0x04, 0x03, 0x0f, 0x00, 0x0b, 0x04, 0x01, 0x0a,           \
            0x0d, 0x01, 0x00, 0x0f, 0x0e, 0x07, 0x09, 0x02,           \
            0x03, 0x0e, 0x05, 0x09, 0x08, 0x0b, 0x06, 0x0c,           \
            0xe4, 0xe2, 0xe1, 0xef, 0xee, 0xe5, 0xeb, 0xe6,           \
            0xe2, 0xe8, 0xec, 0xe3, 0xed, 0xee, 0xe7, 0xe0,           \
            0xe3, 0xe4, 0xea, 0xe9, 0xe5, 0xeb, 0xe0, 0xec,           \
            0xe8, 0xed, 0xef, 0xea, 0xe6, 0xe1, 0xe9, 0xe7,           \
            0xe7, 0xed, 0xea, 0xe6, 0xe2, 0xe8, 0xec, 0xe5,           \
            0xe4, 0xe3, 0xef, 0xe0, 0xeb, 0xe4, 0xe1, 0xea,           \
            0xed, 0xe1, 0xe0, 0xef, 0xee, 0xe7, 0xe9, 0xe2,           \
            0xe3, 0xee, 0xe5, 0xe9, 0xe8, 0xeb, 0xe6, 0xec,           \
            0x94, 0x92, 0x91, 0x9f, 0x9e, 0x95, 0x9b, 0x96,           \
            0x92, 0x98, 0x9c, 0x93, 0x9d, 0x9e, 0x97, 0x90,           \
            0x93, 0x94, 0x9a, 0x99, 0x95, 0x9b, 0x90, 0x9c,           \
            0x98, 0x9d, 0x9f, 0x9a, 0x96, 0x91, 0x99, 0x97,           \
            0x97, 0x9d, 0x9a, 0x96, 0x92, 0x98, 0x9c, 0x95,           \
            0x94, 0x93, 0x9f, 0x90, 0x9b, 0x94, 0x91, 0x9a,           \
            0x9d, 0x91, 0x90, 0x9f, 0x9e, 0x97, 0x99, 0x92,           \
            0x93, 0x9e, 0x95, 0x99, 0x98, 0x9b, 0x96, 0x9c,           \
            0x44, 0x42, 0x41, 0x4f, 0x4e, 0x45, 0x4b, 0x46,           \
            0x42, 0x48, 0x4c, 0x43, 0x4d, 0x4e, 0x47, 0x40,           \
            0x43, 0x44, 0x4a, 0x49, 0x45, 0x4b, 0x40, 0x4c,           \
            0x48, 0x4d, 0x4f, 0x4a, 0x46, 0x41, 0x49, 0x47,           \
            0x47, 0x4d, 0x4a, 0x46, 0x42, 0x48, 0x4c, 0x45,           \
            0x44, 0x43, 0x4f, 0x40, 0x4b, 0x44, 0x41, 0x4a,           \
            0x4d, 0x41, 0x40, 0x4f, 0x4e, 0x47, 0x49, 0x42,           \
            0x43, 0x4e, 0x45, 0x49, 0x48, 0x4b, 0x46, 0x4c,           \
            0x74, 0x72, 0x71, 0x7f, 0x7e, 0x75, 0x7b, 0x76,           \
            0x72, 0x78, 0x7c, 0x73, 0x7d, 0x7e, 0x77, 0x70,           \
            0x73, 0x74, 0x7a, 0x79, 0x75, 0x7b, 0x70, 0x7c,           \
            0x78, 0x7d, 0x7f, 0x7a, 0x76, 0x71, 0x79, 0x77,           \
            0x77, 0x7d, 0x7a, 0x76, 0x72, 0x78, 0x7c, 0x75,           \
            0x74, 0x73, 0x7f, 0x70, 0x7b, 0x74, 0x71, 0x7a,           \
            0x7d, 0x71, 0x70, 0x7f, 0x7e, 0x77, 0x79, 0x72,           \
            0x73, 0x7e, 0x75, 0x79, 0x78, 0x7b, 0x76, 0x7c,           \
            0x84, 0x82, 0x81, 0x8f, 0x8e, 0x85, 0x8b, 0x86,           \
            0x82, 0x88, 0x8c, 0x83, 0x8d, 0x8e, 0x87, 0x80,           \
            0x83, 0x84, 0x8a, 0x89, 0x85, 0x8b, 0x80, 0x8c,           \
            0x88, 0x8d, 0x8f, 0x8a, 0x86, 0x81, 0x89, 0x87,           \
            0x87, 0x8d, 0x8a, 0x86, 0x82, 0x88, 0x8c, 0x85,           \
            0x84, 0x83, 0x8f, 0x80, 0x8b, 0x84, 0x81, 0x8a,           \
            0x8d, 0x81, 0x80, 0x8f, 0x8e, 0x87, 0x89, 0x82,           \
            0x83, 0x8e, 0x85, 0x89, 0x88, 0x8b, 0x86, 0x8c,           \
            0x54, 0x52, 0x51, 0x5f, 0x5e, 0x55, 0x5b, 0x56,           \
            0x52, 0x58, 0x5c, 0x53, 0x5d, 0x5e, 0x57, 0x50,           \
            0x53, 0x54, 0x5a, 0x59, 0x55, 0x5b, 0x50, 0x5c,           \
            0x58, 0x5d, 0x5f, 0x5a, 0x56, 0x51, 0x59, 0x57,           \
            0x57, 0x5d, 0x5a, 0x56, 0x52, 0x58, 0x5c, 0x55,           \
            0x54, 0x53, 0x5f, 0x50, 0x5b, 0x54, 0x51, 0x5a,           \
            0x5d, 0x51, 0x50, 0x5f, 0x5e, 0x57, 0x59, 0x52,           \
            0x53, 0x5e, 0x55, 0x59, 0x58, 0x5b, 0x56, 0x5c,           \
            0xf4, 0xf2, 0xf1, 0xff, 0xfe, 0xf5, 0xfb, 0xf6,           \
            0xf2, 0xf8, 0xfc, 0xf3, 0xfd, 0xfe, 0xf7, 0xf0,           \
            0xf3, 0xf4, 0xfa, 0xf9, 0xf5, 0xfb, 0xf0, 0xfc,           \
            0xf8, 0xfd, 0xff, 0xfa, 0xf6, 0xf1, 0xf9, 0xf7,           \
            0xf7, 0xfd, 0xfa, 0xf6, 0xf2, 0xf8, 0xfc, 0xf5,           \
            0xf4, 0xf3, 0xff, 0xf0, 0xfb, 0xf4, 0xf1, 0xfa,           \
            0xfd, 0xf1, 0xf0, 0xff, 0xfe, 0xf7, 0xf9, 0xf2,           \
            0xf3, 0xfe, 0xf5, 0xf9, 0xf8, 0xfb, 0xf6, 0xfc,           \
            0xc4, 0xc2, 0xc1, 0xcf, 0xce, 0xc5, 0xcb, 0xc6,           \
            0xc2, 0xc8, 0xcc, 0xc3, 0xcd, 0xce, 0xc7, 0xc0,           \
            0xc3, 0xc4, 0xca, 0xc9, 0xc5, 0xcb, 0xc0, 0xcc,           \
            0xc8, 0xcd, 0xcf, 0xca, 0xc6, 0xc1, 0xc9, 0xc7,           \
            0xc7, 0xcd, 0xca, 0xc6, 0xc2, 0xc8, 0xcc, 0xc5,           \
            0xc4, 0xc3, 0xcf, 0xc0, 0xcb, 0xc4, 0xc1, 0xca,           \
            0xcd, 0xc1, 0xc0, 0xcf, 0xce, 0xc7, 0xc9, 0xc2,           \
            0xc3, 0xce, 0xc5, 0xc9, 0xc8, 0xcb, 0xc6, 0xcc,           \
            0x24, 0x22, 0x21, 0x2f, 0x2e, 0x25, 0x2b, 0x26,           \
            0x22, 0x28, 0x2c, 0x23, 0x2d, 0x2e, 0x27, 0x20,           \
            0x23, 0x24, 0x2a, 0x29, 0x25, 0x2b, 0x20, 0x2c,           \
            0x28, 0x2d, 0x2f, 0x2a, 0x26, 0x21, 0x29, 0x27,           \
            0x27, 0x2d, 0x2a, 0x26, 0x22, 0x28, 0x2c, 0x25,           \
            0x24, 0x23, 0x2f, 0x20, 0x2b, 0x24, 0x21, 0x2a,           \
            0x2d, 0x21, 0x20, 0x2f, 0x2e, 0x27, 0x29, 0x22,           \
            0x23, 0x2e, 0x25, 0x29, 0x28, 0x2b, 0x26, 0x2c,           \
            0x64, 0x62, 0x61, 0x6f, 0x6e, 0x65, 0x6b, 0x66,           \
            0x62, 0x68, 0x6c, 0x63, 0x6d, 0x6e, 0x67, 0x60,           \
            0x63, 0x64, 0x6a, 0x69, 0x65, 0x6b, 0x60, 0x6c,           \
            0x68, 0x6d, 0x6f, 0x6a, 0x66, 0x61, 0x69, 0x67,           \
            0x67, 0x6d, 0x6a, 0x66, 0x62, 0x68, 0x6c, 0x65,           \
            0x64, 0x63, 0x6f, 0x60, 0x6b, 0x64, 0x61, 0x6a,           \
            0x6d, 0x61, 0x60, 0x6f, 0x6e, 0x67, 0x69, 0x62,           \
            0x63, 0x6e, 0x65, 0x69, 0x68, 0x6b, 0x66, 0x6c,           \
            0x34, 0x32, 0x31, 0x3f, 0x3e, 0x35, 0x3b, 0x36,           \
            0x32, 0x38, 0x3c, 0x33, 0x3d, 0x3e, 0x37, 0x30,           \
            0x33, 0x34, 0x3a, 0x39, 0x35, 0x3b, 0x30, 0x3c,           \
            0x38, 0x3d, 0x3f, 0x3a, 0x36, 0x31, 0x39, 0x37,           \
            0x37, 0x3d, 0x3a, 0x36, 0x32, 0x38, 0x3c, 0x35,           \
            0x34, 0x33, 0x3f, 0x30, 0x3b, 0x34, 0x31, 0x3a,           \
            0x3d, 0x31, 0x30, 0x3f, 0x3e, 0x37, 0x39, 0x32,           \
            0x33, 0x3e, 0x35, 0x39, 0x38, 0x3b, 0x36, 0x3c,           \
            0xa4, 0xa2, 0xa1, 0xaf, 0xae, 0xa5, 0xab, 0xa6,           \
            0xa2, 0xa8, 0xac, 0xa3, 0xad, 0xae, 0xa7, 0xa0,           \
            0xa3, 0xa4, 0xaa, 0xa9, 0xa5, 0xab, 0xa0, 0xac,           \
            0xa8, 0xad, 0xaf, 0xaa, 0xa6, 0xa1, 0xa9, 0xa7,           \
            0xa7, 0xad, 0xaa, 0xa6, 0xa2, 0xa8, 0xac, 0xa5,           \
            0xa4, 0xa3, 0xaf, 0xa0, 0xab, 0xa4, 0xa1, 0xaa,           \
            0xad, 0xa1, 0xa0, 0xaf, 0xae, 0xa7, 0xa9, 0xa2,           \
            0xa3, 0xae, 0xa5, 0xa9, 0xa8, 0xab, 0xa6, 0xac,           \
            0xd4, 0xd2, 0xd1, 0xdf, 0xde, 0xd5, 0xdb, 0xd6,           \
            0xd2, 0xd8, 0xdc, 0xd3, 0xdd, 0xde, 0xd7, 0xd0,           \
            0xd3, 0xd4, 0xda, 0xd9, 0xd5, 0xdb, 0xd0, 0xdc,           \
            0xd8, 0xdd, 0xdf, 0xda, 0xd6, 0xd1, 0xd9, 0xd7,           \
            0xd7, 0xdd, 0xda, 0xd6, 0xd2, 0xd8, 0xdc, 0xd5,           \
            0xd4, 0xd3, 0xdf, 0xd0, 0xdb, 0xd4, 0xd1, 0xda,           \
            0xdd, 0xd1, 0xd0, 0xdf, 0xde, 0xd7, 0xd9, 0xd2,           \
            0xd3, 0xde, 0xd5, 0xd9, 0xd8, 0xdb, 0xd6, 0xdc,           \
            0x84, 0x82, 0x81, 0x8f, 0x8e, 0x85, 0x8b, 0x86,           \
            0x82, 0x88, 0x8c, 0x83, 0x8d, 0x8e, 0x87, 0x80,           \
            0x83, 0x84, 0x8a, 0x89, 0x85, 0x8b, 0x80, 0x8c,           \
            0x88, 0x8d, 0x8f, 0x8a, 0x86, 0x81, 0x89, 0x87,           \
            0x87, 0x8d, 0x8a, 0x86, 0x82, 0x88, 0x8c, 0x85,           \
            0x84, 0x83, 0x8f, 0x80, 0x8b, 0x84, 0x81, 0x8a,           \
            0x8d, 0x81, 0x80, 0x8f, 0x8e, 0x87, 0x89, 0x82,           \
            0x83, 0x8e, 0x85, 0x89, 0x88, 0x8b, 0x86, 0x8c,           \
            0x74, 0x72, 0x71, 0x7f, 0x7e, 0x75, 0x7b, 0x76,           \
            0x72, 0x78, 0x7c, 0x73, 0x7d, 0x7e, 0x77, 0x70,           \
            0x73, 0x74, 0x7a, 0x79, 0x75, 0x7b, 0x70, 0x7c,           \
            0x78, 0x7d, 0x7f, 0x7a, 0x76, 0x71, 0x79, 0x77,           \
            0x77, 0x7d, 0x7a, 0x76, 0x72, 0x78, 0x7c, 0x75,           \
            0x74, 0x73, 0x7f, 0x70, 0x7b, 0x74, 0x71, 0x7a,           \
            0x7d, 0x71, 0x70, 0x7f, 0x7e, 0x77, 0x79, 0x72,           \
            0x73, 0x7e, 0x75, 0x79, 0x78, 0x7b, 0x76, 0x7c,           \
            0xb4, 0xb2, 0xb1, 0xbf, 0xbe, 0xb5, 0xbb, 0xb6,           \
            0xb2, 0xb8, 0xbc, 0xb3, 0xbd, 0xbe, 0xb7, 0xb0,           \
            0xb3, 0xb4, 0xba, 0xb9, 0xb5, 0xbb, 0xb0, 0xbc,           \
            0xb8, 0xbd, 0xbf, 0xba, 0xb6, 0xb1, 0xb9, 0xb7,           \
            0xb7, 0xbd, 0xba, 0xb6, 0xb2, 0xb8, 0xbc, 0xb5,           \
            0xb4, 0xb3, 0xbf, 0xb0, 0xbb, 0xb4, 0xb1, 0xba,           \
            0xbd, 0xb1, 0xb0, 0xbf, 0xbe, 0xb7, 0xb9, 0xb2,           \
            0xb3, 0xbe, 0xb5, 0xb9, 0xb8, 0xbb, 0xb6, 0xbc,           \
            0x04, 0x02, 0x01, 0x0f, 0x0e, 0x05, 0x0b, 0x06,           \
            0x02, 0x08, 0x0c, 0x03, 0x0d, 0x0e, 0x07, 0x00,           \
            0x03, 0x04, 0x0a, 0x09, 0x05, 0x0b, 0x00, 0x0c,           \
            0x08, 0x0d, 0x0f, 0x0a, 0x06, 0x01, 0x09, 0x07,           \
            0x07, 0x0d, 0x0a, 0x06, 0x02, 0x08, 0x0c, 0x05,           \
            0x04, 0x03, 0x0f, 0x00, 0x0b, 0x04, 0x01, 0x0a,           \
            0x0d, 0x01, 0x00, 0x0f, 0x0e, 0x07, 0x09, 0x02,           \
            0x03, 0x0e, 0x05, 0x09, 0x08, 0x0b, 0x06, 0x0c,           \
            0x44, 0x42, 0x41, 0x4f, 0x4e, 0x45, 0x4b, 0x46,           \
            0x42, 0x48, 0x4c, 0x43, 0x4d, 0x4e, 0x47, 0x40,           \
            0x43, 0x44, 0x4a, 0x49, 0x45, 0x4b, 0x40, 0x4c,           \
            0x48, 0x4d, 0x4f, 0x4a, 0x46, 0x41, 0x49, 0x47,           \
            0x47, 0x4d, 0x4a, 0x46, 0x42, 0x48, 0x4c, 0x45,           \
            0x44, 0x43, 0x4f, 0x40, 0x4b, 0x44, 0x41, 0x4a,           \
            0x4d, 0x41, 0x40, 0x4f, 0x4e, 0x47, 0x49, 0x42,           \
            0x43, 0x4e, 0x45, 0x49, 0x48, 0x4b, 0x46, 0x4c,           \
            0x14, 0x12, 0x11, 0x1f, 0x1e, 0x15, 0x1b, 0x16,           \
            0x12, 0x18, 0x1c, 0x13, 0x1d, 0x1e, 0x17, 0x10,           \
            0x13, 0x14, 0x1a, 0x19, 0x15, 0x1b, 0x10, 0x1c,           \
            0x18, 0x1d, 0x1f, 0x1a, 0x16, 0x11, 0x19, 0x17,           \
            0x17, 0x1d, 0x1a, 0x16, 0x12, 0x18, 0x1c, 0x15,           \
            0x14, 0x13, 0x1f, 0x10, 0x1b, 0x14, 0x11, 0x1a,           \
            0x1d, 0x11, 0x10, 0x1f, 0x1e, 0x17, 0x19, 0x12,           \
            0x13, 0x1e, 0x15, 0x19, 0x18, 0x1b, 0x16, 0x1c,           \
            0xe4, 0xe2, 0xe1, 0xef, 0xee, 0xe5, 0xeb, 0xe6,           \
            0xe2, 0xe8, 0xec, 0xe3, 0xed, 0xee, 0xe7, 0xe0,           \
            0xe3, 0xe4, 0xea, 0xe9, 0xe5, 0xeb, 0xe0, 0xec,           \
            0xe8, 0xed, 0xef, 0xea, 0xe6, 0xe1, 0xe9, 0xe7,           \
            0xe7, 0xed, 0xea, 0xe6, 0xe2, 0xe8, 0xec, 0xe5,           \
            0xe4, 0xe3, 0xef, 0xe0, 0xeb, 0xe4, 0xe1, 0xea,           \
            0xed, 0xe1, 0xe0, 0xef, 0xee, 0xe7, 0xe9, 0xe2,           \
            0xe3, 0xee, 0xe5, 0xe9, 0xe8, 0xeb, 0xe6, 0xec,           \
            0xb4, 0xb2, 0xb1, 0xbf, 0xbe, 0xb5, 0xbb, 0xb6,           \
            0xb2, 0xb8, 0xbc, 0xb3, 0xbd, 0xbe, 0xb7, 0xb0,           \
            0xb3, 0xb4, 0xba, 0xb9, 0xb5, 0xbb, 0xb0, 0xbc,           \
            0xb8, 0xbd, 0xbf, 0xba, 0xb6, 0xb1, 0xb9, 0xb7,           \
            0xb7, 0xbd, 0xba, 0xb6, 0xb2, 0xb8, 0xbc, 0xb5,           \
            0xb4, 0xb3, 0xbf, 0xb0, 0xbb, 0xb4, 0xb1, 0xba,           \
            0xbd, 0xb1, 0xb0, 0xbf, 0xbe, 0xb7, 0xb9, 0xb2,           \
            0xb3, 0xbe, 0xb5, 0xb9, 0xb8, 0xbb, 0xb6, 0xbc,           \
            0xf4, 0xf2, 0xf1, 0xff, 0xfe, 0xf5, 0xfb, 0xf6,           \
            0xf2, 0xf8, 0xfc, 0xf3, 0xfd, 0xfe, 0xf7, 0xf0,           \
            0xf3, 0xf4, 0xfa, 0xf9, 0xf5, 0xfb, 0xf0, 0xfc,           \
            0xf8, 0xfd, 0xff, 0xfa, 0xf6, 0xf1, 0xf9, 0xf7,           \
            0xf7, 0xfd, 0xfa, 0xf6, 0xf2, 0xf8, 0xfc, 0xf5,           \
            0xf4, 0xf3, 0xff, 0xf0, 0xfb, 0xf4, 0xf1, 0xfa,           \
            0xfd, 0xf1, 0xf0, 0xff, 0xfe, 0xf7, 0xf9, 0xf2,           \
            0xf3, 0xfe, 0xf5, 0xf9, 0xf8, 0xfb, 0xf6, 0xfc,           \
            0xa4, 0xa2, 0xa1, 0xaf, 0xae, 0xa5, 0xab, 0xa6,           \
            0xa2, 0xa8, 0xac, 0xa3, 0xad, 0xae, 0xa7, 0xa0,           \
            0xa3, 0xa4, 0xaa, 0xa9, 0xa5, 0xab, 0xa0, 0xac,           \
            0xa8, 0xad, 0xaf, 0xaa, 0xa6, 0xa1, 0xa9, 0xa7,           \
            0xa7, 0xad, 0xaa, 0xa6, 0xa2, 0xa8, 0xac, 0xa5,           \
            0xa4, 0xa3, 0xaf, 0xa0, 0xab, 0xa4, 0xa1, 0xaa,           \
            0xad, 0xa1, 0xa0, 0xaf, 0xae, 0xa7, 0xa9, 0xa2,           \
            0xa3, 0xae, 0xa5, 0xa9, 0xa8, 0xab, 0xa6, 0xac,           \
            0x24, 0x22, 0x21, 0x2f, 0x2e, 0x25, 0x2b, 0x26,           \
            0x22, 0x28, 0x2c, 0x23, 0x2d, 0x2e, 0x27, 0x20,           \
            0x23, 0x24, 0x2a, 0x29, 0x25, 0x2b, 0x20, 0x2c,           \
            0x28, 0x2d, 0x2f, 0x2a, 0x26, 0x21, 0x29, 0x27,           \
            0x27, 0x2d, 0x2a, 0x26, 0x22, 0x28, 0x2c, 0x25,           \
            0x24, 0x23, 0x2f, 0x20, 0x2b, 0x24, 0x21, 0x2a,           \
            0x2d, 0x21, 0x20, 0x2f, 0x2e, 0x27, 0x29, 0x22,           \
            0x23, 0x2e, 0x25, 0x29, 0x28, 0x2b, 0x26, 0x2c,           \
            0x54, 0x52, 0x51, 0x5f, 0x5e, 0x55, 0x5b, 0x56,           \
            0x52, 0x58, 0x5c, 0x53, 0x5d, 0x5e, 0x57, 0x50,           \
            0x53, 0x54, 0x5a, 0x59, 0x55, 0x5b, 0x50, 0x5c,           \
            0x58, 0x5d, 0x5f, 0x5a, 0x56, 0x51, 0x59, 0x57,           \
            0x57, 0x5d, 0x5a, 0x56, 0x52, 0x58, 0x5c, 0x55,           \
            0x54, 0x53, 0x5f, 0x50, 0x5b, 0x54, 0x51, 0x5a,           \
            0x5d, 0x51, 0x50, 0x5f, 0x5e, 0x57, 0x59, 0x52,           \
            0x53, 0x5e, 0x55, 0x59, 0x58, 0x5b, 0x56, 0x5c,           \
            0x14, 0x12, 0x11, 0x1f, 0x1e, 0x15, 0x1b, 0x16,           \
            0x12, 0x18, 0x1c, 0x13, 0x1d, 0x1e, 0x17, 0x10,           \
            0x13, 0x14, 0x1a, 0x19, 0x15, 0x1b, 0x10, 0x1c,           \
            0x18, 0x1d, 0x1f, 0x1a, 0x16, 0x11, 0x19, 0x17,           \
            0x17, 0x1d, 0x1a, 0x16, 0x12, 0x18, 0x1c, 0x15,           \
            0x14, 0x13, 0x1f, 0x10, 0x1b, 0x14, 0x11, 0x1a,           \
            0x1d, 0x11, 0x10, 0x1f, 0x1e, 0x17, 0x19, 0x12,           \
            0x13, 0x1e, 0x15, 0x19, 0x18, 0x1b, 0x16, 0x1c,           \
            0xc4, 0xc2, 0xc1, 0xcf, 0xce, 0xc5, 0xcb, 0xc6,           \
            0xc2, 0xc8, 0xcc, 0xc3, 0xcd, 0xce, 0xc7, 0xc0,           \
            0xc3, 0xc4, 0xca, 0xc9, 0xc5, 0xcb, 0xc0, 0xcc,           \
            0xc8, 0xcd, 0xcf, 0xca, 0xc6, 0xc1, 0xc9, 0xc7,           \
            0xc7, 0xcd, 0xca, 0xc6, 0xc2, 0xc8, 0xcc, 0xc5,           \
            0xc4, 0xc3, 0xcf, 0xc0, 0xcb, 0xc4, 0xc1, 0xca,           \
            0xcd, 0xc1, 0xc0, 0xcf, 0xce, 0xc7, 0xc9, 0xc2,           \
            0xc3, 0xce, 0xc5, 0xc9, 0xc8, 0xcb, 0xc6, 0xcc,           \
            0xd4, 0xd2, 0xd1, 0xdf, 0xde, 0xd5, 0xdb, 0xd6,           \
            0xd2, 0xd8, 0xdc, 0xd3, 0xdd, 0xde, 0xd7, 0xd0,           \
            0xd3, 0xd4, 0xda, 0xd9, 0xd5, 0xdb, 0xd0, 0xdc,           \
            0xd8, 0xdd, 0xdf, 0xda, 0xd6, 0xd1, 0xd9, 0xd7,           \
            0xd7, 0xdd, 0xda, 0xd6, 0xd2, 0xd8, 0xdc, 0xd5,           \
            0xd4, 0xd3, 0xdf, 0xd0, 0xdb, 0xd4, 0xd1, 0xda,           \
            0xdd, 0xd1, 0xd0, 0xdf, 0xde, 0xd7, 0xd9, 0xd2,           \
            0xd3, 0xde, 0xd5, 0xd9, 0xd8, 0xdb, 0xd6, 0xdc,           \
            0x64, 0x62, 0x61, 0x6f, 0x6e, 0x65, 0x6b, 0x66,           \
            0x62, 0x68, 0x6c, 0x63, 0x6d, 0x6e, 0x67, 0x60,           \
            0x63, 0x64, 0x6a, 0x69, 0x65, 0x6b, 0x60, 0x6c,           \
            0x68, 0x6d, 0x6f, 0x6a, 0x66, 0x61, 0x69, 0x67,           \
            0x67, 0x6d, 0x6a, 0x66, 0x62, 0x68, 0x6c, 0x65,           \
            0x64, 0x63, 0x6f, 0x60, 0x6b, 0x64, 0x61, 0x6a,           \
            0x6d, 0x61, 0x60, 0x6f, 0x6e, 0x67, 0x69, 0x62,           \
            0x63, 0x6e, 0x65, 0x69, 0x68, 0x6b, 0x66, 0x6c,           \
            0x54, 0x52, 0x51, 0x5f, 0x5e, 0x55, 0x5b, 0x56,           \
            0x52, 0x58, 0x5c, 0x53, 0x5d, 0x5e, 0x57, 0x50,           \
            0x53, 0x54, 0x5a, 0x59, 0x55, 0x5b, 0x50, 0x5c,           \
            0x58, 0x5d, 0x5f, 0x5a, 0x56, 0x51, 0x59, 0x57,           \
            0x57, 0x5d, 0x5a, 0x56, 0x52, 0x58, 0x5c, 0x55,           \
            0x54, 0x53, 0x5f, 0x50, 0x5b, 0x54, 0x51, 0x5a,           \
            0x5d, 0x51, 0x50, 0x5f, 0x5e, 0x57, 0x59, 0x52,           \
            0x53, 0x5e, 0x55, 0x59, 0x58, 0x5b, 0x56, 0x5c,           \
            0x24, 0x22, 0x21, 0x2f, 0x2e, 0x25, 0x2b, 0x26,           \
            0x22, 0x28, 0x2c, 0x23, 0x2d, 0x2e, 0x27, 0x20,           \
            0x23, 0x24, 0x2a, 0x29, 0x25, 0x2b, 0x20, 0x2c,           \
            0x28, 0x2d, 0x2f, 0x2a, 0x26, 0x21, 0x29, 0x27,           \
            0x27, 0x2d, 0x2a, 0x26, 0x22, 0x28, 0x2c, 0x25,           \
            0x24, 0x23, 0x2f, 0x20, 0x2b, 0x24, 0x21, 0x2a,           \
            0x2d, 0x21, 0x20, 0x2f, 0x2e, 0x27, 0x29, 0x22,           \
            0x23, 0x2e, 0x25, 0x29, 0x28, 0x2b, 0x26, 0x2c,           \
            0x64, 0x62, 0x61, 0x6f, 0x6e, 0x65, 0x6b, 0x66,           \
            0x62, 0x68, 0x6c, 0x63, 0x6d, 0x6e, 0x67, 0x60,           \
            0x63, 0x64, 0x6a, 0x69, 0x65, 0x6b, 0x60, 0x6c,           \
            0x68, 0x6d, 0x6f, 0x6a, 0x66, 0x61, 0x69, 0x67,           \
            0x67, 0x6d, 0x6a, 0x66, 0x62, 0x68, 0x6c, 0x65,           \
            0x64, 0x63, 0x6f, 0x60, 0x6b, 0x64, 0x61, 0x6a,           \
            0x6d, 0x61, 0x60, 0x6f, 0x6e, 0x67, 0x69, 0x62,           \
            0x63, 0x6e, 0x65, 0x69, 0x68, 0x6b, 0x66, 0x6c,           \
            0xd4, 0xd2, 0xd1, 0xdf, 0xde, 0xd5, 0xdb, 0xd6,           \
            0xd2, 0xd8, 0xdc, 0xd3, 0xdd, 0xde, 0xd7, 0xd0,           \
            0xd3, 0xd4, 0xda, 0xd9, 0xd5, 0xdb, 0xd0, 0xdc,           \
            0xd8, 0xdd, 0xdf, 0xda, 0xd6, 0xd1, 0xd9, 0xd7,           \
            0xd7, 0xdd, 0xda, 0xd6, 0xd2, 0xd8, 0xdc, 0xd5,           \
            0xd4, 0xd3, 0xdf, 0xd0, 0xdb, 0xd4, 0xd1, 0xda,           \
            0xdd, 0xd1, 0xd0, 0xdf, 0xde, 0xd7, 0xd9, 0xd2,           \
            0xd3, 0xde, 0xd5, 0xd9, 0xd8, 0xdb, 0xd6, 0xdc,           \
            0xe4, 0xe2, 0xe1, 0xef, 0xee, 0xe5, 0xeb, 0xe6,           \
            0xe2, 0xe8, 0xec, 0xe3, 0xed, 0xee, 0xe7, 0xe0,           \
            0xe3, 0xe4, 0xea, 0xe9, 0xe5, 0xeb, 0xe0, 0xec,           \
            0xe8, 0xed, 0xef, 0xea, 0xe6, 0xe1, 0xe9, 0xe7,           \
            0xe7, 0xed, 0xea, 0xe6, 0xe2, 0xe8, 0xec, 0xe5,           \
            0xe4, 0xe3, 0xef, 0xe0, 0xeb, 0xe4, 0xe1, 0xea,           \
            0xed, 0xe1, 0xe0, 0xef, 0xee, 0xe7, 0xe9, 0xe2,           \
            0xe3, 0xee, 0xe5, 0xe9, 0xe8, 0xeb, 0xe6, 0xec,           \
            0x94, 0x92, 0x91, 0x9f, 0x9e, 0x95, 0x9b, 0x96,           \
            0x92, 0x98, 0x9c, 0x93, 0x9d, 0x9e, 0x97, 0x90,           \
            0x93, 0x94, 0x9a, 0x99, 0x95, 0x9b, 0x90, 0x9c,           \
            0x98, 0x9d, 0x9f, 0x9a, 0x96, 0x91, 0x99, 0x97,           \
            0x97, 0x9d, 0x9a, 0x96, 0x92, 0x98, 0x9c, 0x95,           \
            0x94, 0x93, 0x9f, 0x90, 0x9b, 0x94, 0x91, 0x9a,           \
            0x9d, 0x91, 0x90, 0x9f, 0x9e, 0x97, 0x99, 0x92,           \
            0x93, 0x9e, 0x95, 0x99, 0x98, 0x9b, 0x96, 0x9c,           \
            0x04, 0x02, 0x01, 0x0f, 0x0e, 0x05, 0x0b, 0x06,           \
            0x02, 0x08, 0x0c, 0x03, 0x0d, 0x0e, 0x07, 0x00,           \
            0x03, 0x04, 0x0a, 0x09, 0x05, 0x0b, 0x00, 0x0c,           \
            0x08, 0x0d, 0x0f, 0x0a, 0x06, 0x01, 0x09, 0x07,           \
            0x07, 0x0d, 0x0a, 0x06, 0x02, 0x08, 0x0c, 0x05,           \
            0x04, 0x03, 0x0f, 0x00, 0x0b, 0x04, 0x01, 0x0a,           \
            0x0d, 0x01, 0x00, 0x0f, 0x0e, 0x07, 0x09, 0x02,           \
            0x03, 0x0e, 0x05, 0x09, 0x08, 0x0b, 0x06, 0x0c,           \
            0x64, 0x62, 0x61, 0x6f, 0x6e, 0x65, 0x6b, 0x66,           \
            0x62, 0x68, 0x6c, 0x63, 0x6d, 0x6e, 0x67, 0x60,           \
            0x63, 0x64, 0x6a, 0x69, 0x65, 0x6b, 0x60, 0x6c,           \
            0x68, 0x6d, 0x6f, 0x6a, 0x66, 0x61, 0x69, 0x67,           \
            0x67, 0x6d, 0x6a, 0x66, 0x62, 0x68, 0x6c, 0x65,           \
            0x64, 0x63, 0x6f, 0x60, 0x6b, 0x64, 0x61, 0x6a,           \
            0x6d, 0x61, 0x60, 0x6f, 0x6e, 0x67, 0x69, 0x62,           \
            0x63, 0x6e, 0x65, 0x69, 0x68, 0x6b, 0x66, 0x6c,           \
            0x24, 0x22, 0x21, 0x2f, 0x2e, 0x25, 0x2b, 0x26,           \
            0x22, 0x28, 0x2c, 0x23, 0x2d, 0x2e, 0x27, 0x20,           \
            0x23, 0x24, 0x2a, 0x29, 0x25, 0x2b, 0x20, 0x2c,           \
            0x28, 0x2d, 0x2f, 0x2a, 0x26, 0x21, 0x29, 0x27,           \
            0x27, 0x2d, 0x2a, 0x26, 0x22, 0x28, 0x2c, 0x25,           \
            0x24, 0x23, 0x2f, 0x20, 0x2b, 0x24, 0x21, 0x2a,           \
            0x2d, 0x21, 0x20, 0x2f, 0x2e, 0x27, 0x29, 0x22,           \
            0x23, 0x2e, 0x25, 0x29, 0x28, 0x2b, 0x26, 0x2c,           \
            0x44, 0x42, 0x41, 0x4f, 0x4e, 0x45, 0x4b, 0x46,           \
            0x42, 0x48, 0x4c, 0x43, 0x4d, 0x4e, 0x47, 0x40,           \
            0x43, 0x44, 0x4a, 0x49, 0x45, 0x4b, 0x40, 0x4c,           \
            0x48, 0x4d, 0x4f, 0x4a, 0x46, 0x41, 0x49, 0x47,           \
            0x47, 0x4d, 0x4a, 0x46, 0x42, 0x48, 0x4c, 0x45,           \
            0x44, 0x43, 0x4f, 0x40, 0x4b, 0x44, 0x41, 0x4a,           \
            0x4d, 0x41, 0x40, 0x4f, 0x4e, 0x47, 0x49, 0x42,           \
            0x43, 0x4e, 0x45, 0x49, 0x48, 0x4b, 0x46, 0x4c,           \
            0xb4, 0xb2, 0xb1, 0xbf, 0xbe, 0xb5, 0xbb, 0xb6,           \
            0xb2, 0xb8, 0xbc, 0xb3, 0xbd, 0xbe, 0xb7, 0xb0,           \
            0xb3, 0xb4, 0xba, 0xb9, 0xb5, 0xbb, 0xb0, 0xbc,           \
            0xb8, 0xbd, 0xbf, 0xba, 0xb6, 0xb1, 0xb9, 0xb7,           \
            0xb7, 0xbd, 0xba, 0xb6, 0xb2, 0xb8, 0xbc, 0xb5,           \
            0xb4, 0xb3, 0xbf, 0xb0, 0xbb, 0xb4, 0xb1, 0xba,           \
            0xbd, 0xb1, 0xb0, 0xbf, 0xbe, 0xb7, 0xb9, 0xb2,           \
            0xb3, 0xbe, 0xb5, 0xb9, 0xb8, 0xbb, 0xb6, 0xbc,           \
            0x84, 0x82, 0x81, 0x8f, 0x8e, 0x85, 0x8b, 0x86,           \
            0x82, 0x88, 0x8c, 0x83, 0x8d, 0x8e, 0x87, 0x80,           \
            0x83, 0x84, 0x8a, 0x89, 0x85, 0x8b, 0x80, 0x8c,           \
            0x88, 0x8d, 0x8f, 0x8a, 0x86, 0x81, 0x89, 0x87,           \
            0x87, 0x8d, 0x8a, 0x86, 0x82, 0x88, 0x8c, 0x85,           \
            0x84, 0x83, 0x8f, 0x80, 0x8b, 0x84, 0x81, 0x8a,           \
            0x8d, 0x81, 0x80, 0x8f, 0x8e, 0x87, 0x89, 0x82,           \
            0x83, 0x8e, 0x85, 0x89, 0x88, 0x8b, 0x86, 0x8c,           \
            0x94, 0x92, 0x91, 0x9f, 0x9e, 0x95, 0x9b, 0x96,           \
            0x92, 0x98, 0x9c, 0x93, 0x9d, 0x9e, 0x97, 0x90,           \
            0x93, 0x94, 0x9a, 0x99, 0x95, 0x9b, 0x90, 0x9c,           \
            0x98, 0x9d, 0x9f, 0x9a, 0x96, 0x91, 0x99, 0x97,           \
            0x97, 0x9d, 0x9a, 0x96, 0x92, 0x98, 0x9c, 0x95,           \
            0x94, 0x93, 0x9f, 0x90, 0x9b, 0x94, 0x91, 0x9a,           \
            0x9d, 0x91, 0x90, 0x9f, 0x9e, 0x97, 0x99, 0x92,           \
            0x93, 0x9e, 0x95, 0x99, 0x98, 0x9b, 0x96, 0x9c,           \
            0xf4, 0xf2, 0xf1, 0xff, 0xfe, 0xf5, 0xfb, 0xf6,           \
            0xf2, 0xf8, 0xfc, 0xf3, 0xfd, 0xfe, 0xf7, 0xf0,           \
            0xf3, 0xf4, 0xfa, 0xf9, 0xf5, 0xfb, 0xf0, 0xfc,           \
            0xf8, 0xfd, 0xff, 0xfa, 0xf6, 0xf1, 0xf9, 0xf7,           \
            0xf7, 0xfd, 0xfa, 0xf6, 0xf2, 0xf8, 0xfc, 0xf5,           \
            0xf4, 0xf3, 0xff, 0xf0, 0xfb, 0xf4, 0xf1, 0xfa,           \
            0xfd, 0xf1, 0xf0, 0xff, 0xfe, 0xf7, 0xf9, 0xf2,           \
            0xf3, 0xfe, 0xf5, 0xf9, 0xf8, 0xfb, 0xf6, 0xfc,           \
            0xc4, 0xc2, 0xc1, 0xcf, 0xce, 0xc5, 0xcb, 0xc6,           \
            0xc2, 0xc8, 0xcc, 0xc3, 0xcd, 0xce, 0xc7, 0xc0,           \
            0xc3, 0xc4, 0xca, 0xc9, 0xc5, 0xcb, 0xc0, 0xcc,           \
            0xc8, 0xcd, 0xcf, 0xca, 0xc6, 0xc1, 0xc9, 0xc7,           \
            0xc7, 0xcd, 0xca, 0xc6, 0xc2, 0xc8, 0xcc, 0xc5,           \
            0xc4, 0xc3, 0xcf, 0xc0, 0xcb, 0xc4, 0xc1, 0xca,           \
            0xcd, 0xc1, 0xc0, 0xcf, 0xce, 0xc7, 0xc9, 0xc2,           \
            0xc3, 0xce, 0xc5, 0xc9, 0xc8, 0xcb, 0xc6, 0xcc,           \
            0x14, 0x12, 0x11, 0x1f, 0x1e, 0x15, 0x1b, 0x16,           \
            0x12, 0x18, 0x1c, 0x13, 0x1d, 0x1e, 0x17, 0x10,           \
            0x13, 0x14, 0x1a, 0x19, 0x15, 0x1b, 0x10, 0x1c,           \
            0x18, 0x1d, 0x1f, 0x1a, 0x16, 0x11, 0x19, 0x17,           \
            0x17, 0x1d, 0x1a, 0x16, 0x12, 0x18, 0x1c, 0x15,           \
            0x14, 0x13, 0x1f, 0x10, 0x1b, 0x14, 0x11, 0x1a,           \
            0x1d, 0x11, 0x10, 0x1f, 0x1e, 0x17, 0x19, 0x12,           \
            0x13, 0x1e, 0x15, 0x19, 0x18, 0x1b, 0x16, 0x1c,           \
            0xf4, 0xf2, 0xf1, 0xff, 0xfe, 0xf5, 0xfb, 0xf6,           \
            0xf2, 0xf8, 0xfc, 0xf3, 0xfd, 0xfe, 0xf7, 0xf0,           \
            0xf3, 0xf4, 0xfa, 0xf9, 0xf5, 0xfb, 0xf0, 0xfc,           \
            0xf8, 0xfd, 0xff, 0xfa, 0xf6, 0xf1, 0xf9, 0xf7,           \
            0xf7, 0xfd, 0xfa, 0xf6, 0xf2, 0xf8, 0xfc, 0xf5,           \
            0xf4, 0xf3, 0xff, 0xf0, 0xfb, 0xf4, 0xf1, 0xfa,           \
            0xfd, 0xf1, 0xf0, 0xff, 0xfe, 0xf7, 0xf9, 0xf2,           \
            0xf3, 0xfe, 0xf5, 0xf9, 0xf8, 0xfb, 0xf6, 0xfc,           \
            0xc4, 0xc2, 0xc1, 0xcf, 0xce, 0xc5, 0xcb, 0xc6,           \
            0xc2, 0xc8, 0xcc, 0xc3, 0xcd, 0xce, 0xc7, 0xc0,           \
            0xc3, 0xc4, 0xca, 0xc9, 0xc5, 0xcb, 0xc0, 0xcc,           \
            0xc8, 0xcd, 0xcf, 0xca, 0xc6, 0xc1, 0xc9, 0xc7,           \
            0xc7, 0xcd, 0xca, 0xc6, 0xc2, 0xc8, 0xcc, 0xc5,           \
            0xc4, 0xc3, 0xcf, 0xc0, 0xcb, 0xc4, 0xc1, 0xca,           \
            0xcd, 0xc1, 0xc0, 0xcf, 0xce, 0xc7, 0xc9, 0xc2,           \
            0xc3, 0xce, 0xc5, 0xc9, 0xc8, 0xcb, 0xc6, 0xcc,           \
            0x84, 0x82, 0x81, 0x8f, 0x8e, 0x85, 0x8b, 0x86,           \
            0x82, 0x88, 0x8c, 0x83, 0x8d, 0x8e, 0x87, 0x80,           \
            0x83, 0x84, 0x8a, 0x89, 0x85, 0x8b, 0x80, 0x8c,           \
            0x88, 0x8d, 0x8f, 0x8a, 0x86, 0x81, 0x89, 0x87,           \
            0x87, 0x8d, 0x8a, 0x86, 0x82, 0x88, 0x8c, 0x85,           \
            0x84, 0x83, 0x8f, 0x80, 0x8b, 0x84, 0x81, 0x8a,           \
            0x8d, 0x81, 0x80, 0x8f, 0x8e, 0x87, 0x89, 0x82,           \
            0x83, 0x8e, 0x85, 0x89, 0x88, 0x8b, 0x86, 0x8c,           \
            0x74, 0x72, 0x71, 0x7f, 0x7e, 0x75, 0x7b, 0x76,           \
            0x72, 0x78, 0x7c, 0x73, 0x7d, 0x7e, 0x77, 0x70,           \
            0x73, 0x74, 0x7a, 0x79, 0x75, 0x7b, 0x70, 0x7c,           \
            0x78, 0x7d, 0x7f, 0x7a, 0x76, 0x71, 0x79, 0x77,           \
            0x77, 0x7d, 0x7a, 0x76, 0x72, 0x78, 0x7c, 0x75,           \
            0x74, 0x73, 0x7f, 0x70, 0x7b, 0x74, 0x71, 0x7a,           \
            0x7d, 0x71, 0x70, 0x7f, 0x7e, 0x77, 0x79, 0x72,           \
            0x73, 0x7e, 0x75, 0x79, 0x78, 0x7b, 0x76, 0x7c,           \
            0x34, 0x32, 0x31, 0x3f, 0x3e, 0x35, 0x3b, 0x36,           \
            0x32, 0x38, 0x3c, 0x33, 0x3d, 0x3e, 0x37, 0x30,           \
            0x33, 0x34, 0x3a, 0x39, 0x35, 0x3b, 0x30, 0x3c,           \
            0x38, 0x3d, 0x3f, 0x3a, 0x36, 0x31, 0x39, 0x37,           \
            0x37, 0x3d, 0x3a, 0x36, 0x32, 0x38, 0x3c, 0x35,           \
            0x34, 0x33, 0x3f, 0x30, 0x3b, 0x34, 0x31, 0x3a,           \
            0x3d, 0x31, 0x30, 0x3f, 0x3e, 0x37, 0x39, 0x32,           \
            0x33, 0x3e, 0x35, 0x39, 0x38, 0x3b, 0x36, 0x3c,           \
            0xa4, 0xa2, 0xa1, 0xaf, 0xae, 0xa5, 0xab, 0xa6,           \
            0xa2, 0xa8, 0xac, 0xa3, 0xad, 0xae, 0xa7, 0xa0,           \
            0xa3, 0xa4, 0xaa, 0xa9, 0xa5, 0xab, 0xa0, 0xac,           \
            0xa8, 0xad, 0xaf, 0xaa, 0xa6, 0xa1, 0xa9, 0xa7,           \
            0xa7, 0xad, 0xaa, 0xa6, 0xa2, 0xa8, 0xac, 0xa5,           \
            0xa4, 0xa3, 0xaf, 0xa0, 0xab, 0xa4, 0xa1, 0xaa,           \
            0xad, 0xa1, 0xa0, 0xaf, 0xae, 0xa7, 0xa9, 0xa2,           \
            0xa3, 0xae, 0xa5, 0xa9, 0xa8, 0xab, 0xa6, 0xac,           \
            0xd4, 0xd2, 0xd1, 0xdf, 0xde, 0xd5, 0xdb, 0xd6,           \
            0xd2, 0xd8, 0xdc, 0xd3, 0xdd, 0xde, 0xd7, 0xd0,           \
            0xd3, 0xd4, 0xda, 0xd9, 0xd5, 0xdb, 0xd0, 0xdc,           \
            0xd8, 0xdd, 0xdf, 0xda, 0xd6, 0xd1, 0xd9, 0xd7,           \
            0xd7, 0xdd, 0xda, 0xd6, 0xd2, 0xd8, 0xdc, 0xd5,           \
            0xd4, 0xd3, 0xdf, 0xd0, 0xdb, 0xd4, 0xd1, 0xda,           \
            0xdd, 0xd1, 0xd0, 0xdf, 0xde, 0xd7, 0xd9, 0xd2,           \
            0xd3, 0xde, 0xd5, 0xd9, 0xd8, 0xdb, 0xd6, 0xdc,           \
            0x04, 0x02, 0x01, 0x0f, 0x0e, 0x05, 0x0b, 0x06,           \
            0x02, 0x08, 0x0c, 0x03, 0x0d, 0x0e, 0x07, 0x00,           \
            0x03, 0x04, 0x0a, 0x09, 0x05, 0x0b, 0x00, 0x0c,           \
            0x08, 0x0d, 0x0f, 0x0a, 0x06, 0x01, 0x09, 0x07,           \
            0x07, 0x0d, 0x0a, 0x06, 0x02, 0x08, 0x0c, 0x05,           \
            0x04, 0x03, 0x0f, 0x00, 0x0b, 0x04, 0x01, 0x0a,           \
            0x0d, 0x01, 0x00, 0x0f, 0x0e, 0x07, 0x09, 0x02,           \
            0x03, 0x0e, 0x05, 0x09, 0x08, 0x0b, 0x06, 0x0c,           \
            0x44, 0x42, 0x41, 0x4f, 0x4e, 0x45, 0x4b, 0x46,           \
            0x42, 0x48, 0x4c, 0x43, 0x4d, 0x4e, 0x47, 0x40,           \
            0x43, 0x44, 0x4a, 0x49, 0x45, 0x4b, 0x40, 0x4c,           \
            0x48, 0x4d, 0x4f, 0x4a, 0x46, 0x41, 0x49, 0x47,           \
            0x47, 0x4d, 0x4a, 0x46, 0x42, 0x48, 0x4c, 0x45,           \
            0x44, 0x43, 0x4f, 0x40, 0x4b, 0x44, 0x41, 0x4a,           \
            0x4d, 0x41, 0x40, 0x4f, 0x4e, 0x47, 0x49, 0x42,           \
            0x43, 0x4e, 0x45, 0x49, 0x48, 0x4b, 0x46, 0x4c,           \
            0x34, 0x32, 0x31, 0x3f, 0x3e, 0x35, 0x3b, 0x36,           \
            0x32, 0x38, 0x3c, 0x33, 0x3d, 0x3e, 0x37, 0x30,           \
            0x33, 0x34, 0x3a, 0x39, 0x35, 0x3b, 0x30, 0x3c,           \
            0x38, 0x3d, 0x3f, 0x3a, 0x36, 0x31, 0x39, 0x37,           \
            0x37, 0x3d, 0x3a, 0x36, 0x32, 0x38, 0x3c, 0x35,           \
            0x34, 0x33, 0x3f, 0x30, 0x3b, 0x34, 0x31, 0x3a,           \
            0x3d, 0x31, 0x30, 0x3f, 0x3e, 0x37, 0x39, 0x32,           \
            0x33, 0x3e, 0x35, 0x39, 0x38, 0x3b, 0x36, 0x3c,           \
            0x74, 0x72, 0x71, 0x7f, 0x7e, 0x75, 0x7b, 0x76,           \
            0x72, 0x78, 0x7c, 0x73, 0x7d, 0x7e, 0x77, 0x70,           \
            0x73, 0x74, 0x7a, 0x79, 0x75, 0x7b, 0x70, 0x7c,           \
            0x78, 0x7d, 0x7f, 0x7a, 0x76, 0x71, 0x79, 0x77,           \
            0x77, 0x7d, 0x7a, 0x76, 0x72, 0x78, 0x7c, 0x75,           \
            0x74, 0x73, 0x7f, 0x70, 0x7b, 0x74, 0x71, 0x7a,           \
            0x7d, 0x71, 0x70, 0x7f, 0x7e, 0x77, 0x79, 0x72,           \
            0x73, 0x7e, 0x75, 0x79, 0x78, 0x7b, 0x76, 0x7c,           \
            0xe4, 0xe2, 0xe1, 0xef, 0xee, 0xe5, 0xeb, 0xe6,           \
            0xe2, 0xe8, 0xec, 0xe3, 0xed, 0xee, 0xe7, 0xe0,           \
            0xe3, 0xe4, 0xea, 0xe9, 0xe5, 0xeb, 0xe0, 0xec,           \
            0xe8, 0xed, 0xef, 0xea, 0xe6, 0xe1, 0xe9, 0xe7,           \
            0xe7, 0xed, 0xea, 0xe6, 0xe2, 0xe8, 0xec, 0xe5,           \
            0xe4, 0xe3, 0xef, 0xe0, 0xeb, 0xe4, 0xe1, 0xea,           \
            0xed, 0xe1, 0xe0, 0xef, 0xee, 0xe7, 0xe9, 0xe2,           \
            0xe3, 0xee, 0xe5, 0xe9, 0xe8, 0xeb, 0xe6, 0xec,           \
            0xa4, 0xa2, 0xa1, 0xaf, 0xae, 0xa5, 0xab, 0xa6,           \
            0xa2, 0xa8, 0xac, 0xa3, 0xad, 0xae, 0xa7, 0xa0,           \
            0xa3, 0xa4, 0xaa, 0xa9, 0xa5, 0xab, 0xa0, 0xac,           \
            0xa8, 0xad, 0xaf, 0xaa, 0xa6, 0xa1, 0xa9, 0xa7,           \
            0xa7, 0xad, 0xaa, 0xa6, 0xa2, 0xa8, 0xac, 0xa5,           \
            0xa4, 0xa3, 0xaf, 0xa0, 0xab, 0xa4, 0xa1, 0xaa,           \
            0xad, 0xa1, 0xa0, 0xaf, 0xae, 0xa7, 0xa9, 0xa2,           \
            0xa3, 0xae, 0xa5, 0xa9, 0xa8, 0xab, 0xa6, 0xac,           \
            0x54, 0x52, 0x51, 0x5f, 0x5e, 0x55, 0x5b, 0x56,           \
            0x52, 0x58, 0x5c, 0x53, 0x5d, 0x5e, 0x57, 0x50,           \
            0x53, 0x54, 0x5a, 0x59, 0x55, 0x5b, 0x50, 0x5c,           \
            0x58, 0x5d, 0x5f, 0x5a, 0x56, 0x51, 0x59, 0x57,           \
            0x57, 0x5d, 0x5a, 0x56, 0x52, 0x58, 0x5c, 0x55,           \
            0x54, 0x53, 0x5f, 0x50, 0x5b, 0x54, 0x51, 0x5a,           \
            0x5d, 0x51, 0x50, 0x5f, 0x5e, 0x57, 0x59, 0x52,           \
            0x53, 0x5e, 0x55, 0x59, 0x58, 0x5b, 0x56, 0x5c,           \
            0x14, 0x12, 0x11, 0x1f, 0x1e, 0x15, 0x1b, 0x16,           \
            0x12, 0x18, 0x1c, 0x13, 0x1d, 0x1e, 0x17, 0x10,           \
            0x13, 0x14, 0x1a, 0x19, 0x15, 0x1b, 0x10, 0x1c,           \
            0x18, 0x1d, 0x1f, 0x1a, 0x16, 0x11, 0x19, 0x17,           \
            0x17, 0x1d, 0x1a, 0x16, 0x12, 0x18, 0x1c, 0x15,           \
            0x14, 0x13, 0x1f, 0x10, 0x1b, 0x14, 0x11, 0x1a,           \
            0x1d, 0x11, 0x10, 0x1f, 0x1e, 0x17, 0x19, 0x12,           \
            0x13, 0x1e, 0x15, 0x19, 0x18, 0x1b, 0x16, 0x1c,           \
            0xb4, 0xb2, 0xb1, 0xbf, 0xbe, 0xb5, 0xbb, 0xb6,           \
            0xb2, 0xb8, 0xbc, 0xb3, 0xbd, 0xbe, 0xb7, 0xb0,           \
            0xb3, 0xb4, 0xba, 0xb9, 0xb5, 0xbb, 0xb0, 0xbc,           \
            0xb8, 0xbd, 0xbf, 0xba, 0xb6, 0xb1, 0xb9, 0xb7,           \
            0xb7, 0xbd, 0xba, 0xb6, 0xb2, 0xb8, 0xbc, 0xb5,           \
            0xb4, 0xb3, 0xbf, 0xb0, 0xbb, 0xb4, 0xb1, 0xba,           \
            0xbd, 0xb1, 0xb0, 0xbf, 0xbe, 0xb7, 0xb9, 0xb2,           \
            0xb3, 0xbe, 0xb5, 0xb9, 0xb8, 0xbb, 0xb6, 0xbc}

#define SBOX_TBL_6_7 {0xb2, 0xb8, 0xbc, 0xb5, 0xbf, 0xb3, 0xba, 0xb0, \
            0xb4, 0xbd, 0xb9, 0xb6, 0xb1, 0xbe, 0xb6, 0xb9,           \
            0xbd, 0xb2, 0xb3, 0xbf, 0xb0, 0xbc, 0xb5, 0xba,           \
            0xb7, 0xbb, 0xbe, 0xb1, 0xbb, 0xb7, 0xb8, 0xb4,           \
            0xbb, 0xb6, 0xb7, 0xb9, 0xb2, 0xb8, 0xb4, 0xb7,           \
            0xbd, 0xbb, 0xba, 0xb0, 0xb8, 0xb5, 0xb1, 0xbc,           \
            0xb0, 0xbd, 0xbc, 0xba, 0xb9, 0xb2, 0xbf, 0xb4,           \
            0xbe, 0xb1, 0xb3, 0xbf, 0xb5, 0xbe, 0xb6, 0xb3,           \
            0xe2, 0xe8, 0xec, 0xe5, 0xef, 0xe3, 0xea, 0xe0,           \
            0xe4, 0xed, 0xe9, 0xe6, 0xe1, 0xee, 0xe6, 0xe9,           \
            0xed, 0xe2, 0xe3, 0xef, 0xe0, 0xec, 0xe5, 0xea,           \
            0xe7, 0xeb, 0xee, 0xe1, 0xeb, 0xe7, 0xe8, 0xe4,           \
            0xeb, 0xe6, 0xe7, 0xe9, 0xe2, 0xe8, 0xe4, 0xe7,           \
            0xed, 0xeb, 0xea, 0xe0, 0xe8, 0xe5, 0xe1, 0xec,           \
            0xe0, 0xed, 0xec, 0xea, 0xe9, 0xe2, 0xef, 0xe4,           \
            0xee, 0xe1, 0xe3, 0xef, 0xe5, 0xee, 0xe6, 0xe3,           \
            0x52, 0x58, 0x5c, 0x55, 0x5f, 0x53, 0x5a, 0x50,           \
            0x54, 0x5d, 0x59, 0x56, 0x51, 0x5e, 0x56, 0x59,           \
            0x5d, 0x52, 0x53, 0x5f, 0x50, 0x5c, 0x55, 0x5a,           \
            0x57, 0x5b, 0x5e, 0x51, 0x5b, 0x57, 0x58, 0x54,           \
            0x5b, 0x56, 0x57, 0x59, 0x52, 0x58, 0x54, 0x57,           \
            0x5d, 0x5b, 0x5a, 0x50, 0x58, 0x55, 0x51, 0x5c,           \
            0x50, 0x5d, 0x5c, 0x5a, 0x59, 0x52, 0x5f, 0x54,           \
            0x5e, 0x51, 0x53, 0x5f, 0x55, 0x5e, 0x56, 0x53,           \
            0x02, 0x08, 0x0c, 0x05, 0x0f, 0x03, 0x0a, 0x00,           \
            0x04, 0x0d, 0x09, 0x06, 0x01, 0x0e, 0x06, 0x09,           \
            0x0d, 0x02, 0x03, 0x0f, 0x00, 0x0c, 0x05, 0x0a,           \
            0x07, 0x0b, 0x0e, 0x01, 0x0b, 0x07, 0x08, 0x04,           \
            0x0b, 0x06, 0x07, 0x09, 0x02, 0x08, 0x04, 0x07,           \
            0x0d, 0x0b, 0x0a, 0x00, 0x08, 0x05, 0x01, 0x0c,           \
            0x00, 0x0d, 0x0c, 0x0a, 0x09, 0x02, 0x0f, 0x04,           \
            0x0e, 0x01, 0x03, 0x0f, 0x05, 0x0e, 0x06, 0x03,           \
            0x62, 0x68, 0x6c, 0x65, 0x6f, 0x63, 0x6a, 0x60,           \
            0x64, 0x6d, 0x69, 0x66, 0x61, 0x6e, 0x66, 0x69,           \
            0x6d, 0x62, 0x63, 0x6f, 0x60, 0x6c, 0x65, 0x6a,           \
            0x67, 0x6b, 0x6e, 0x61, 0x6b, 0x67, 0x68, 0x64,           \
            0x6b, 0x66, 0x67, 0x69, 0x62, 0x68, 0x64, 0x67,           \
            0x6d, 0x6b, 0x6a, 0x60, 0x68, 0x65, 0x61, 0x6c,           \
            0x60, 0x6d, 0x6c, 0x6a, 0x69, 0x62, 0x6f, 0x64,           \
            0x6e, 0x61, 0x63, 0x6f, 0x65, 0x6e, 0x66, 0x63,           \
            0x92, 0x98, 0x9c, 0x95, 0x9f, 0x93, 0x9a, 0x90,           \
            0x94, 0x9d, 0x99, 0x96, 0x91, 0x9e, 0x96, 0x99,           \
            0x9d, 0x92, 0x93, 0x9f, 0x90, 0x9c, 0x95, 0x9a,           \
            0x97, 0x9b, 0x9e, 0x91, 0x9b, 0x97, 0x98, 0x94,           \
            0x9b, 0x96, 0x97, 0x99, 0x92, 0x98, 0x94, 0x97,           \
            0x9d, 0x9b, 0x9a, 0x90, 0x98, 0x95, 0x91, 0x9c,           \
            0x90, 0x9d, 0x9c, 0x9a, 0x99, 0x92, 0x9f, 0x94,           \
            0x9e, 0x91, 0x93, 0x9f, 0x95, 0x9e, 0x96, 0x93,           \
            0xa2, 0xa8, 0xac, 0xa5, 0xaf, 0xa3, 0xaa, 0xa0,           \
            0xa4, 0xad, 0xa9, 0xa6, 0xa1, 0xae, 0xa6, 0xa9,           \
            0xad, 0xa2, 0xa3, 0xaf, 0xa0, 0xac, 0xa5, 0xaa,           \
            0xa7, 0xab, 0xae, 0xa1, 0xab, 0xa7, 0xa8, 0xa4,           \
            0xab, 0xa6, 0xa7, 0xa9, 0xa2, 0xa8, 0xa4, 0xa7,           \
            0xad, 0xab, 0xaa, 0xa0, 0xa8, 0xa5, 0xa1, 0xac,           \
            0xa0, 0xad, 0xac, 0xaa, 0xa9, 0xa2, 0xaf, 0xa4,           \
            0xae, 0xa1, 0xa3, 0xaf, 0xa5, 0xae, 0xa6, 0xa3,           \
            0xf2, 0xf8, 0xfc, 0xf5, 0xff, 0xf3, 0xfa, 0xf0,           \
            0xf4, 0xfd, 0xf9, 0xf6, 0xf1, 0xfe, 0xf6, 0xf9,           \
            0xfd, 0xf2, 0xf3, 0xff, 0xf0, 0xfc, 0xf5, 0xfa,           \
            0xf7, 0xfb, 0xfe, 0xf1, 0xfb, 0xf7, 0xf8, 0xf4,           \
            0xfb, 0xf6, 0xf7, 0xf9, 0xf2, 0xf8, 0xf4, 0xf7,           \
            0xfd, 0xfb, 0xfa, 0xf0, 0xf8, 0xf5, 0xf1, 0xfc,           \
            0xf0, 0xfd, 0xfc, 0xfa, 0xf9, 0xf2, 0xff, 0xf4,           \
            0xfe, 0xf1, 0xf3, 0xff, 0xf5, 0xfe, 0xf6, 0xf3,           \
            0x12, 0x18, 0x1c, 0x15, 0x1f, 0x13, 0x1a, 0x10,           \
            0x14, 0x1d, 0x19, 0x16, 0x11, 0x1e, 0x16, 0x19,           \
            0x1d, 0x12, 0x13, 0x1f, 0x10, 0x1c, 0x15, 0x1a,           \
            0x17, 0x1b, 0x1e, 0x11, 0x1b, 0x17, 0x18, 0x14,           \
            0x1b, 0x16, 0x17, 0x19, 0x12, 0x18, 0x14, 0x17,           \
            0x1d, 0x1b, 0x1a, 0x10, 0x18, 0x15, 0x11, 0x1c,           \
            0x10, 0x1d, 0x1c, 0x1a, 0x19, 0x12, 0x1f, 0x14,           \
            0x1e, 0x11, 0x13, 0x1f, 0x15, 0x1e, 0x16, 0x13,           \
            0x22, 0x28, 0x2c, 0x25, 0x2f, 0x23, 0x2a, 0x20,           \
            0x24, 0x2d, 0x29, 0x26, 0x21, 0x2e, 0x26, 0x29,           \
            0x2d, 0x22, 0x23, 0x2f, 0x20, 0x2c, 0x25, 0x2a,           \
            0x27, 0x2b, 0x2e, 0x21, 0x2b, 0x27, 0x28, 0x24,           \
            0x2b, 0x26, 0x27, 0x29, 0x22, 0x28, 0x24, 0x27,           \
            0x2d, 0x2b, 0x2a, 0x20, 0x28, 0x25, 0x21, 0x2c,           \
            0x20, 0x2d, 0x2c, 0x2a, 0x29, 0x22, 0x2f, 0x24,           \
            0x2e, 0x21, 0x23, 0x2f, 0x25, 0x2e, 0x26, 0x23,           \
            0xc2, 0xc8, 0xcc, 0xc5, 0xcf, 0xc3, 0xca, 0xc0,           \
            0xc4, 0xcd, 0xc9, 0xc6, 0xc1, 0xce, 0xc6, 0xc9,           \
            0xcd, 0xc2, 0xc3, 0xcf, 0xc0, 0xcc, 0xc5, 0xca,           \
            0xc7, 0xcb, 0xce, 0xc1, 0xcb, 0xc7, 0xc8, 0xc4,           \
            0xcb, 0xc6, 0xc7, 0xc9, 0xc2, 0xc8, 0xc4, 0xc7,           \
            0xcd, 0xcb, 0xca, 0xc0, 0xc8, 0xc5, 0xc1, 0xcc,           \
            0xc0, 0xcd, 0xcc, 0xca, 0xc9, 0xc2, 0xcf, 0xc4,           \
            0xce, 0xc1, 0xc3, 0xcf, 0xc5, 0xce, 0xc6, 0xc3,           \
            0x52, 0x58, 0x5c, 0x55, 0x5f, 0x53, 0x5a, 0x50,           \
            0x54, 0x5d, 0x59, 0x56, 0x51, 0x5e, 0x56, 0x59,           \
            0x5d, 0x52, 0x53, 0x5f, 0x50, 0x5c, 0x55, 0x5a,           \
            0x57, 0x5b, 0x5e, 0x51, 0x5b, 0x57, 0x58, 0x54,           \
            0x5b, 0x56, 0x57, 0x59, 0x52, 0x58, 0x54, 0x57,           \
            0x5d, 0x5b, 0x5a, 0x50, 0x58, 0x55, 0x51, 0x5c,           \
            0x50, 0x5d, 0x5c, 0x5a, 0x59, 0x52, 0x5f, 0x54,           \
            0x5e, 0x51, 0x53, 0x5f, 0x55, 0x5e, 0x56, 0x53,           \
            0xd2, 0xd8, 0xdc, 0xd5, 0xdf, 0xd3, 0xda, 0xd0,           \
            0xd4, 0xdd, 0xd9, 0xd6, 0xd1, 0xde, 0xd6, 0xd9,           \
            0xdd, 0xd2, 0xd3, 0xdf, 0xd0, 0xdc, 0xd5, 0xda,           \
            0xd7, 0xdb, 0xde, 0xd1, 0xdb, 0xd7, 0xd8, 0xd4,           \
            0xdb, 0xd6, 0xd7, 0xd9, 0xd2, 0xd8, 0xd4, 0xd7,           \
            0xdd, 0xdb, 0xda, 0xd0, 0xd8, 0xd5, 0xd1, 0xdc,           \
            0xd0, 0xdd, 0xdc, 0xda, 0xd9, 0xd2, 0xdf, 0xd4,           \
            0xde, 0xd1, 0xd3, 0xdf, 0xd5, 0xde, 0xd6, 0xd3,           \
            0x72, 0x78, 0x7c, 0x75, 0x7f, 0x73, 0x7a, 0x70,           \
            0x74, 0x7d, 0x79, 0x76, 0x71, 0x7e, 0x76, 0x79,           \
            0x7d, 0x72, 0x73, 0x7f, 0x70, 0x7c, 0x75, 0x7a,           \
            0x77, 0x7b, 0x7e, 0x71, 0x7b, 0x77, 0x78, 0x74,           \
            0x7b, 0x76, 0x77, 0x79, 0x72, 0x78, 0x74, 0x77,           \
            0x7d, 0x7b, 0x7a, 0x70, 0x78, 0x75, 0x71, 0x7c,           \
            0x70, 0x7d, 0x7c, 0x7a, 0x79, 0x72, 0x7f, 0x74,           \
            0x7e, 0x71, 0x73, 0x7f, 0x75, 0x7e, 0x76, 0x73,           \
            0x32, 0x38, 0x3c, 0x35, 0x3f, 0x33, 0x3a, 0x30,           \
            0x34, 0x3d, 0x39, 0x36, 0x31, 0x3e, 0x36, 0x39,           \
            0x3d, 0x32, 0x33, 0x3f, 0x30, 0x3c, 0x35, 0x3a,           \
            0x37, 0x3b, 0x3e, 0x31, 0x3b, 0x37, 0x38, 0x34,           \
            0x3b, 0x36, 0x37, 0x39, 0x32, 0x38, 0x34, 0x37,           \
            0x3d, 0x3b, 0x3a, 0x30, 0x38, 0x35, 0x31, 0x3c,           \
            0x30, 0x3d, 0x3c, 0x3a, 0x39, 0x32, 0x3f, 0x34,           \
            0x3e, 0x31, 0x33, 0x3f, 0x35, 0x3e, 0x36, 0x33,           \
            0xa2, 0xa8, 0xac, 0xa5, 0xaf, 0xa3, 0xaa, 0xa0,           \
            0xa4, 0xad, 0xa9, 0xa6, 0xa1, 0xae, 0xa6, 0xa9,           \
            0xad, 0xa2, 0xa3, 0xaf, 0xa0, 0xac, 0xa5, 0xaa,           \
            0xa7, 0xab, 0xae, 0xa1, 0xab, 0xa7, 0xa8, 0xa4,           \
            0xab, 0xa6, 0xa7, 0xa9, 0xa2, 0xa8, 0xa4, 0xa7,           \
            0xad, 0xab, 0xaa, 0xa0, 0xa8, 0xa5, 0xa1, 0xac,           \
            0xa0, 0xad, 0xac, 0xaa, 0xa9, 0xa2, 0xaf, 0xa4,           \
            0xae, 0xa1, 0xa3, 0xaf, 0xa5, 0xae, 0xa6, 0xa3,           \
            0x42, 0x48, 0x4c, 0x45, 0x4f, 0x43, 0x4a, 0x40,           \
            0x44, 0x4d, 0x49, 0x46, 0x41, 0x4e, 0x46, 0x49,           \
            0x4d, 0x42, 0x43, 0x4f, 0x40, 0x4c, 0x45, 0x4a,           \
            0x47, 0x4b, 0x4e, 0x41, 0x4b, 0x47, 0x48, 0x44,           \
            0x4b, 0x46, 0x47, 0x49, 0x42, 0x48, 0x44, 0x47,           \
            0x4d, 0x4b, 0x4a, 0x40, 0x48, 0x45, 0x41, 0x4c,           \
            0x40, 0x4d, 0x4c, 0x4a, 0x49, 0x42, 0x4f, 0x44,           \
            0x4e, 0x41, 0x43, 0x4f, 0x45, 0x4e, 0x46, 0x43,           \
            0xd2, 0xd8, 0xdc, 0xd5, 0xdf, 0xd3, 0xda, 0xd0,           \
            0xd4, 0xdd, 0xd9, 0xd6, 0xd1, 0xde, 0xd6, 0xd9,           \
            0xdd, 0xd2, 0xd3, 0xdf, 0xd0, 0xdc, 0xd5, 0xda,           \
            0xd7, 0xdb, 0xde, 0xd1, 0xdb, 0xd7, 0xd8, 0xd4,           \
            0xdb, 0xd6, 0xd7, 0xd9, 0xd2, 0xd8, 0xd4, 0xd7,           \
            0xdd, 0xdb, 0xda, 0xd0, 0xd8, 0xd5, 0xd1, 0xdc,           \
            0xd0, 0xdd, 0xdc, 0xda, 0xd9, 0xd2, 0xdf, 0xd4,           \
            0xde, 0xd1, 0xd3, 0xdf, 0xd5, 0xde, 0xd6, 0xd3,           \
            0x92, 0x98, 0x9c, 0x95, 0x9f, 0x93, 0x9a, 0x90,           \
            0x94, 0x9d, 0x99, 0x96, 0x91, 0x9e, 0x96, 0x99,           \
            0x9d, 0x92, 0x93, 0x9f, 0x90, 0x9c, 0x95, 0x9a,           \
            0x97, 0x9b, 0x9e, 0x91, 0x9b, 0x97, 0x98, 0x94,           \
            0x9b, 0x96, 0x97, 0x99, 0x92, 0x98, 0x94, 0x97,           \
            0x9d, 0x9b, 0x9a, 0x90, 0x98, 0x95, 0x91, 0x9c,           \
            0x90, 0x9d, 0x9c, 0x9a, 0x99, 0x92, 0x9f, 0x94,           \
            0x9e, 0x91, 0x93, 0x9f, 0x95, 0x9e, 0x96, 0x93,           \
            0x62, 0x68, 0x6c, 0x65, 0x6f, 0x63, 0x6a, 0x60,           \
            0x64, 0x6d, 0x69, 0x66, 0x61, 0x6e, 0x66, 0x69,           \
            0x6d, 0x62, 0x63, 0x6f, 0x60, 0x6c, 0x65, 0x6a,           \
            0x67, 0x6b, 0x6e, 0x61, 0x6b, 0x67, 0x68, 0x64,           \
            0x6b, 0x66, 0x67, 0x69, 0x62, 0x68, 0x64, 0x67,           \
            0x6d, 0x6b, 0x6a, 0x60, 0x68, 0x65, 0x61, 0x6c,           \
            0x60, 0x6d, 0x6c, 0x6a, 0x69, 0x62, 0x6f, 0x64,           \
            0x6e, 0x61, 0x63, 0x6f, 0x65, 0x6e, 0x66, 0x63,           \
            0xf2, 0xf8, 0xfc, 0xf5, 0xff, 0xf3, 0xfa, 0xf0,           \
            0xf4, 0xfd, 0xf9, 0xf6, 0xf1, 0xfe, 0xf6, 0xf9,           \
            0xfd, 0xf2, 0xf3, 0xff, 0xf0, 0xfc, 0xf5, 0xfa,           \
            0xf7, 0xfb, 0xfe, 0xf1, 0xfb, 0xf7, 0xf8, 0xf4,           \
            0xfb, 0xf6, 0xf7, 0xf9, 0xf2, 0xf8, 0xf4, 0xf7,           \
            0xfd, 0xfb, 0xfa, 0xf0, 0xf8, 0xf5, 0xf1, 0xfc,           \
            0xf0, 0xfd, 0xfc, 0xfa, 0xf9, 0xf2, 0xff, 0xf4,           \
            0xfe, 0xf1, 0xf3, 0xff, 0xf5, 0xfe, 0xf6, 0xf3,           \
            0x32, 0x38, 0x3c, 0x35, 0x3f, 0x33, 0x3a, 0x30,           \
            0x34, 0x3d, 0x39, 0x36, 0x31, 0x3e, 0x36, 0x39,           \
            0x3d, 0x32, 0x33, 0x3f, 0x30, 0x3c, 0x35, 0x3a,           \
            0x37, 0x3b, 0x3e, 0x31, 0x3b, 0x37, 0x38, 0x34,           \
            0x3b, 0x36, 0x37, 0x39, 0x32, 0x38, 0x34, 0x37,           \
            0x3d, 0x3b, 0x3a, 0x30, 0x38, 0x35, 0x31, 0x3c,           \
            0x30, 0x3d, 0x3c, 0x3a, 0x39, 0x32, 0x3f, 0x34,           \
            0x3e, 0x31, 0x33, 0x3f, 0x35, 0x3e, 0x36, 0x33,           \
            0x02, 0x08, 0x0c, 0x05, 0x0f, 0x03, 0x0a, 0x00,           \
            0x04, 0x0d, 0x09, 0x06, 0x01, 0x0e, 0x06, 0x09,           \
            0x0d, 0x02, 0x03, 0x0f, 0x00, 0x0c, 0x05, 0x0a,           \
            0x07, 0x0b, 0x0e, 0x01, 0x0b, 0x07, 0x08, 0x04,           \
            0x0b, 0x06, 0x07, 0x09, 0x02, 0x08, 0x04, 0x07,           \
            0x0d, 0x0b, 0x0a, 0x00, 0x08, 0x05, 0x01, 0x0c,           \
            0x00, 0x0d, 0x0c, 0x0a, 0x09, 0x02, 0x0f, 0x04,           \
            0x0e, 0x01, 0x03, 0x0f, 0x05, 0x0e, 0x06, 0x03,           \
            0xc2, 0xc8, 0xcc, 0xc5, 0xcf, 0xc3, 0xca, 0xc0,           \
            0xc4, 0xcd, 0xc9, 0xc6, 0xc1, 0xce, 0xc6, 0xc9,           \
            0xcd, 0xc2, 0xc3, 0xcf, 0xc0, 0xcc, 0xc5, 0xca,           \
            0xc7, 0xcb, 0xce, 0xc1, 0xcb, 0xc7, 0xc8, 0xc4,           \
            0xcb, 0xc6, 0xc7, 0xc9, 0xc2, 0xc8, 0xc4, 0xc7,           \
            0xcd, 0xcb, 0xca, 0xc0, 0xc8, 0xc5, 0xc1, 0xcc,           \
            0xc0, 0xcd, 0xcc, 0xca, 0xc9, 0xc2, 0xcf, 0xc4,           \
            0xce, 0xc1, 0xc3, 0xcf, 0xc5, 0xce, 0xc6, 0xc3,           \
            0x22, 0x28, 0x2c, 0x25, 0x2f, 0x23, 0x2a, 0x20,           \
            0x24, 0x2d, 0x29, 0x26, 0x21, 0x2e, 0x26, 0x29,           \
            0x2d, 0x22, 0x23, 0x2f, 0x20, 0x2c, 0x25, 0x2a,           \
            0x27, 0x2b, 0x2e, 0x21, 0x2b, 0x27, 0x28, 0x24,           \
            0x2b, 0x26, 0x27, 0x29, 0x22, 0x28, 0x24, 0x27,           \
            0x2d, 0x2b, 0x2a, 0x20, 0x28, 0x25, 0x21, 0x2c,           \
            0x20, 0x2d, 0x2c, 0x2a, 0x29, 0x22, 0x2f, 0x24,           \
            0x2e, 0x21, 0x23, 0x2f, 0x25, 0x2e, 0x26, 0x23,           \
            0x82, 0x88, 0x8c, 0x85, 0x8f, 0x83, 0x8a, 0x80,           \
            0x84, 0x8d, 0x89, 0x86, 0x81, 0x8e, 0x86, 0x89,           \
            0x8d, 0x82, 0x83, 0x8f, 0x80, 0x8c, 0x85, 0x8a,           \
            0x87, 0x8b, 0x8e, 0x81, 0x8b, 0x87, 0x88, 0x84,           \
            0x8b, 0x86, 0x87, 0x89, 0x82, 0x88, 0x84, 0x87,           \
            0x8d, 0x8b, 0x8a, 0x80, 0x88, 0x85, 0x81, 0x8c,           \
            0x80, 0x8d, 0x8c, 0x8a, 0x89, 0x82, 0x8f, 0x84,           \
            0x8e, 0x81, 0x83, 0x8f, 0x85, 0x8e, 0x86, 0x83,           \
            0x72, 0x78, 0x7c, 0x75, 0x7f, 0x73, 0x7a, 0x70,           \
            0x74, 0x7d, 0x79, 0x76, 0x71, 0x7e, 0x76, 0x79,           \
            0x7d, 0x72, 0x73, 0x7f, 0x70, 0x7c, 0x75, 0x7a,           \
            0x77, 0x7b, 0x7e, 0x71, 0x7b, 0x77, 0x78, 0x74,           \
            0x7b, 0x76, 0x77, 0x79, 0x72, 0x78, 0x74, 0x77,           \
            0x7d, 0x7b, 0x7a, 0x70, 0x78, 0x75, 0x71, 0x7c,           \
            0x70, 0x7d, 0x7c, 0x7a, 0x79, 0x72, 0x7f, 0x74,           \
            0x7e, 0x71, 0x73, 0x7f, 0x75, 0x7e, 0x76, 0x73,           \
            0xb2, 0xb8, 0xbc, 0xb5, 0xbf, 0xb3, 0xba, 0xb0,           \
            0xb4, 0xbd, 0xb9, 0xb6, 0xb1, 0xbe, 0xb6, 0xb9,           \
            0xbd, 0xb2, 0xb3, 0xbf, 0xb0, 0xbc, 0xb5, 0xba,           \
            0xb7, 0xbb, 0xbe, 0xb1, 0xbb, 0xb7, 0xb8, 0xb4,           \
            0xbb, 0xb6, 0xb7, 0xb9, 0xb2, 0xb8, 0xb4, 0xb7,           \
            0xbd, 0xbb, 0xba, 0xb0, 0xb8, 0xb5, 0xb1, 0xbc,           \
            0xb0, 0xbd, 0xbc, 0xba, 0xb9, 0xb2, 0xbf, 0xb4,           \
            0xbe, 0xb1, 0xb3, 0xbf, 0xb5, 0xbe, 0xb6, 0xb3,           \
            0x82, 0x88, 0x8c, 0x85, 0x8f, 0x83, 0x8a, 0x80,           \
            0x84, 0x8d, 0x89, 0x86, 0x81, 0x8e, 0x86, 0x89,           \
            0x8d, 0x82, 0x83, 0x8f, 0x80, 0x8c, 0x85, 0x8a,           \
            0x87, 0x8b, 0x8e, 0x81, 0x8b, 0x87, 0x88, 0x84,           \
            0x8b, 0x86, 0x87, 0x89, 0x82, 0x88, 0x84, 0x87,           \
            0x8d, 0x8b, 0x8a, 0x80, 0x88, 0x85, 0x81, 0x8c,           \
            0x80, 0x8d, 0x8c, 0x8a, 0x89, 0x82, 0x8f, 0x84,           \
            0x8e, 0x81, 0x83, 0x8f, 0x85, 0x8e, 0x86, 0x83,           \
            0x42, 0x48, 0x4c, 0x45, 0x4f, 0x43, 0x4a, 0x40,           \
            0x44, 0x4d, 0x49, 0x46, 0x41, 0x4e, 0x46, 0x49,           \
            0x4d, 0x42, 0x43, 0x4f, 0x40, 0x4c, 0x45, 0x4a,           \
            0x47, 0x4b, 0x4e, 0x41, 0x4b, 0x47, 0x48, 0x44,           \
            0x4b, 0x46, 0x47, 0x49, 0x42, 0x48, 0x44, 0x47,           \
            0x4d, 0x4b, 0x4a, 0x40, 0x48, 0x45, 0x41, 0x4c,           \
            0x40, 0x4d, 0x4c, 0x4a, 0x49, 0x42, 0x4f, 0x44,           \
            0x4e, 0x41, 0x43, 0x4f, 0x45, 0x4e, 0x46, 0x43,           \
            0xe2, 0xe8, 0xec, 0xe5, 0xef, 0xe3, 0xea, 0xe0,           \
            0xe4, 0xed, 0xe9, 0xe6, 0xe1, 0xee, 0xe6, 0xe9,           \
            0xed, 0xe2, 0xe3, 0xef, 0xe0, 0xec, 0xe5, 0xea,           \
            0xe7, 0xeb, 0xee, 0xe1, 0xeb, 0xe7, 0xe8, 0xe4,           \
            0xeb, 0xe6, 0xe7, 0xe9, 0xe2, 0xe8, 0xe4, 0xe7,           \
            0xed, 0xeb, 0xea, 0xe0, 0xe8, 0xe5, 0xe1, 0xec,           \
            0xe0, 0xed, 0xec, 0xea, 0xe9, 0xe2, 0xef, 0xe4,           \
            0xee, 0xe1, 0xe3, 0xef, 0xe5, 0xee, 0xe6, 0xe3,           \
            0x12, 0x18, 0x1c, 0x15, 0x1f, 0x13, 0x1a, 0x10,           \
            0x14, 0x1d, 0x19, 0x16, 0x11, 0x1e, 0x16, 0x19,           \
            0x1d, 0x12, 0x13, 0x1f, 0x10, 0x1c, 0x15, 0x1a,           \
            0x17, 0x1b, 0x1e, 0x11, 0x1b, 0x17, 0x18, 0x14,           \
            0x1b, 0x16, 0x17, 0x19, 0x12, 0x18, 0x14, 0x17,           \
            0x1d, 0x1b, 0x1a, 0x10, 0x18, 0x15, 0x11, 0x1c,           \
            0x10, 0x1d, 0x1c, 0x1a, 0x19, 0x12, 0x1f, 0x14,           \
            0x1e, 0x11, 0x13, 0x1f, 0x15, 0x1e, 0x16, 0x13,           \
            0x82, 0x88, 0x8c, 0x85, 0x8f, 0x83, 0x8a, 0x80,           \
            0x84, 0x8d, 0x89, 0x86, 0x81, 0x8e, 0x86, 0x89,           \
            0x8d, 0x82, 0x83, 0x8f, 0x80, 0x8c, 0x85, 0x8a,           \
            0x87, 0x8b, 0x8e, 0x81, 0x8b, 0x87, 0x88, 0x84,           \
            0x8b, 0x86, 0x87, 0x89, 0x82, 0x88, 0x84, 0x87,           \
            0x8d, 0x8b, 0x8a, 0x80, 0x88, 0x85, 0x81, 0x8c,           \
            0x80, 0x8d, 0x8c, 0x8a, 0x89, 0x82, 0x8f, 0x84,           \
            0x8e, 0x81, 0x83, 0x8f, 0x85, 0x8e, 0x86, 0x83,           \
            0x42, 0x48, 0x4c, 0x45, 0x4f, 0x43, 0x4a, 0x40,           \
            0x44, 0x4d, 0x49, 0x46, 0x41, 0x4e, 0x46, 0x49,           \
            0x4d, 0x42, 0x43, 0x4f, 0x40, 0x4c, 0x45, 0x4a,           \
            0x47, 0x4b, 0x4e, 0x41, 0x4b, 0x47, 0x48, 0x44,           \
            0x4b, 0x46, 0x47, 0x49, 0x42, 0x48, 0x44, 0x47,           \
            0x4d, 0x4b, 0x4a, 0x40, 0x48, 0x45, 0x41, 0x4c,           \
            0x40, 0x4d, 0x4c, 0x4a, 0x49, 0x42, 0x4f, 0x44,           \
            0x4e, 0x41, 0x43, 0x4f, 0x45, 0x4e, 0x46, 0x43,           \
            0x32, 0x38, 0x3c, 0x35, 0x3f, 0x33, 0x3a, 0x30,           \
            0x34, 0x3d, 0x39, 0x36, 0x31, 0x3e, 0x36, 0x39,           \
            0x3d, 0x32, 0x33, 0x3f, 0x30, 0x3c, 0x35, 0x3a,           \
            0x37, 0x3b, 0x3e, 0x31, 0x3b, 0x37, 0x38, 0x34,           \
            0x3b, 0x36, 0x37, 0x39, 0x32, 0x38, 0x34, 0x37,           \
            0x3d, 0x3b, 0x3a, 0x30, 0x38, 0x35, 0x31, 0x3c,           \
            0x30, 0x3d, 0x3c, 0x3a, 0x39, 0x32, 0x3f, 0x34,           \
            0x3e, 0x31, 0x33, 0x3f, 0x35, 0x3e, 0x36, 0x33,           \
            0xf2, 0xf8, 0xfc, 0xf5, 0xff, 0xf3, 0xfa, 0xf0,           \
            0xf4, 0xfd, 0xf9, 0xf6, 0xf1, 0xfe, 0xf6, 0xf9,           \
            0xfd, 0xf2, 0xf3, 0xff, 0xf0, 0xfc, 0xf5, 0xfa,           \
            0xf7, 0xfb, 0xfe, 0xf1, 0xfb, 0xf7, 0xf8, 0xf4,           \
            0xfb, 0xf6, 0xf7, 0xf9, 0xf2, 0xf8, 0xf4, 0xf7,           \
            0xfd, 0xfb, 0xfa, 0xf0, 0xf8, 0xf5, 0xf1, 0xfc,           \
            0xf0, 0xfd, 0xfc, 0xfa, 0xf9, 0xf2, 0xff, 0xf4,           \
            0xfe, 0xf1, 0xf3, 0xff, 0xf5, 0xfe, 0xf6, 0xf3,           \
            0x52, 0x58, 0x5c, 0x55, 0x5f, 0x53, 0x5a, 0x50,           \
            0x54, 0x5d, 0x59, 0x56, 0x51, 0x5e, 0x56, 0x59,           \
            0x5d, 0x52, 0x53, 0x5f, 0x50, 0x5c, 0x55, 0x5a,           \
            0x57, 0x5b, 0x5e, 0x51, 0x5b, 0x57, 0x58, 0x54,           \
            0x5b, 0x56, 0x57, 0x59, 0x52, 0x58, 0x54, 0x57,           \
            0x5d, 0x5b, 0x5a, 0x50, 0x58, 0x55, 0x51, 0x5c,           \
            0x50, 0x5d, 0x5c, 0x5a, 0x59, 0x52, 0x5f, 0x54,           \
            0x5e, 0x51, 0x53, 0x5f, 0x55, 0x5e, 0x56, 0x53,           \
            0x22, 0x28, 0x2c, 0x25, 0x2f, 0x23, 0x2a, 0x20,           \
            0x24, 0x2d, 0x29, 0x26, 0x21, 0x2e, 0x26, 0x29,           \
            0x2d, 0x22, 0x23, 0x2f, 0x20, 0x2c, 0x25, 0x2a,           \
            0x27, 0x2b, 0x2e, 0x21, 0x2b, 0x27, 0x28, 0x24,           \
            0x2b, 0x26, 0x27, 0x29, 0x22, 0x28, 0x24, 0x27,           \
            0x2d, 0x2b, 0x2a, 0x20, 0x28, 0x25, 0x21, 0x2c,           \
            0x20, 0x2d, 0x2c, 0x2a, 0x29, 0x22, 0x2f, 0x24,           \
            0x2e, 0x21, 0x23, 0x2f, 0x25, 0x2e, 0x26, 0x23,           \
            0x02, 0x08, 0x0c, 0x05, 0x0f, 0x03, 0x0a, 0x00,           \
            0x04, 0x0d, 0x09, 0x06, 0x01, 0x0e, 0x06, 0x09,           \
            0x0d, 0x02, 0x03, 0x0f, 0x00, 0x0c, 0x05, 0x0a,           \
            0x07, 0x0b, 0x0e, 0x01, 0x0b, 0x07, 0x08, 0x04,           \
            0x0b, 0x06, 0x07, 0x09, 0x02, 0x08, 0x04, 0x07,           \
            0x0d, 0x0b, 0x0a, 0x00, 0x08, 0x05, 0x01, 0x0c,           \
            0x00, 0x0d, 0x0c, 0x0a, 0x09, 0x02, 0x0f, 0x04,           \
            0x0e, 0x01, 0x03, 0x0f, 0x05, 0x0e, 0x06, 0x03,           \
            0xc2, 0xc8, 0xcc, 0xc5, 0xcf, 0xc3, 0xca, 0xc0,           \
            0xc4, 0xcd, 0xc9, 0xc6, 0xc1, 0xce, 0xc6, 0xc9,           \
            0xcd, 0xc2, 0xc3, 0xcf, 0xc0, 0xcc, 0xc5, 0xca,           \
            0xc7, 0xcb, 0xce, 0xc1, 0xcb, 0xc7, 0xc8, 0xc4,           \
            0xcb, 0xc6, 0xc7, 0xc9, 0xc2, 0xc8, 0xc4, 0xc7,           \
            0xcd, 0xcb, 0xca, 0xc0, 0xc8, 0xc5, 0xc1, 0xcc,           \
            0xc0, 0xcd, 0xcc, 0xca, 0xc9, 0xc2, 0xcf, 0xc4,           \
            0xce, 0xc1, 0xc3, 0xcf, 0xc5, 0xce, 0xc6, 0xc3,           \
            0xb2, 0xb8, 0xbc, 0xb5, 0xbf, 0xb3, 0xba, 0xb0,           \
            0xb4, 0xbd, 0xb9, 0xb6, 0xb1, 0xbe, 0xb6, 0xb9,           \
            0xbd, 0xb2, 0xb3, 0xbf, 0xb0, 0xbc, 0xb5, 0xba,           \
            0xb7, 0xbb, 0xbe, 0xb1, 0xbb, 0xb7, 0xb8, 0xb4,           \
            0xbb, 0xb6, 0xb7, 0xb9, 0xb2, 0xb8, 0xb4, 0xb7,           \
            0xbd, 0xbb, 0xba, 0xb0, 0xb8, 0xb5, 0xb1, 0xbc,           \
            0xb0, 0xbd, 0xbc, 0xba, 0xb9, 0xb2, 0xbf, 0xb4,           \
            0xbe, 0xb1, 0xb3, 0xbf, 0xb5, 0xbe, 0xb6, 0xb3,           \
            0x72, 0x78, 0x7c, 0x75, 0x7f, 0x73, 0x7a, 0x70,           \
            0x74, 0x7d, 0x79, 0x76, 0x71, 0x7e, 0x76, 0x79,           \
            0x7d, 0x72, 0x73, 0x7f, 0x70, 0x7c, 0x75, 0x7a,           \
            0x77, 0x7b, 0x7e, 0x71, 0x7b, 0x77, 0x78, 0x74,           \
            0x7b, 0x76, 0x77, 0x79, 0x72, 0x78, 0x74, 0x77,           \
            0x7d, 0x7b, 0x7a, 0x70, 0x78, 0x75, 0x71, 0x7c,           \
            0x70, 0x7d, 0x7c, 0x7a, 0x79, 0x72, 0x7f, 0x74,           \
            0x7e, 0x71, 0x73, 0x7f, 0x75, 0x7e, 0x76, 0x73,           \
            0x62, 0x68, 0x6c, 0x65, 0x6f, 0x63, 0x6a, 0x60,           \
            0x64, 0x6d, 0x69, 0x66, 0x61, 0x6e, 0x66, 0x69,           \
            0x6d, 0x62, 0x63, 0x6f, 0x60, 0x6c, 0x65, 0x6a,           \
            0x67, 0x6b, 0x6e, 0x61, 0x6b, 0x67, 0x68, 0x64,           \
            0x6b, 0x66, 0x67, 0x69, 0x62, 0x68, 0x64, 0x67,           \
            0x6d, 0x6b, 0x6a, 0x60, 0x68, 0x65, 0x61, 0x6c,           \
            0x60, 0x6d, 0x6c, 0x6a, 0x69, 0x62, 0x6f, 0x64,           \
            0x6e, 0x61, 0x63, 0x6f, 0x65, 0x6e, 0x66, 0x63,           \
            0x92, 0x98, 0x9c, 0x95, 0x9f, 0x93, 0x9a, 0x90,           \
            0x94, 0x9d, 0x99, 0x96, 0x91, 0x9e, 0x96, 0x99,           \
            0x9d, 0x92, 0x93, 0x9f, 0x90, 0x9c, 0x95, 0x9a,           \
            0x97, 0x9b, 0x9e, 0x91, 0x9b, 0x97, 0x98, 0x94,           \
            0x9b, 0x96, 0x97, 0x99, 0x92, 0x98, 0x94, 0x97,           \
            0x9d, 0x9b, 0x9a, 0x90, 0x98, 0x95, 0x91, 0x9c,           \
            0x90, 0x9d, 0x9c, 0x9a, 0x99, 0x92, 0x9f, 0x94,           \
            0x9e, 0x91, 0x93, 0x9f, 0x95, 0x9e, 0x96, 0x93,           \
            0xe2, 0xe8, 0xec, 0xe5, 0xef, 0xe3, 0xea, 0xe0,           \
            0xe4, 0xed, 0xe9, 0xe6, 0xe1, 0xee, 0xe6, 0xe9,           \
            0xed, 0xe2, 0xe3, 0xef, 0xe0, 0xec, 0xe5, 0xea,           \
            0xe7, 0xeb, 0xee, 0xe1, 0xeb, 0xe7, 0xe8, 0xe4,           \
            0xeb, 0xe6, 0xe7, 0xe9, 0xe2, 0xe8, 0xe4, 0xe7,           \
            0xed, 0xeb, 0xea, 0xe0, 0xe8, 0xe5, 0xe1, 0xec,           \
            0xe0, 0xed, 0xec, 0xea, 0xe9, 0xe2, 0xef, 0xe4,           \
            0xee, 0xe1, 0xe3, 0xef, 0xe5, 0xee, 0xe6, 0xe3,           \
            0x12, 0x18, 0x1c, 0x15, 0x1f, 0x13, 0x1a, 0x10,           \
            0x14, 0x1d, 0x19, 0x16, 0x11, 0x1e, 0x16, 0x19,           \
            0x1d, 0x12, 0x13, 0x1f, 0x10, 0x1c, 0x15, 0x1a,           \
            0x17, 0x1b, 0x1e, 0x11, 0x1b, 0x17, 0x18, 0x14,           \
            0x1b, 0x16, 0x17, 0x19, 0x12, 0x18, 0x14, 0x17,           \
            0x1d, 0x1b, 0x1a, 0x10, 0x18, 0x15, 0x11, 0x1c,           \
            0x10, 0x1d, 0x1c, 0x1a, 0x19, 0x12, 0x1f, 0x14,           \
            0x1e, 0x11, 0x13, 0x1f, 0x15, 0x1e, 0x16, 0x13,           \
            0x92, 0x98, 0x9c, 0x95, 0x9f, 0x93, 0x9a, 0x90,           \
            0x94, 0x9d, 0x99, 0x96, 0x91, 0x9e, 0x96, 0x99,           \
            0x9d, 0x92, 0x93, 0x9f, 0x90, 0x9c, 0x95, 0x9a,           \
            0x97, 0x9b, 0x9e, 0x91, 0x9b, 0x97, 0x98, 0x94,           \
            0x9b, 0x96, 0x97, 0x99, 0x92, 0x98, 0x94, 0x97,           \
            0x9d, 0x9b, 0x9a, 0x90, 0x98, 0x95, 0x91, 0x9c,           \
            0x90, 0x9d, 0x9c, 0x9a, 0x99, 0x92, 0x9f, 0x94,           \
            0x9e, 0x91, 0x93, 0x9f, 0x95, 0x9e, 0x96, 0x93,           \
            0x62, 0x68, 0x6c, 0x65, 0x6f, 0x63, 0x6a, 0x60,           \
            0x64, 0x6d, 0x69, 0x66, 0x61, 0x6e, 0x66, 0x69,           \
            0x6d, 0x62, 0x63, 0x6f, 0x60, 0x6c, 0x65, 0x6a,           \
            0x67, 0x6b, 0x6e, 0x61, 0x6b, 0x67, 0x68, 0x64,           \
            0x6b, 0x66, 0x67, 0x69, 0x62, 0x68, 0x64, 0x67,           \
            0x6d, 0x6b, 0x6a, 0x60, 0x68, 0x65, 0x61, 0x6c,           \
            0x60, 0x6d, 0x6c, 0x6a, 0x69, 0x62, 0x6f, 0x64,           \
            0x6e, 0x61, 0x63, 0x6f, 0x65, 0x6e, 0x66, 0x63,           \
            0xf2, 0xf8, 0xfc, 0xf5, 0xff, 0xf3, 0xfa, 0xf0,           \
            0xf4, 0xfd, 0xf9, 0xf6, 0xf1, 0xfe, 0xf6, 0xf9,           \
            0xfd, 0xf2, 0xf3, 0xff, 0xf0, 0xfc, 0xf5, 0xfa,           \
            0xf7, 0xfb, 0xfe, 0xf1, 0xfb, 0xf7, 0xf8, 0xf4,           \
            0xfb, 0xf6, 0xf7, 0xf9, 0xf2, 0xf8, 0xf4, 0xf7,           \
            0xfd, 0xfb, 0xfa, 0xf0, 0xf8, 0xf5, 0xf1, 0xfc,           \
            0xf0, 0xfd, 0xfc, 0xfa, 0xf9, 0xf2, 0xff, 0xf4,           \
            0xfe, 0xf1, 0xf3, 0xff, 0xf5, 0xfe, 0xf6, 0xf3,           \
            0x82, 0x88, 0x8c, 0x85, 0x8f, 0x83, 0x8a, 0x80,           \
            0x84, 0x8d, 0x89, 0x86, 0x81, 0x8e, 0x86, 0x89,           \
            0x8d, 0x82, 0x83, 0x8f, 0x80, 0x8c, 0x85, 0x8a,           \
            0x87, 0x8b, 0x8e, 0x81, 0x8b, 0x87, 0x88, 0x84,           \
            0x8b, 0x86, 0x87, 0x89, 0x82, 0x88, 0x84, 0x87,           \
            0x8d, 0x8b, 0x8a, 0x80, 0x88, 0x85, 0x81, 0x8c,           \
            0x80, 0x8d, 0x8c, 0x8a, 0x89, 0x82, 0x8f, 0x84,           \
            0x8e, 0x81, 0x83, 0x8f, 0x85, 0x8e, 0x86, 0x83,           \
            0xa2, 0xa8, 0xac, 0xa5, 0xaf, 0xa3, 0xaa, 0xa0,           \
            0xa4, 0xad, 0xa9, 0xa6, 0xa1, 0xae, 0xa6, 0xa9,           \
            0xad, 0xa2, 0xa3, 0xaf, 0xa0, 0xac, 0xa5, 0xaa,           \
            0xa7, 0xab, 0xae, 0xa1, 0xab, 0xa7, 0xa8, 0xa4,           \
            0xab, 0xa6, 0xa7, 0xa9, 0xa2, 0xa8, 0xa4, 0xa7,           \
            0xad, 0xab, 0xaa, 0xa0, 0xa8, 0xa5, 0xa1, 0xac,           \
            0xa0, 0xad, 0xac, 0xaa, 0xa9, 0xa2, 0xaf, 0xa4,           \
            0xae, 0xa1, 0xa3, 0xaf, 0xa5, 0xae, 0xa6, 0xa3,           \
            0x32, 0x38, 0x3c, 0x35, 0x3f, 0x33, 0x3a, 0x30,           \
            0x34, 0x3d, 0x39, 0x36, 0x31, 0x3e, 0x36, 0x39,           \
            0x3d, 0x32, 0x33, 0x3f, 0x30, 0x3c, 0x35, 0x3a,           \
            0x37, 0x3b, 0x3e, 0x31, 0x3b, 0x37, 0x38, 0x34,           \
            0x3b, 0x36, 0x37, 0x39, 0x32, 0x38, 0x34, 0x37,           \
            0x3d, 0x3b, 0x3a, 0x30, 0x38, 0x35, 0x31, 0x3c,           \
            0x30, 0x3d, 0x3c, 0x3a, 0x39, 0x32, 0x3f, 0x34,           \
            0x3e, 0x31, 0x33, 0x3f, 0x35, 0x3e, 0x36, 0x33,           \
            0xc2, 0xc8, 0xcc, 0xc5, 0xcf, 0xc3, 0xca, 0xc0,           \
            0xc4, 0xcd, 0xc9, 0xc6, 0xc1, 0xce, 0xc6, 0xc9,           \
            0xcd, 0xc2, 0xc3, 0xcf, 0xc0, 0xcc, 0xc5, 0xca,           \
            0xc7, 0xcb, 0xce, 0xc1, 0xcb, 0xc7, 0xc8, 0xc4,           \
            0xcb, 0xc6, 0xc7, 0xc9, 0xc2, 0xc8, 0xc4, 0xc7,           \
            0xcd, 0xcb, 0xca, 0xc0, 0xc8, 0xc5, 0xc1, 0xcc,           \
            0xc0, 0xcd, 0xcc, 0xca, 0xc9, 0xc2, 0xcf, 0xc4,           \
            0xce, 0xc1, 0xc3, 0xcf, 0xc5, 0xce, 0xc6, 0xc3,           \
            0x52, 0x58, 0x5c, 0x55, 0x5f, 0x53, 0x5a, 0x50,           \
            0x54, 0x5d, 0x59, 0x56, 0x51, 0x5e, 0x56, 0x59,           \
            0x5d, 0x52, 0x53, 0x5f, 0x50, 0x5c, 0x55, 0x5a,           \
            0x57, 0x5b, 0x5e, 0x51, 0x5b, 0x57, 0x58, 0x54,           \
            0x5b, 0x56, 0x57, 0x59, 0x52, 0x58, 0x54, 0x57,           \
            0x5d, 0x5b, 0x5a, 0x50, 0x58, 0x55, 0x51, 0x5c,           \
            0x50, 0x5d, 0x5c, 0x5a, 0x59, 0x52, 0x5f, 0x54,           \
            0x5e, 0x51, 0x53, 0x5f, 0x55, 0x5e, 0x56, 0x53,           \
            0x72, 0x78, 0x7c, 0x75, 0x7f, 0x73, 0x7a, 0x70,           \
            0x74, 0x7d, 0x79, 0x76, 0x71, 0x7e, 0x76, 0x79,           \
            0x7d, 0x72, 0x73, 0x7f, 0x70, 0x7c, 0x75, 0x7a,           \
            0x77, 0x7b, 0x7e, 0x71, 0x7b, 0x77, 0x78, 0x74,           \
            0x7b, 0x76, 0x77, 0x79, 0x72, 0x78, 0x74, 0x77,           \
            0x7d, 0x7b, 0x7a, 0x70, 0x78, 0x75, 0x71, 0x7c,           \
            0x70, 0x7d, 0x7c, 0x7a, 0x79, 0x72, 0x7f, 0x74,           \
            0x7e, 0x71, 0x73, 0x7f, 0x75, 0x7e, 0x76, 0x73,           \
            0xa2, 0xa8, 0xac, 0xa5, 0xaf, 0xa3, 0xaa, 0xa0,           \
            0xa4, 0xad, 0xa9, 0xa6, 0xa1, 0xae, 0xa6, 0xa9,           \
            0xad, 0xa2, 0xa3, 0xaf, 0xa0, 0xac, 0xa5, 0xaa,           \
            0xa7, 0xab, 0xae, 0xa1, 0xab, 0xa7, 0xa8, 0xa4,           \
            0xab, 0xa6, 0xa7, 0xa9, 0xa2, 0xa8, 0xa4, 0xa7,           \
            0xad, 0xab, 0xaa, 0xa0, 0xa8, 0xa5, 0xa1, 0xac,           \
            0xa0, 0xad, 0xac, 0xaa, 0xa9, 0xa2, 0xaf, 0xa4,           \
            0xae, 0xa1, 0xa3, 0xaf, 0xa5, 0xae, 0xa6, 0xa3,           \
            0x12, 0x18, 0x1c, 0x15, 0x1f, 0x13, 0x1a, 0x10,           \
            0x14, 0x1d, 0x19, 0x16, 0x11, 0x1e, 0x16, 0x19,           \
            0x1d, 0x12, 0x13, 0x1f, 0x10, 0x1c, 0x15, 0x1a,           \
            0x17, 0x1b, 0x1e, 0x11, 0x1b, 0x17, 0x18, 0x14,           \
            0x1b, 0x16, 0x17, 0x19, 0x12, 0x18, 0x14, 0x17,           \
            0x1d, 0x1b, 0x1a, 0x10, 0x18, 0x15, 0x11, 0x1c,           \
            0x10, 0x1d, 0x1c, 0x1a, 0x19, 0x12, 0x1f, 0x14,           \
            0x1e, 0x11, 0x13, 0x1f, 0x15, 0x1e, 0x16, 0x13,           \
            0xe2, 0xe8, 0xec, 0xe5, 0xef, 0xe3, 0xea, 0xe0,           \
            0xe4, 0xed, 0xe9, 0xe6, 0xe1, 0xee, 0xe6, 0xe9,           \
            0xed, 0xe2, 0xe3, 0xef, 0xe0, 0xec, 0xe5, 0xea,           \
            0xe7, 0xeb, 0xee, 0xe1, 0xeb, 0xe7, 0xe8, 0xe4,           \
            0xeb, 0xe6, 0xe7, 0xe9, 0xe2, 0xe8, 0xe4, 0xe7,           \
            0xed, 0xeb, 0xea, 0xe0, 0xe8, 0xe5, 0xe1, 0xec,           \
            0xe0, 0xed, 0xec, 0xea, 0xe9, 0xe2, 0xef, 0xe4,           \
            0xee, 0xe1, 0xe3, 0xef, 0xe5, 0xee, 0xe6, 0xe3,           \
            0xd2, 0xd8, 0xdc, 0xd5, 0xdf, 0xd3, 0xda, 0xd0,           \
            0xd4, 0xdd, 0xd9, 0xd6, 0xd1, 0xde, 0xd6, 0xd9,           \
            0xdd, 0xd2, 0xd3, 0xdf, 0xd0, 0xdc, 0xd5, 0xda,           \
            0xd7, 0xdb, 0xde, 0xd1, 0xdb, 0xd7, 0xd8, 0xd4,           \
            0xdb, 0xd6, 0xd7, 0xd9, 0xd2, 0xd8, 0xd4, 0xd7,           \
            0xdd, 0xdb, 0xda, 0xd0, 0xd8, 0xd5, 0xd1, 0xdc,           \
            0xd0, 0xdd, 0xdc, 0xda, 0xd9, 0xd2, 0xdf, 0xd4,           \
            0xde, 0xd1, 0xd3, 0xdf, 0xd5, 0xde, 0xd6, 0xd3,           \
            0x02, 0x08, 0x0c, 0x05, 0x0f, 0x03, 0x0a, 0x00,           \
            0x04, 0x0d, 0x09, 0x06, 0x01, 0x0e, 0x06, 0x09,           \
            0x0d, 0x02, 0x03, 0x0f, 0x00, 0x0c, 0x05, 0x0a,           \
            0x07, 0x0b, 0x0e, 0x01, 0x0b, 0x07, 0x08, 0x04,           \
            0x0b, 0x06, 0x07, 0x09, 0x02, 0x08, 0x04, 0x07,           \
            0x0d, 0x0b, 0x0a, 0x00, 0x08, 0x05, 0x01, 0x0c,           \
            0x00, 0x0d, 0x0c, 0x0a, 0x09, 0x02, 0x0f, 0x04,           \
            0x0e, 0x01, 0x03, 0x0f, 0x05, 0x0e, 0x06, 0x03,           \
            0x22, 0x28, 0x2c, 0x25, 0x2f, 0x23, 0x2a, 0x20,           \
            0x24, 0x2d, 0x29, 0x26, 0x21, 0x2e, 0x26, 0x29,           \
            0x2d, 0x22, 0x23, 0x2f, 0x20, 0x2c, 0x25, 0x2a,           \
            0x27, 0x2b, 0x2e, 0x21, 0x2b, 0x27, 0x28, 0x24,           \
            0x2b, 0x26, 0x27, 0x29, 0x22, 0x28, 0x24, 0x27,           \
            0x2d, 0x2b, 0x2a, 0x20, 0x28, 0x25, 0x21, 0x2c,           \
            0x20, 0x2d, 0x2c, 0x2a, 0x29, 0x22, 0x2f, 0x24,           \
            0x2e, 0x21, 0x23, 0x2f, 0x25, 0x2e, 0x26, 0x23,           \
            0xb2, 0xb8, 0xbc, 0xb5, 0xbf, 0xb3, 0xba, 0xb0,           \
            0xb4, 0xbd, 0xb9, 0xb6, 0xb1, 0xbe, 0xb6, 0xb9,           \
            0xbd, 0xb2, 0xb3, 0xbf, 0xb0, 0xbc, 0xb5, 0xba,           \
            0xb7, 0xbb, 0xbe, 0xb1, 0xbb, 0xb7, 0xb8, 0xb4,           \
            0xbb, 0xb6, 0xb7, 0xb9, 0xb2, 0xb8, 0xb4, 0xb7,           \
            0xbd, 0xbb, 0xba, 0xb0, 0xb8, 0xb5, 0xb1, 0xbc,           \
            0xb0, 0xbd, 0xbc, 0xba, 0xb9, 0xb2, 0xbf, 0xb4,           \
            0xbe, 0xb1, 0xb3, 0xbf, 0xb5, 0xbe, 0xb6, 0xb3,           \
            0x42, 0x48, 0x4c, 0x45, 0x4f, 0x43, 0x4a, 0x40,           \
            0x44, 0x4d, 0x49, 0x46, 0x41, 0x4e, 0x46, 0x49,           \
            0x4d, 0x42, 0x43, 0x4f, 0x40, 0x4c, 0x45, 0x4a,           \
            0x47, 0x4b, 0x4e, 0x41, 0x4b, 0x47, 0x48, 0x44,           \
            0x4b, 0x46, 0x47, 0x49, 0x42, 0x48, 0x44, 0x47,           \
            0x4d, 0x4b, 0x4a, 0x40, 0x48, 0x45, 0x41, 0x4c,           \
            0x40, 0x4d, 0x4c, 0x4a, 0x49, 0x42, 0x4f, 0x44,           \
            0x4e, 0x41, 0x43, 0x4f, 0x45, 0x4e, 0x46, 0x43,           \
            0xd2, 0xd8, 0xdc, 0xd5, 0xdf, 0xd3, 0xda, 0xd0,           \
            0xd4, 0xdd, 0xd9, 0xd6, 0xd1, 0xde, 0xd6, 0xd9,           \
            0xdd, 0xd2, 0xd3, 0xdf, 0xd0, 0xdc, 0xd5, 0xda,           \
            0xd7, 0xdb, 0xde, 0xd1, 0xdb, 0xd7, 0xd8, 0xd4,           \
            0xdb, 0xd6, 0xd7, 0xd9, 0xd2, 0xd8, 0xd4, 0xd7,           \
            0xdd, 0xdb, 0xda, 0xd0, 0xd8, 0xd5, 0xd1, 0xdc,           \
            0xd0, 0xdd, 0xdc, 0xda, 0xd9, 0xd2, 0xdf, 0xd4,           \
            0xde, 0xd1, 0xd3, 0xdf, 0xd5, 0xde, 0xd6, 0xd3}

/* K[0] */
#define KSHFL0_0 (vec_uchar16){1, 6, 4, 7, 6, 2, 4, 7, 0, 1, 2, 5, 0, 4, 3, 3}
#define KMASK0_0 (vec_uchar16){0x20, 0x10, 0x20, 0x08, 0x40, 0x40, 0x40, 0x40, 0x20, 0x40, 0x10, 0x20, 0x10, 0x10, 0x20, 0x40}
#define KSHFL0_1 (vec_uchar16){5, 7, 7, 0, 4, 3, 2, 5, 2, 3, 4, 6, 4, 0, 5, 3}
#define KMASK0_1 (vec_uchar16){0x08, 0x20, 0x10, 0x40, 0x08, 0x10, 0x20, 0x40, 0x02, 0x08, 0x01, 0x02, 0x04, 0x08, 0x01, 0x02}
#define KSHFL0_2 (vec_uchar16){0, 6, 2, 3, 7, 2, 4, 7, 1, 2, 5, 1, 1, 7, 6, 3}
#define KMASK0_2 (vec_uchar16){0x04, 0x04, 0x01, 0x04, 0x04, 0x04, 0x02, 0x01, 0x01, 0x08, 0x04, 0x02, 0x04, 0x02, 0x01, 0x01}

/* K[1] */
#define KSHFL1_0 (vec_uchar16){0, 5, 3, 6, 5, 1, 3, 6, 7, 0, 1, 4, 7, 3, 2, 2}
#define KMASK1_0 (vec_uchar16){0x20, 0x10, 0x20, 0x08, 0x40, 0x40, 0x40, 0x40, 0x10, 0x40, 0x10, 0x20, 0x08, 0x10, 0x20, 0x40}
#define KSHFL1_1 (vec_uchar16){4, 6, 6, 7, 7, 2, 1, 4, 1, 2, 3, 5, 3, 7, 4, 2}
#define KMASK1_1 (vec_uchar16){0x08, 0x20, 0x10, 0x20, 0x40, 0x10, 0x20, 0x40, 0x02, 0x08, 0x01, 0x02, 0x04, 0x01, 0x01, 0x02}
#define KSHFL1_2 (vec_uchar16){3, 5, 1, 2, 6, 1, 3, 6, 0, 1, 4, 0, 0, 6, 5, 2}
#define KMASK1_2 (vec_uchar16){0x08, 0x04, 0x01, 0x04, 0x04, 0x04, 0x02, 0x01, 0x01, 0x08, 0x04, 0x02, 0x04, 0x02, 0x01, 0x01}

/* K[2] */
#define KSHFL2_0 (vec_uchar16){6, 3, 1, 4, 3, 7, 1, 4, 5, 6, 7, 2, 5, 1, 0, 0}
#define KMASK2_0 (vec_uchar16){0x10, 0x10, 0x20, 0x08, 0x40, 0x20, 0x40, 0x40, 0x10, 0x20, 0x08, 0x20, 0x08, 0x10, 0x20, 0x40}
#define KSHFL2_1 (vec_uchar16){6, 4, 4, 5, 5, 0, 7, 2, 7, 0, 1, 3, 1, 5, 2, 0}
#define KMASK2_1 (vec_uchar16){0x40, 0x20, 0x10, 0x20, 0x40, 0x10, 0x10, 0x40, 0x04, 0x08, 0x01, 0x02, 0x04, 0x01, 0x01, 0x02}
#define KSHFL2_2 (vec_uchar16){1, 3, 7, 0, 4, 3, 1, 4, 6, 7, 2, 6, 2, 4, 3, 0}
#define KMASK2_2 (vec_uchar16){0x08, 0x04, 0x02, 0x04, 0x04, 0x08, 0x02, 0x01, 0x02, 0x01, 0x04, 0x04, 0x08, 0x02, 0x01, 0x01}

/* K[3] */
#define KSHFL3_0 (vec_uchar16){4, 1, 7, 6, 1, 5, 7, 2, 3, 4, 5, 0, 7, 7, 6, 6}
#define KMASK3_0 (vec_uchar16){0x10, 0x10, 0x10, 0x40, 0x40, 0x20, 0x20, 0x40, 0x10, 0x20, 0x08, 0x20, 0x40, 0x08, 0x10, 0x20}
#define KSHFL3_1 (vec_uchar16){4, 2, 2, 3, 3, 6, 5, 0, 5, 6, 7, 1, 3, 3, 0, 6}
#define KMASK3_1 (vec_uchar16){0x40, 0x20, 0x10, 0x20, 0x40, 0x08, 0x10, 0x40, 0x04, 0x01, 0x02, 0x02, 0x08, 0x01, 0x01, 0x04}
#define KSHFL3_2 (vec_uchar16){7, 1, 5, 2, 2, 1, 7, 2, 4, 5, 0, 4, 0, 2, 1, 6}
#define KMASK3_2 (vec_uchar16){0x01, 0x04, 0x02, 0x08, 0x04, 0x08, 0x04, 0x01, 0x02, 0x01, 0x04, 0x04, 0x08, 0x02, 0x01, 0x02}

/* K[4] */
#define KSHFL4_0 (vec_uchar16){2, 7, 5, 4, 7, 3, 5, 0, 1, 2, 7, 6, 5, 5, 4, 4}
#define KMASK4_0 (vec_uchar16){0x10, 0x08, 0x10, 0x40, 0x20, 0x20, 0x20, 0x40, 0x10, 0x20, 0x40, 0x10, 0x40, 0x08, 0x10, 0x20}
#define KSHFL4_1 (vec_uchar16){2, 0, 0, 1, 1, 4, 3, 6, 3, 4, 5, 7, 1, 1, 6, 4}
#define KMASK4_1 (vec_uchar16){0x40, 0x20, 0x10, 0x20, 0x40, 0x08, 0x10, 0x20, 0x04, 0x01, 0x02, 0x04, 0x08, 0x01, 0x02, 0x04}
#define KSHFL4_2 (vec_uchar16){5, 3, 3, 0, 0, 7, 5, 0, 2, 3, 2, 2, 6, 0, 7, 4}
#define KMASK4_2 (vec_uchar16){0x01, 0x08, 0x02, 0x08, 0x04, 0x01, 0x04, 0x01, 0x02, 0x01, 0x08, 0x04, 0x01, 0x02, 0x02, 0x02}

/* K[5] */
#define KSHFL5_0 (vec_uchar16){0, 5, 3, 2, 5, 1, 3, 6, 7, 0, 5, 4, 3, 7, 2, 2}
#define KMASK5_0 (vec_uchar16){0x10, 0x08, 0x10, 0x40, 0x20, 0x20, 0x20, 0x20, 0x08, 0x20, 0x40, 0x10, 0x40, 0x40, 0x10, 0x20}
#define KSHFL5_1 (vec_uchar16){0, 6, 6, 7, 7, 6, 1, 4, 1, 2, 3, 5, 7, 7, 4, 2}
#define KMASK5_1 (vec_uchar16){0x40, 0x10, 0x08, 0x10, 0x20, 0x40, 0x10, 0x20, 0x04, 0x01, 0x02, 0x04, 0x01, 0x02, 0x02, 0x04}
#define KSHFL5_2 (vec_uchar16){3, 1, 1, 6, 2, 5, 3, 6, 0, 1, 0, 0, 4, 6, 5, 2}
#define KMASK5_2 (vec_uchar16){0x01, 0x08, 0x02, 0x01, 0x08, 0x01, 0x04, 0x02, 0x02, 0x01, 0x08, 0x04, 0x01, 0x04, 0x02, 0x02}

/* K[6] */
#define KSHFL6_0 (vec_uchar16){6, 7, 1, 0, 3, 7, 1, 4, 5, 6, 3, 2, 1, 5, 0, 0}
#define KMASK6_0 (vec_uchar16){0x08, 0x40, 0x10, 0x40, 0x20, 0x10, 0x20, 0x20, 0x08, 0x10, 0x40, 0x10, 0x40, 0x40, 0x10, 0x20}
#define KSHFL6_1 (vec_uchar16){6, 4, 4, 5, 5, 4, 7, 2, 3, 0, 1, 3, 5, 5, 2, 0}
#define KMASK6_1 (vec_uchar16){0x20, 0x10, 0x08, 0x10, 0x20, 0x40, 0x08, 0x20, 0x08, 0x01, 0x02, 0x04, 0x01, 0x02, 0x02, 0x04}
#define KSHFL6_2 (vec_uchar16){1, 7, 7, 4, 0, 3, 1, 4, 6, 7, 6, 2, 2, 4, 3, 0}
#define KMASK6_2 (vec_uchar16){0x01, 0x01, 0x04, 0x01, 0x08, 0x01, 0x04, 0x02, 0x04, 0x02, 0x01, 0x08, 0x01, 0x04, 0x02, 0x02}

/* K[7] */
#define KSHFL7_0 (vec_uchar16){4, 5, 7, 6, 1, 5, 7, 2, 7, 4, 1, 0, 7, 3, 6, 6}
#define KMASK7_0 (vec_uchar16){0x08, 0x40, 0x08, 0x20, 0x20, 0x10, 0x10, 0x20, 0x40, 0x10, 0x40, 0x10, 0x20, 0x40, 0x08, 0x10}
#define KSHFL7_1 (vec_uchar16){4, 2, 6, 3, 3, 2, 5, 0, 1, 6, 7, 1, 3, 3, 0, 2}
#define KMASK7_1 (vec_uchar16){0x20, 0x10, 0x40, 0x10, 0x20, 0x40, 0x08, 0x20, 0x08, 0x02, 0x04, 0x04, 0x01, 0x02, 0x02, 0x08}
#define KSHFL7_2 (vec_uchar16){7, 5, 5, 2, 6, 1, 3, 2, 4, 5, 4, 0, 0, 2, 1, 6}
#define KMASK7_2 (vec_uchar16){0x02, 0x01, 0x04, 0x01, 0x01, 0x01, 0x08, 0x02, 0x04, 0x02, 0x01, 0x08, 0x01, 0x04, 0x02, 0x04}

/* K[8] */
#define KSHFL8_0 (vec_uchar16){7, 4, 6, 5, 0, 4, 6, 1, 6, 3, 0, 7, 6, 2, 5, 5}
#define KMASK8_0 (vec_uchar16){0x40, 0x40, 0x08, 0x20, 0x20, 0x10, 0x10, 0x20, 0x40, 0x10, 0x40, 0x08, 0x20, 0x40, 0x08, 0x10}
#define KSHFL8_1 (vec_uchar16){3, 1, 5, 2, 2, 1, 4, 7, 0, 5, 6, 0, 2, 2, 7, 1}
#define KMASK8_1 (vec_uchar16){0x20, 0x10, 0x40, 0x10, 0x20, 0x40, 0x08, 0x10, 0x08, 0x02, 0x04, 0x04, 0x01, 0x02, 0x04, 0x08}
#define KSHFL8_2 (vec_uchar16){6, 4, 4, 1, 5, 0, 2, 1, 3, 4, 3, 7, 7, 1, 0, 5}
#define KMASK8_2 (vec_uchar16){0x02, 0x01, 0x04, 0x01, 0x01, 0x01, 0x08, 0x02, 0x04, 0x02, 0x01, 0x01, 0x02, 0x04, 0x02, 0x04}

/* K[9] */
#define KSHFL9_0 (vec_uchar16){5, 2, 4, 3, 6, 2, 4, 7, 4, 1, 6, 5, 4, 0, 7, 3}
#define KMASK9_0 (vec_uchar16){0x40, 0x40, 0x08, 0x20, 0x10, 0x10, 0x10, 0x10, 0x40, 0x10, 0x20, 0x08, 0x20, 0x40, 0x40, 0x10}
#define KSHFL9_1 (vec_uchar16){1, 7, 3, 0, 0, 7, 6, 5, 6, 3, 4, 2, 0, 0, 5, 7}
#define KMASK9_1 (vec_uchar16){0x20, 0x08, 0x40, 0x10, 0x20, 0x20, 0x40, 0x10, 0x01, 0x02, 0x04, 0x08, 0x01, 0x02, 0x04, 0x01}
#define KSHFL9_2 (vec_uchar16){4, 2, 2, 7, 3, 6, 0, 7, 1, 2, 1, 5, 5, 3, 6, 3}
#define KMASK9_2 (vec_uchar16){0x02, 0x01, 0x04, 0x02, 0x01, 0x02, 0x08, 0x04, 0x04, 0x02, 0x01, 0x01, 0x02, 0x08, 0x04, 0x04}

/* K[10] */
#define KSHFL10_0 (vec_uchar16){3, 0, 6, 1, 4, 0, 2, 5, 2, 7, 4, 7, 2, 6, 5, 1}
#define KMASK10_0 (vec_uchar16){0x40, 0x40, 0x40, 0x20, 0x10, 0x10, 0x10, 0x10, 0x40, 0x08, 0x20, 0x40, 0x20, 0x20, 0x40, 0x10}
#define KSHFL10_1 (vec_uchar16){7, 5, 1, 6, 6, 5, 4, 3, 4, 1, 2, 0, 6, 6, 3, 5}
#define KMASK10_1 (vec_uchar16){0x10, 0x08, 0x40, 0x08, 0x10, 0x20, 0x40, 0x10, 0x01, 0x02, 0x04, 0x08, 0x02, 0x04, 0x04, 0x01}
#define KSHFL10_2 (vec_uchar16){2, 0, 0, 5, 1, 4, 6, 5, 3, 0, 7, 3, 3, 1, 4, 1}
#define KMASK10_2 (vec_uchar16){0x02, 0x01, 0x04, 0x02, 0x01, 0x02, 0x01, 0x04, 0x08, 0x02, 0x02, 0x01, 0x02, 0x08, 0x04, 0x04}

/* K[11] */
#define KSHFL11_0 (vec_uchar16){1, 6, 4, 7, 2, 6, 0, 3, 0, 5, 2, 5, 0, 4, 3, 7}
#define KMASK11_0 (vec_uchar16){0x40, 0x20, 0x40, 0x10, 0x10, 0x08, 0x10, 0x10, 0x40, 0x08, 0x20, 0x40, 0x20, 0x20, 0x40, 0x08}
#define KSHFL11_1 (vec_uchar16){5, 7, 7, 4, 4, 3, 2, 1, 2, 7, 0, 6, 4, 4, 1, 3}
#define KMASK11_1 (vec_uchar16){0x10, 0x40, 0x20, 0x08, 0x10, 0x20, 0x40, 0x10, 0x01, 0x04, 0x04, 0x01, 0x02, 0x04, 0x04, 0x01}
#define KSHFL11_2 (vec_uchar16){0, 6, 2, 3, 7, 2, 4, 3, 1, 6, 5, 1, 1, 7, 2, 3}
#define KMASK11_2 (vec_uchar16){0x02, 0x02, 0x08, 0x02, 0x02, 0x02, 0x01, 0x04, 0x08, 0x04, 0x02, 0x01, 0x02, 0x01, 0x04, 0x08}

/* K[12] */
#define KSHFL12_0 (vec_uchar16){7, 4, 2, 5, 0, 4, 6, 1, 6, 7, 0, 3, 6, 2, 1, 5}
#define KMASK12_0 (vec_uchar16){0x20, 0x20, 0x40, 0x10, 0x10, 0x08, 0x08, 0x10, 0x20, 0x40, 0x20, 0x40, 0x10, 0x20, 0x40, 0x08}
#define KSHFL12_1 (vec_uchar16){3, 5, 5, 6, 2, 1, 0, 7, 0, 5, 2, 4, 2, 2, 3, 1}
#define KMASK12_1 (vec_uchar16){0x10, 0x40, 0x20, 0x40, 0x10, 0x20, 0x40, 0x08, 0x01, 0x04, 0x08, 0x01, 0x02, 0x04, 0x08, 0x01}
#define KSHFL12_2 (vec_uchar16){6, 4, 0, 1, 5, 0, 2, 1, 7, 4, 3, 7, 7, 5, 0, 1}
#define KMASK12_2 (vec_uchar16){0x04, 0x02, 0x08, 0x02, 0x02, 0x02, 0x01, 0x04, 0x01, 0x04, 0x02, 0x02, 0x04, 0x01, 0x04, 0x08}

/* K[13] */
#define KSHFL13_0 (vec_uchar16){5, 2, 0, 3, 6, 6, 4, 7, 4, 5, 6, 1, 4, 0, 7, 7}
#define KMASK13_0 (vec_uchar16){0x20, 0x20, 0x40, 0x10, 0x08, 0x40, 0x08, 0x08, 0x20, 0x40, 0x10, 0x40, 0x10, 0x20, 0x20, 0x40}
#define KSHFL13_1 (vec_uchar16){1, 3, 3, 4, 0, 7, 6, 5, 6, 3, 0, 2, 0, 0, 1, 7}
#define KMASK13_1 (vec_uchar16){0x10, 0x40, 0x20, 0x40, 0x10, 0x10, 0x20, 0x08, 0x02, 0x04, 0x08, 0x01, 0x02, 0x04, 0x08, 0x02}
#define KSHFL13_2 (vec_uchar16){4, 2, 6, 7, 3, 6, 0, 3, 5, 2, 1, 5, 5, 3, 2, 7}
#define KMASK13_2 (vec_uchar16){0x04, 0x02, 0x01, 0x04, 0x02, 0x04, 0x01, 0x08, 0x01, 0x04, 0x02, 0x02, 0x04, 0x01, 0x08, 0x01}

/* K[14] */
#define KSHFL14_0 (vec_uchar16){3, 0, 6, 1, 4, 4, 6, 5, 2, 3, 4, 7, 2, 6, 5, 5}
#define KMASK14_0 (vec_uchar16){0x20, 0x20, 0x20, 0x10, 0x08, 0x40, 0x40, 0x08, 0x20, 0x40, 0x10, 0x20, 0x10, 0x10, 0x20, 0x40}
#define KSHFL14_1 (vec_uchar16){7, 1, 1, 2, 6, 5, 4, 7, 4, 1, 6, 0, 6, 2, 7, 5}
#define KMASK14_1 (vec_uchar16){0x08, 0x40, 0x20, 0x40, 0x08, 0x10, 0x20, 0x40, 0x02, 0x04, 0x01, 0x01, 0x04, 0x08, 0x01, 0x02}
#define KSHFL14_2 (vec_uchar16){2, 0, 4, 5, 1, 4, 6, 1, 3, 0, 7, 3, 3, 1, 0, 5}
#define KMASK14_2 (vec_uchar16){0x04, 0x02, 0x01, 0x04, 0x02, 0x04, 0x02, 0x08, 0x01, 0x04, 0x04, 0x02, 0x04, 0x01, 0x08, 0x01}

/* K[15] */
#define KSHFL15_0 (vec_uchar16){2, 7, 5, 0, 7, 3, 5, 4, 1, 2, 3, 6, 1, 5, 4, 4}
#define KMASK15_0 (vec_uchar16){0x20, 0x10, 0x20, 0x10, 0x40, 0x40, 0x40, 0x08, 0x20, 0x40, 0x10, 0x20, 0x10, 0x10, 0x20, 0x40}
#define KSHFL15_1 (vec_uchar16){6, 0, 0, 1, 5, 4, 3, 6, 3, 0, 5, 7, 5, 1, 6, 4}
#define KMASK15_1 (vec_uchar16){0x08, 0x40, 0x20, 0x40, 0x08, 0x10, 0x20, 0x40, 0x02, 0x04, 0x01, 0x02, 0x04, 0x08, 0x01, 0x02}
#define KSHFL15_2 (vec_uchar16){1, 7, 3, 4, 0, 3, 5, 0, 2, 3, 6, 2, 2, 0, 7, 4}
#define KMASK15_2 (vec_uchar16){0x04, 0x04, 0x01, 0x04, 0x02, 0x04, 0x02, 0x08, 0x01, 0x08, 0x04, 0x02, 0x04, 0x01, 0x01, 0x01}

#define IPSHFL (vec_uchar16){7, 6, 5, 4, 3, 2, 1, 0, 7, 6, 5, 4, 3, 2, 1, 0}
#define IPLMASK0 (vec_uchar16){0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08}
#define IPLMASK1 (vec_uchar16){0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80}
#define IPRMASK0 (vec_uchar16){0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04}
#define IPRMASK1 (vec_uchar16){0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40}


#define FPSHFL (vec_uchar16){4, 0, 5, 1, 6, 2, 7, 3, 4, 0, 5, 1, 6, 2, 7, 3}
#define FPMASK0 (vec_uchar16){0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40}
#define FPMASK1 (vec_uchar16){0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10}
#define FPMASK2 (vec_uchar16){0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04}
#define FPMASK3 (vec_uchar16){0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01}


#define ESHFL0_BASE (vec_uchar16){3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 1, 1}
#define ESHFL1_BASE (vec_uchar16){1, 1, 1, 1, 1, 1, 1, 2, 9, 9, 9, 9, 9, 9, 9, 9}
#define ESHFL2_BASE (vec_uchar16){1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 2, 3, 3, 3}
#define ESHFL3_BASE (vec_uchar16){3, 3, 3, 3, 3, 3, 3, 0, 9, 9, 9, 9, 9, 9, 9, 9}
#define EMASK0_BASE (vec_uchar16){0x80, 0x01, 0x02, 0x04, 0x08, 0x10, 0x08, 0x10, 0x20, 0x40, 0x80, 0x01, 0x80, 0x01, 0x02, 0x04}
#define EMASK1_BASE (vec_uchar16){0x08, 0x10, 0x08, 0x10, 0x20, 0x40, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}
#define EMASK2_BASE (vec_uchar16){0x80, 0x01, 0x02, 0x04, 0x08, 0x10, 0x08, 0x10, 0x20, 0x40, 0x80, 0x01, 0x80, 0x01, 0x02, 0x04}
#define EMASK3_BASE (vec_uchar16){0x08, 0x10, 0x08, 0x10, 0x20, 0x40, 0x80, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}

#define PSHFL0 (vec_uchar16){1, 0, 2, 2, 3, 1, 3, 2, 0, 1, 2, 3, 0, 2, 3, 1}
#define PSHFL1 (vec_uchar16){0, 0, 2, 1, 3, 3, 0, 1, 2, 1, 3, 0, 2, 1, 0, 3}
#define PMASK0 (vec_uchar16){0x80, 0x40, 0x08, 0x10, 0x10, 0x08, 0x08, 0x01, 0x01, 0x40, 0x40, 0x02, 0x10, 0x02, 0x40, 0x02}
#define PMASK1 (vec_uchar16){0x02, 0x80, 0x80, 0x20, 0x80, 0x04, 0x04, 0x01, 0x04, 0x10, 0x20, 0x20, 0x20, 0x04, 0x08, 0x01}


#endif // _SPE_CRYPT_H_
